import click
import paramiko
from paramiko import SSHClient
from paramiko.rsakey import RSAKey
from scp import SCPClient

from aws.cli.utils import (
    exec_command,
    key_file_path,
    username,
    files_to_upload,
    target_folder,
    poetry_install_command,
    poetry_path, tensorboard_install_command,
)
from src.utils.logger import logger

log = logger()


@click.command()
@click.option("--instance-ip", required=True, help="IP address of EC2 instance")
@click.option("--username", default=username, help="Username of a user to log in with")
@click.option("--key-file-path", default=key_file_path, help="Path to SSH key file")
@click.option(
    "--files-to-upload",
    multiple=True,
    default=files_to_upload,
    help="Files to sync to EC2 instance",
)
@click.option(
    "--target-folder", default=target_folder, help="Target folder on EC2 instance"
)
def prepare_instance(
    instance_ip, username, key_file_path, files_to_upload, target_folder
):
    log.info(f"Connecting to: {instance_ip}")

    ssh = SSHClient()
    ssh.set_missing_host_key_policy(paramiko.AutoAddPolicy())
    key = RSAKey.from_private_key_file(key_file_path)
    ssh.connect(hostname=instance_ip, username=username, pkey=key)
    log.info(f"Connected!")

    log.info(f"Install Poetry")
    exec_command(ssh, log, poetry_install_command)

    log.info(f"Create target folder: {target_folder}")
    exec_command(ssh, log, f"mkdir -p {target_folder}")

    log.info(f"Upload project files")
    scp = SCPClient(ssh.get_transport())
    for file in files_to_upload:
        log.info(f"Uploading: {file}")
        scp.put(file, recursive=True, remote_path=target_folder)
    scp.close()

    log.info(f"Installing dependencies")
    exec_command(ssh, log, f"cd {target_folder} && {poetry_path} install")

    log.info(f"Install Tensorboad")
    exec_command(ssh, log, tensorboard_install_command)

    log.info("All done!")


if __name__ == "__main__":
    prepare_instance()
