import copy
import datetime
import json
import os
import random
import time
from itertools import combinations

import click
import torch
from torch import nn

from experiments.e1.src.encoder import NpEncoder
from experiments.e1.src.evaluation import (
    EvaluationConfig,
    TransformerType,
    evaluate_population,
    genotype_difference_factor,
)
from experiments.e1.src.genotype import initialize_population
from experiments.e1.src.operators import random_parameter_mutation, uniform_crossover
from experiments.e1.src.selection import tournament_selection, SelectionMode
from experiments.e1.src.visualization import simple_plot
from src.dataset.DatasetType import DatasetType
from src.dataset.LibraDataset import LibraCategory
from src.evolutionary.dataset.bootstrap import (
    boostrap_synthetic_dataset,
    boostrap_libra_dataset,
)
from src.utils.helpers import mean

# What to evolve #
# * embedding dim
# * patch size
# * head count
# * number of layers
# * forward expansion
# * dropout


def get_sample_config_libra():
    input_dim = 200
    output_dim = 50
    train_test_ratio = 0.8
    dataset_length = 0
    dataset_type = DatasetType.LIBRA
    libra_categories = [
        LibraCategory.ECONOMICS,
        LibraCategory.FINANCE,
        LibraCategory.HUMAN,
        LibraCategory.NATURE,
    ]
    dataset_filter = [
        [71, 86, 69, 75, 70, 83, 80, 74, 81, 77],
        [63, 96, 49, 39, 71, 18, 99, 66, 82, 17],
        [86, 56, 46, 21, 67, 5, 2, 92, 68, 97],
        [0, 91, 39, 70, 23, 42, 73, 1, 53, 24],
    ]
    min_training_examples = 100
    max_values_per_dataset = 5000

    transformer_type = TransformerType.ENCODER_ONLY
    train_from_scratch = False
    batch_size = 16
    split_size = 0
    epochs = 1

    iterations = 1
    population_count = 4
    tournament_size = 2
    mutation_probability = 0.1
    crossover_probability = 0.3

    config = {
        "dataset_type": dataset_type.value,
        "dataset_length": dataset_length,
        "train_test_ratio": train_test_ratio,
        "libra_categories": libra_categories,
        "libra_dataset_filter": dataset_filter,
        "libra_min_training_examples": min_training_examples,
        "libra_max_values_per_dataset": max_values_per_dataset,
        "input_dim": input_dim,
        "output_dim": output_dim,
        "batch_size": batch_size,
        "split_size": split_size,
        "epochs": epochs,
        "iterations": iterations,
        "population_count": population_count,
        "tournament_size": tournament_size,
        "mutation_probability": mutation_probability,
        "crossover_probability": crossover_probability,
        "transformer_type": transformer_type,
        "train_from_scratch": train_from_scratch,
    }
    return config


def get_sample_config_synthetic():
    input_dim = 200
    output_dim = 50
    dataset_length = 5000
    train_test_ratio = 0.8
    dataset_type = DatasetType.SYNTHETIC

    transformer_type = TransformerType.ENCODER_ONLY
    train_from_scratch = False
    batch_size = 16
    split_size = 0
    epochs = 1

    iterations = 1
    population_count = 4
    tournament_size = 2
    mutation_probability = 0.1
    crossover_probability = 0.3

    config = {
        "dataset_type": dataset_type.value,
        "synthetic_length": dataset_length,
        "train_test_ratio": train_test_ratio,
        "input_dim": input_dim,
        "output_dim": output_dim,
        "batch_size": batch_size,
        "split_size": split_size,
        "epochs": epochs,
        "iterations": iterations,
        "population_count": population_count,
        "tournament_size": tournament_size,
        "mutation_probability": mutation_probability,
        "crossover_probability": crossover_probability,
        "transformer_type": transformer_type,
        "train_from_scratch": train_from_scratch,
    }
    return config


@click.command()
@click.option("--config-path", help="Config json path")
def main(config_path: str):
    if config_path is None or config_path == "":
        print("*** Default config ***")
        config = get_sample_config_synthetic()
    else:
        print("*** Loaded config ***")
        file = open(config_path)
        config = json.load(file)

    print(json.dumps(config, indent=4))

    if config["dataset_type"] == DatasetType.SYNTHETIC.value:
        train_set, test_set = boostrap_synthetic_dataset(
            config["input_dim"],
            config["output_dim"],
            config["synthetic_length"],
            config["train_test_ratio"],
        )
    elif config["dataset_type"] == DatasetType.LIBRA.value:
        train_set, test_set = boostrap_libra_dataset(
            config["input_dim"],
            config["output_dim"],
            config["libra_categories"],
            config["libra_dataset_filter"],
            config["libra_min_training_examples"],
            config["libra_max_values_per_dataset"],
            config["train_test_ratio"],
        )
    else:
        raise ValueError("Unknown dataset type")

    # HERE THE MAGIC HAPPENS
    run_evolution(config, train_set, test_set)


def run_evolution(config: dict, train_set, test_set):
    epochs = config["epochs"]
    input_dim = config["input_dim"]
    output_dim = config["output_dim"]
    split_size = config["split_size"]
    batch_size = config["batch_size"]
    train_from_scratch = config["train_from_scratch"]

    iterations = config["iterations"]
    population_count = config["population_count"]
    tournament_size = config["tournament_size"]
    transformer_type = config["transformer_type"]
    mutation_probability = config["mutation_probability"]
    crossover_probability = config["crossover_probability"]

    # Training
    criterion = nn.SmoothL1Loss(reduction="sum")
    optimizer = torch.optim.AdamW(
        [torch.zeros(1, 1)], lr=0, betas=(0.9, 0.98), eps=1e-9
    )
    device = "cuda" if torch.cuda.is_available() else "cpu"
    evaluation_config = EvaluationConfig(
        batch_size,
        input_dim,
        output_dim,
        epochs,
        train_set,
        test_set,
        criterion,
        optimizer,
        device,
        split_size,
    )

    # Population
    population = initialize_population(population_count)

    # Stats
    parent_fitness = []
    survivor_fitness = []
    difference_factor = []

    # Evolution
    evolution_start = time.time()
    iteration_duration_history = []
    for i in range(iterations):
        iteration_start = time.time()
        iterations_left = iterations - (i + 1)
        mean_iteration_duration = mean(iteration_duration_history)
        evolution_time_left = mean_iteration_duration * iterations_left

        print("#######################################################")
        print(
            f"###### Iteration: {i + 1}, time left: {evolution_time_left:.2f}s ({(evolution_time_left/3600):.2f}h) ######"
        )
        print("#######################################################")

        # Select parents
        _, parent1 = tournament_selection(
            population,
            tournament_size,
            SelectionMode.BEST_FITNESS,
            transformer_type,
            evaluation_config,
            train_from_scratch,
        )
        _, parent2 = tournament_selection(
            population,
            tournament_size,
            SelectionMode.BEST_FITNESS,
            transformer_type,
            evaluation_config,
            train_from_scratch,
        )

        # Mean fitness of parents
        parent_fitness.append(mean([parent1.fitness(), parent2.fitness()]))

        # Diversity in population
        pairs = list(combinations(population, 2))
        factors = [genotype_difference_factor(g1, g2) for g1, g2 in pairs]
        factor_mean = mean(factors)
        difference_factor.append(factor_mean)

        # Crossover and mutation
        child1 = copy.deepcopy(parent1)
        child2 = copy.deepcopy(parent2)

        if random.random() < crossover_probability:
            child1, child2 = uniform_crossover(parent1, parent2)

        if random.random() < mutation_probability:
            random_parameter_mutation(child1)

        if random.random() < mutation_probability:
            random_parameter_mutation(child2)

        # Let parents compete with children
        _, survivor1 = tournament_selection(
            [parent1, child1],
            2,
            SelectionMode.BEST_FITNESS,
            transformer_type,
            evaluation_config,
            train_from_scratch,
        )
        _, survivor2 = tournament_selection(
            [parent2, child2],
            2,
            SelectionMode.BEST_FITNESS,
            transformer_type,
            evaluation_config,
            train_from_scratch,
        )

        # Remove individuals with the worst fitness
        to_remove_index1, _ = tournament_selection(
            population,
            tournament_size,
            SelectionMode.WORST_FITNESS,
            transformer_type,
            evaluation_config,
            train_from_scratch,
        )
        population.pop(to_remove_index1)

        to_remove_index2, _ = tournament_selection(
            population,
            tournament_size,
            SelectionMode.WORST_FITNESS,
            transformer_type,
            evaluation_config,
            train_from_scratch,
        )
        population.pop(to_remove_index2)

        # Append survivors to population
        population.append(survivor1)
        population.append(survivor2)

        # Mean fitness of survivors
        survivor_fitness.append(mean([survivor1.fitness(), survivor2.fitness()]))

        # Save and print results
        if i == iterations - 1:
            evaluate_population(
                population, transformer_type, evaluation_config, train_from_scratch
            )
            sorted_population = list(
                sorted(population, key=lambda g: g.fitness(), reverse=True)
            )

            # Save results
            save_results_as_json(
                sorted_population, survivor_fitness, difference_factor, config
            )

            # Show results
            simple_plot(survivor_fitness)
            simple_plot(difference_factor)

        iteration_duration = time.time() - iteration_start
        iteration_duration_history.append(iteration_duration)

    evolution_duration = time.time() - evolution_start
    print(f"### Evolution took: {evolution_duration:.2f}s ###")


def save_results_as_json(
    sorted_population,
    survivor_fitness,
    difference_factor,
    config,
    top_k_genotypes: int = 10,
    directory: str = "./results",
):
    timestamp = datetime.datetime.now().strftime("%d-%m-%Y_%H:%M:%S")
    if not os.path.exists(directory):
        os.mkdir(directory)

    best_genotype = sorted_population[0].to_dict()
    top_genotypes = [g.to_dict() for g in sorted_population[:top_k_genotypes]]

    data = {
        "config": config,
        "best_genotype": best_genotype,
        "top_genotypes": top_genotypes,
        "survivor_fitness": survivor_fitness,
        "difference_factor": difference_factor,
    }

    filename = os.path.join(directory, f"{timestamp}.json")
    with open(filename, "w") as outfile:
        json.dump(data, outfile, cls=NpEncoder, indent=4)


if __name__ == "__main__":
    main()
