import time
import unittest
from dataclasses import dataclass
from enum import Enum
from typing import List, Tuple

from torch import nn

from experiments.e1.src.genotype import Genotype
from experiments.e1.src.operators import random_parameter_mutation
from src.training.NoamOptimizer import NoamOptimizer
from src.training.TransformerLossCompute import TransformerLossCompute
from src.transformers.decoder_only.TimeSeriesDecoderTransformer import (
    TimeSeriesDecoderTransformer,
)
from src.transformers.decoder_only.train_decoder_only import (
    run_epoch as run_epoch_do,
    batch_iterator as batch_iterator_do,
)
from src.transformers.encoder_decoder.TimeSeriesTransformer import TimeSeriesTransformer
from src.transformers.encoder_decoder.train_encoder_decoder import (
    run_epoch as run_epoch_ed,
    batch_iterator as batch_iterator_ed,
)
from src.transformers.encoder_decoder.train_encoder_decoder_informer import (
    run_epoch as run_epoch_edi,
    batch_iterator as batch_iterator_edi,
)
from src.transformers.encoder_only.TimeSeriesEncoderTransformer import (
    TimeSeriesEncoderTransformer,
)
from src.transformers.encoder_only.train_encoder_only import (
    run_epoch as run_epoch_eo,
    batch_iterator as batch_iterator_eo,
)
from src.utils.helpers import is_nan


class TransformerType(str, Enum):
    ENCODER_DECODER_INFORMER = "encoder_decoder_informer"
    ENCODER_DECODER = "encoder_decoder"
    DECODER_ONLY = "decoder_only"
    ENCODER_ONLY = "encoder_only"
    ENCODER_ONLY_SIMPLE = "encoder_only_simple"


def genotype_to_phenotype(
    genotype: Genotype,
    transformer_type: TransformerType,
    input_dim=None,
    output_dim=None,
) -> nn.Module:
    embedding_dim = genotype.chromosome(0).value()
    patch_size = genotype.chromosome(1).value()
    head_count = genotype.chromosome(2).value()
    layer_count = genotype.chromosome(3).value()
    forward_expansion = genotype.chromosome(4).value()
    dropout = genotype.chromosome(5).value()

    transformer = None
    if (
        transformer_type == TransformerType.ENCODER_DECODER
        or transformer_type == TransformerType.ENCODER_DECODER_INFORMER
    ):
        transformer = TimeSeriesTransformer.create(
            layer_count,
            layer_count,
            embedding_dim,
            forward_expansion,
            head_count,
            dropout,
            patch_size,
        )
    elif transformer_type == TransformerType.DECODER_ONLY:
        transformer = TimeSeriesDecoderTransformer.create(
            layer_count,
            embedding_dim,
            forward_expansion,
            head_count,
            dropout,
            patch_size,
        )
    elif (
        transformer_type == TransformerType.ENCODER_ONLY
        or transformer_type == TransformerType.ENCODER_ONLY_SIMPLE
    ):
        use_full_regression = transformer_type == TransformerType.ENCODER_ONLY
        transformer = TimeSeriesEncoderTransformer.create(
            layer_count,
            embedding_dim,
            input_dim,
            output_dim,
            forward_expansion,
            head_count,
            dropout,
            patch_size,
            use_full_regression=use_full_regression,
        )
    return transformer


@dataclass
class EvaluationConfig:
    batch_size: int
    input_dim: int
    output_dim: int
    epochs: int
    train_set: list
    test_set: list
    criterion: any
    optimizer: any
    device: any
    split_size: int


def evaluate_transformer(
    model: nn.Module, transformer_type: TransformerType, config: EvaluationConfig
) -> Tuple[List[float], List[float]]:
    # NOTE: Very important!!! Add model parameters to the optimizer
    config.optimizer.add_param_group({"params": [p for p in model.parameters()]})

    scheduled_optimizer = NoamOptimizer(
        config.optimizer,
        model.embedding_dim(),
        1000,
    )
    scheduled_optimizer.reset()  # Reset the warmup steps

    # Train
    eval_loss_per_epoch = []
    train_loss_per_epoch = []
    model = model.to(config.device)
    for epoch in range(config.epochs):
        print(f"$$$ Epoch: {epoch + 1} $$$")
        start = time.time()

        model.train()
        patch_size = model.patch_size()
        if transformer_type == TransformerType.ENCODER_DECODER_INFORMER:
            train_loss = run_epoch_edi(
                batch_iterator_edi(
                    config.train_set,
                    config.batch_size,
                    config.split_size,
                    config.device,
                ),
                model,
                TransformerLossCompute(
                    model.get_regression(), config.criterion, scheduled_optimizer
                ),
            )
        elif transformer_type == TransformerType.ENCODER_DECODER:
            train_loss = run_epoch_ed(
                batch_iterator_ed(
                    config.train_set, config.batch_size, patch_size, config.device
                ),
                model,
                TransformerLossCompute(
                    model.get_regression(), config.criterion, scheduled_optimizer
                ),
            )
        elif transformer_type == TransformerType.DECODER_ONLY:
            train_loss = run_epoch_do(
                batch_iterator_do(
                    config.train_set, config.batch_size, patch_size, config.device
                ),
                model,
                TransformerLossCompute(
                    model.get_regression(), config.criterion, scheduled_optimizer
                ),
            )
        elif (
            transformer_type == TransformerType.ENCODER_ONLY
            or transformer_type == TransformerType.ENCODER_ONLY_SIMPLE
        ):
            train_loss = run_epoch_eo(
                batch_iterator_eo(
                    config.train_set, config.batch_size, config.input_dim, config.device
                ),
                model,
                TransformerLossCompute(
                    model.get_regression(), config.criterion, scheduled_optimizer
                ),
                0,
            )
        else:
            raise ValueError("Unknown transformer type")
        print(f">> Train acc loss: {train_loss}")

        model.eval()
        if transformer_type == TransformerType.ENCODER_DECODER_INFORMER:
            eval_loss = run_epoch_edi(
                batch_iterator_edi(
                    config.test_set, config.batch_size, config.split_size, config.device
                ),
                model,
                TransformerLossCompute(model.get_regression(), config.criterion, None),
            )
        elif transformer_type == TransformerType.ENCODER_DECODER:
            eval_loss = run_epoch_ed(
                batch_iterator_ed(
                    config.test_set, config.batch_size, patch_size, config.device
                ),
                model,
                TransformerLossCompute(model.get_regression(), config.criterion, None),
            )
        elif transformer_type == TransformerType.DECODER_ONLY:
            eval_loss = run_epoch_do(
                batch_iterator_do(
                    config.test_set, config.batch_size, patch_size, config.device
                ),
                model,
                TransformerLossCompute(model.get_regression(), config.criterion, None),
            )
        elif (
            transformer_type == TransformerType.ENCODER_ONLY
            or transformer_type == TransformerType.ENCODER_ONLY_SIMPLE
        ):
            eval_loss = run_epoch_eo(
                batch_iterator_eo(
                    config.test_set, config.batch_size, config.input_dim, config.device
                ),
                model,
                TransformerLossCompute(model.get_regression(), config.criterion, None),
                0,
            )
        else:
            raise ValueError("Unknown transformer type")
        print(f">> Eval acc loss: {eval_loss}")
        print(f">>> Epoch took: {time.time() - start}s")

        # If the resulting loss is NaN, set it to inf to signify bad model
        train_loss_value = train_loss.detach().cpu().numpy()
        if is_nan(train_loss_value):
            train_loss_value = float("inf")

        eval_loss_value = eval_loss.detach().cpu().numpy()
        if is_nan(eval_loss_value):
            eval_loss_value = float("inf")

        train_loss_per_epoch.append(float(train_loss_value))
        eval_loss_per_epoch.append(float(eval_loss_value))

    # NOTE: Very important!!! Resets optimizer and removes the current model parameters from it!
    config.optimizer.param_groups.clear()  # optim.param_groups = []
    config.optimizer.state.clear()  # optim.state = defaultdict(dict)
    config.optimizer.zero_grad()

    return train_loss_per_epoch, eval_loss_per_epoch


def evaluate_population(
    genotypes: List[Genotype],
    transformer_type: TransformerType,
    config: EvaluationConfig,
    retrain: bool,
):
    for genotype in genotypes:
        if genotype.is_dirty() or retrain:
            phenotype = genotype_to_phenotype(
                genotype, transformer_type, config.input_dim, config.output_dim
            )
            train_loss_list, eval_loss_list = evaluate_transformer(
                phenotype, transformer_type, config
            )
            genotype.set_loss(train_loss_list[-1], eval_loss_list[-1])


def genotype_difference_factor(genotype1: Genotype, genotype2: Genotype) -> float:
    if genotype1.chromosome_length() != genotype2.chromosome_length():
        raise ValueError("Chromosomes need to have an equal length")

    factor = 0
    for idx in range(genotype1.chromosome_length()):
        parameter1 = genotype1.chromosome(idx)
        parameter2 = genotype2.chromosome(idx)
        allowed_values = parameter1.get_allowed_values()
        parameter1_index = allowed_values.index(parameter1.value())
        parameter2_index = allowed_values.index(parameter2.value())
        factor += abs(parameter1_index - parameter2_index) / (len(allowed_values) - 1)

    factor /= genotype1.chromosome_length()
    return factor


class TestEvaluation(unittest.TestCase):
    def test_difference_factor(self):
        g1 = Genotype.create_default()
        g2 = Genotype.create_default()
        g3 = Genotype.create_default()
        random_parameter_mutation(g3)

        factor1 = genotype_difference_factor(g1, g2)
        self.assertEqual(factor1, 0.0)

        factor2 = genotype_difference_factor(g2, g3)
        self.assertEqual(factor2 > 0.0, True)
