import json
import random
import unittest
from typing import List, Tuple, Union

import numpy as np

from experiments.e1.src.encoder import NpEncoder


class ConstantParameter:
    def __init__(self, value):
        self._value = value
        self._allowed_values = [value]

    def value(self):
        return self._value

    def set_value(self, value):
        if value not in self._allowed_values:
            raise ValueError("Value not allowed")
        self._value = value

    def get_allowed_values(self) -> list:
        return self._allowed_values

    # Serialize/deserialize
    def to_dict(self) -> dict:
        data = {
            "value": self._value,
            "allowed_values": self._allowed_values,
        }
        return data


class RangeParameter:
    def __init__(self, initial_value, allowed_range, value_count: int = 100):
        self._value = initial_value
        self._allowed_range = list(allowed_range)

        if isinstance(initial_value, int):
            self._allowed_values = list(
                np.arange(allowed_range[0], allowed_range[1] + 1)
            )
            self._allowed_values = [int(v) for v in self._allowed_values]
        elif isinstance(initial_value, float):
            self._allowed_values = list(
                np.round(
                    np.linspace(allowed_range[0], allowed_range[1], value_count), 3
                )
            )
            self._allowed_values = [float(v) for v in self._allowed_values]
        self._value_count = len(self._allowed_values)

    def value(self):
        return self._value

    def set_value(self, value):
        if value not in self._allowed_values:
            raise ValueError("Value not allowed")
        self._value = value

    def allowed_values(self) -> list:
        return self._allowed_values

    # Serialize
    def to_dict(self) -> dict:
        data = {
            "value": self._value,
            "allowed_range": self._allowed_range,
            "value_count": self._value_count,
        }
        return data

    @staticmethod
    def from_dict(data: dict):
        value = data["value"]
        allowed_range = data["allowed_range"]
        value_count = data["value_count"]
        return RangeParameter(value, allowed_range, value_count)


class SetParameter:
    def __init__(self, initial_value, allowed_values):
        self._value = initial_value
        self._allowed_values = allowed_values

    def value(self):
        return self._value

    def set_value(self, value):
        if value not in self._allowed_values:
            raise ValueError("Value not allowed")
        self._value = value

    def allowed_values(self) -> list:
        return self._allowed_values

    # Serialize
    def to_dict(self) -> dict:
        data = {
            "value": self._value,
            "allowed_values": self._allowed_values,
        }
        return data

    @staticmethod
    def from_dict(data: dict):
        value = data["value"]
        allowed_values = data["allowed_values"]
        return SetParameter(value, allowed_values)


class Genotype:
    def __init__(self, chromosome):
        self._chromosome = chromosome
        self._train_loss = None
        self._test_loss = None

    @staticmethod
    def create_default():
        # (embedding dim, patch_size, head count, number of layers, forward expansion, dropout)
        embedding_dim = SetParameter(8, [8, 16, 32, 64, 128])
        patch_size = SetParameter(1, [1, 5, 10, 25, 50])
        head_count = SetParameter(2, [1, 2, 4, 8])
        layer_count = RangeParameter(1, (1, 16))
        forward_expansion = RangeParameter(1, (1, 8))
        dropout = RangeParameter(0.1, (0.0, 0.5), 21)

        chromosome = [
            embedding_dim,
            patch_size,
            head_count,
            layer_count,
            forward_expansion,
            dropout,
        ]
        return Genotype(chromosome)

    @staticmethod
    def create_constant(
        embedding_dim, patch_size, head_count, layer_count, forward_expansion, dropout
    ):
        # (embedding dim, patch_size, head count, number of layers, forward expansion, dropout)
        embedding_dim = ConstantParameter(embedding_dim)
        patch_size = ConstantParameter(patch_size)
        head_count = ConstantParameter(head_count)
        layer_count = ConstantParameter(layer_count)
        forward_expansion = ConstantParameter(forward_expansion)
        dropout = ConstantParameter(dropout)

        chromosome = [
            embedding_dim,
            patch_size,
            head_count,
            layer_count,
            forward_expansion,
            dropout,
        ]
        return Genotype(chromosome)

    # Serialize
    def to_dict(self) -> dict:
        data = {
            "embedding_dim": self._chromosome[0].to_dict(),
            "patch_size": self._chromosome[1].to_dict(),
            "head_count": self._chromosome[2].to_dict(),
            "layer_count": self._chromosome[3].to_dict(),
            "forward_expansion": self._chromosome[4].to_dict(),
            "dropout": self._chromosome[5].to_dict(),
            "train_loss": self.loss()[0],
            "test_loss": self.loss()[1],
            "fitness": self.fitness(),
        }
        return data

    @staticmethod
    def from_dict(data: dict):
        embedding_dim_param = SetParameter(
            data["embedding_dim"]["value"], data["embedding_dim"]["allowed_values"]
        )
        patch_size_param = SetParameter(
            data["patch_size"]["value"], data["patch_size"]["allowed_values"]
        )
        head_count_param = SetParameter(
            data["head_count"]["value"], data["head_count"]["allowed_values"]
        )
        layer_count_param = RangeParameter(
            data["layer_count"]["value"],
            data["layer_count"]["allowed_range"],
            data["layer_count"]["value_count"],
        )
        forward_expansion_param = RangeParameter(
            data["forward_expansion"]["value"],
            data["forward_expansion"]["allowed_range"],
            data["forward_expansion"]["value_count"],
        )
        dropout_param = RangeParameter(
            data["dropout"]["value"],
            data["dropout"]["allowed_range"],
            data["dropout"]["value_count"],
        )

        # (embedding dim, number of layers, head count, forward expansion, dropout)
        chromosome = [
            embedding_dim_param,
            patch_size_param,
            head_count_param,
            layer_count_param,
            forward_expansion_param,
            dropout_param,
        ]
        genotype = Genotype(chromosome)
        genotype._train_loss = data["train_loss"]
        genotype._test_loss = data["test_loss"]
        genotype._fitness = data["fitness"]

        return Genotype(chromosome)

    # Chromosome
    def chromosome_length(self):
        return len(self._chromosome)

    def chromosome(self, idx=None):
        if idx is not None:
            return self._chromosome[idx]

        return self._chromosome

    def values(self):
        return [p.get_value() for p in self._chromosome]

    # Fitness
    def fitness(self) -> Union[float, None]:
        if self._test_loss is None:
            return None

        return self._test_loss * (-1)

    def loss(self) -> Tuple[Union[float, None], Union[float, None]]:
        return self._train_loss, self._test_loss

    def set_loss(self, train_loss: float, test_loss: float):
        self._train_loss = train_loss
        self._test_loss = test_loss

    def is_dirty(self) -> bool:
        return self._train_loss is None or self._test_loss is None

    def mark_dirty(self):
        self._train_loss = None
        self._test_loss = None


def randomize_genotype(genotype: Genotype):
    for p in genotype.chromosome():
        values = p.allowed_values()
        chosen = random.choice(values)
        p.set_value(chosen)


def initialize_population(count: int) -> List[Genotype]:
    population = []
    for _ in range(count):
        genotype = Genotype.create_default()
        randomize_genotype(genotype)
        population.append(genotype)
    return population


def serialize_population(population: List[Genotype]) -> str:
    genotype_data_list = []
    for genotype in population:
        genotype_dict = genotype.to_dict()
        genotype_data_list.append(genotype_dict)
    return json.dumps(genotype_data_list, cls=NpEncoder)


def deserialize_population(str: str) -> List[Genotype]:
    genotype_data_list = json.loads(str)
    population = []
    for data in genotype_data_list:
        genotype = Genotype.from_dict(data)
        population.append(genotype)
    return population


class TestGenotype(unittest.TestCase):
    def test_initialization(self):
        count = 10
        population = initialize_population(count)
        self.assertEqual(len(population), count)

    def test_to_dict(self):
        genotype = Genotype.create_default()
        jsn = genotype.to_dict()
        self.assertNotEqual(jsn, {})

    def test_serialize_deserialize(self):
        population_before = [Genotype.create_default()]
        json_str = serialize_population(population_before)
        population_after = deserialize_population(json_str)

        dict_before = population_before[0].to_dict()
        dict_after = population_after[0].to_dict()
        self.assertEqual(dict_before, dict_after)
