import datetime
import json
import os
from enum import Enum

import torch
from torch import nn

from experiments.e1.src.encoder import NpEncoder
from experiments.e1.src.evaluation import (
    EvaluationConfig,
    evaluate_transformer,
    TransformerType,
)
from experiments.e1.src.visualization import make_prediction, plot_prediction
from src.dataset.LibraDataset import LibraCategory
from src.evolutionary.core.conversion import genotype_to_phenotype
from src.evolutionary.dataset.bootstrap import (
    boostrap_synthetic_dataset,
    boostrap_libra_dataset,
)
from src.evolutionary.genotype.Genotype import Genotype
from src.training.evaluation import rmse, mae
from src.utils.helpers import mean


class DatasetType(str, Enum):
    SYNTHETIC = "synthetic"
    LIBRA = "libra"


def train(
    batch_size,
    input_dim,
    output_dim,
    train_set,
    test_set,
    epochs,
    device,
    genotype_path,
):
    # Training
    criterion = nn.SmoothL1Loss(reduction="sum")
    optimizer = torch.optim.AdamW(
        [torch.zeros(1, 1)], lr=0.001, betas=(0.9, 0.98), eps=1e-9
    )
    evaluation_config = EvaluationConfig(
        batch_size,
        input_dim,
        output_dim,
        epochs,
        train_set,
        test_set,
        criterion,
        optimizer,
        device,
        0,
    )

    # Genotype and model
    genotype = Genotype.load_from_json(genotype_path)
    model = genotype_to_phenotype(genotype, input_dim, output_dim, device)

    transformer_type = TransformerType.ENCODER_ONLY
    train_loss_per_epoch, test_loss_per_epoch = evaluate_transformer(
        model, transformer_type, evaluation_config
    )

    return model, train_loss_per_epoch, test_loss_per_epoch


def main():
    # Datasets
    batch_size = 16
    input_dim = 200
    output_dim = 50

    dataset_type = DatasetType.LIBRA
    train_test_ratio = 0.8
    dataset_length = 5000
    libra_categories = [LibraCategory.HUMAN]
    max_included_datasets = 20
    max_values_per_dataset = 3000

    device = "cuda" if torch.cuda.is_available() else "cpu"

    if dataset_type == DatasetType.SYNTHETIC:
        train_set, test_set = boostrap_synthetic_dataset(
            input_dim,
            output_dim,
            dataset_length,
            train_test_ratio,
        )
    elif dataset_type == DatasetType.LIBRA:
        train_set, test_set = boostrap_libra_dataset(
            input_dim,
            output_dim,
            libra_categories,
            None,
            max_included_datasets,
            max_values_per_dataset,
            train_test_ratio,
        )
    else:
        raise ValueError("Unknown dataset type")

    # Training
    epochs = 5
    # TODO: Load the best genotype from json
    genotype_path = "randm-all-best-genotype33.json"
    model, train_loss, test_loss = train(
        batch_size,
        input_dim,
        output_dim,
        train_set,
        test_set,
        epochs,
        device,
        genotype_path,
    )

    # Calculate RMSE and MAE error
    print(f"$$$ Running evaluation $$$ ")
    mae_list = []
    rmse_list = []
    for idx, example in enumerate(test_set[:1000]):
        print(f"$ {idx + 1}/{len(test_set)}")
        source = example["x"]
        target = example["y"]

        patch_size = model.patch_size()
        prediction = make_prediction(
            model,
            TransformerType.ENCODER_ONLY,
            source,
            device,
            input_dim,
            output_dim,
            patch_size,
            0,
        )
        mae_list.append(mae(prediction, target))
        rmse_list.append(rmse(prediction, target))

    print(f"$$$ MAE: {mean(mae_list)} $$$ ")
    print(f"$$$ RMSE: {mean(rmse_list)} $$$ ")

    # Metadata to store
    metadata = {
        "dataset_type": dataset_type.value,
        "libra_categories": libra_categories,
        "train_test_ratio": train_test_ratio,
        "max_included_datasets": max_included_datasets,
        "input_dim": input_dim,
        "output_dim": output_dim,
        "batch_size": batch_size,
        "epochs": epochs,
        "mean_train_loss_per_epoch": train_loss,
        "mean_test_loss_per_epoch": test_loss,
        "test_rmse_mean": mean(rmse_list),
        "test_mae_mean": mean(mae_list),
    }
    save_results_as_json(metadata)

    # Plot example
    # TODO: Select examples to show
    for i in range(10):
        example = test_set[i]
        source = example["x"]
        target = example["y"]

        patch_size = model.patch_size()
        prediction = make_prediction(
            model,
            TransformerType.ENCODER_ONLY,
            source,
            device,
            input_dim,
            output_dim,
            patch_size,
            0,
        )
        plot_prediction(source, target, prediction)


def save_results_as_json(
    data,
    directory: str = "./results",
):
    timestamp = datetime.datetime.now().strftime("%d-%m-%Y_%H:%M:%S")
    if not os.path.exists(directory):
        os.mkdir(directory)

    filename = os.path.join(directory, f"{timestamp}.json")
    with open(filename, "w") as outfile:
        json.dump(data, outfile, cls=NpEncoder, indent=4)


if __name__ == "__main__":
    main()
