import os
import matplotlib.pyplot as plt
import torch
from torch import optim
from torch.nn import functional as F
import numpy as np
from NBEATS.model import NBeatsNet

# Adapted from: https://github.com/Y9008/NBEATS
# Uses library: https://github.com/philipperemy/n-beats


class NeuralBeats:
    def __init__(
        self,
        forecast_length,
        backcast_length,
        save_model=False,
        path="",
        checkpoint_name="NBEATS-checkpoint.th",
        device="cpu",
        batch_size=None,
        thetas_dims=[8, 8],  # [4, 8]
        nb_blocks_per_stack=6,
        share_weights_in_stack=False,
        train_percent=0.8,
        hidden_layer_units=128,
        stack=None,
        Architecture="DRESS",
    ):

        self.forecast_length = forecast_length
        if backcast_length == None:
            self.backcast_length = 3 * self.forecast_length
        else:
            self.backcast_length = backcast_length
        if batch_size == None:
            self.batch_size = int(self.data.shape[0] / 10)
        else:
            self.batch_size = batch_size
        self.hidden_layer_units = hidden_layer_units

        if stack == None:
            self.stacks = [NBeatsNet.GENERIC_BLOCK, NBeatsNet.GENERIC_BLOCK]
        else:
            Dict = {
                1: NBeatsNet.GENERIC_BLOCK,
                2: NBeatsNet.TREND_BLOCK,
                3: NBeatsNet.SEASONALITY_BLOCK,
            }
            self.stacks = [Dict.get(item, item) for item in stack]

        self.CHECKPOINT_NAME = path + checkpoint_name
        self.device = torch.device(
            device
        )  # change to gpu if gpu present for better performance
        self.thetas_dims = thetas_dims
        self.nb_blocks_per_stack = nb_blocks_per_stack
        self.train_size = train_percent
        self.share_weights_in_stack = share_weights_in_stack
        self.Architecture = Architecture

        self.net = NBeatsNet(
            stack_types=self.stacks,
            forecast_length=self.forecast_length,
            thetas_dims=self.thetas_dims,
            nb_blocks_per_stack=self.nb_blocks_per_stack,
            Architecture=self.Architecture,
            backcast_length=self.backcast_length,
            hidden_layer_units=self.hidden_layer_units,
            share_weights_in_stack=self.share_weights_in_stack,
            device=self.device,
        )
        self.parameters = self.net.parameters()
        self.global_step_cl = 0
        self.check_save = save_model
        self.loaded = False
        self.saved = True

    def plot_scatter(self, *args, **kwargs):
        plt.plot(*args, **kwargs)
        plt.scatter(*args, **kwargs)

    def data_generator(self, x_full, y_full, bs):
        def split(arr, size):
            arrays = []
            while len(arr) > size:
                slice_ = arr[:size]
                arrays.append(slice_)
                arr = arr[size:]
            arrays.append(arr)
            return arrays

        while True:
            for rr in split((x_full, y_full), bs):
                yield rr

    def loader(self, model, optimiser, use_checkpoint=False):
        if os.path.exists(self.CHECKPOINT_NAME) and use_checkpoint:
            checkpoint = torch.load(self.CHECKPOINT_NAME)

            model.load_state_dict(checkpoint["model_state_dict"])
            optimiser.load_state_dict(checkpoint["optimizer_state_dict"])
            grad_step = checkpoint["grad_step"]
            if self.loaded:
                self.norm_constant = checkpoint["norm_constant"]
            return grad_step
        return 0

    def saver(self, model, optimiser, grad_step):
        if self.saved:
            torch.save(
                {
                    "grad_step": grad_step,
                    "model_state_dict": model.state_dict(),
                    "optimizer_state_dict": optimiser.state_dict(),
                },
                self.CHECKPOINT_NAME,
            )
        else:
            torch.save(
                {
                    "grad_step": grad_step,
                    "model_state_dict": model.state_dict(),
                    "optimizer_state_dict": optimiser.state_dict(),
                },
                self.CHECKPOINT_NAME,
            )

    def load(self, file=None, optimiser=None):
        if file == None:
            raise Exception("Empty File Name")
        elif file == "":
            raise Exception("Empty File Name")
        else:
            if optimiser == None:
                self.optimiser = optim.Adam(self.net.parameters())
            else:
                self.optimiser = optimiser
            self.CHECKPOINT_NAME = file
            self.loaded = True
            self.global_step_cl = self.loader(self.net, self.optimiser)
            self.CHECKPOINT_NAME = "nbeats-training-checkpoint.th"

    def save(self, file=None):
        if file == None:
            raise Exception("Empty File Name")
        elif file == "":
            raise Exception("Empty File Name")
        else:
            self.CHECKPOINT_NAME = file
            self.saver(self.net, self.optimiser, self.global_step_cl)
            self.saved = True

    def train_100_grad_steps(self, data, train_loss_fn, test_losses):
        if not self.loaded:
            global_step = self.loader(self.net, self.optimiser)
            self.loaded = False
            self.global_step_cl = global_step
        for x_train_batch, y_train_batch in data:
            self.global_step_cl += 1
            self.optimiser.zero_grad()
            self.net.train()
            _, forecast = self.net(
                torch.tensor(np.array(x_train_batch), dtype=torch.float).to(self.device)
            )
            loss = train_loss_fn(
                forecast,
                torch.tensor(np.array(y_train_batch), dtype=torch.float).to(
                    self.device
                ),
            )
            loss.backward()
            self.optimiser.step()
            if self.verbose:
                if self.global_step_cl % 30 == 0:
                    print(
                        f"grad_step = {str(self.global_step_cl).zfill(6)}, tr_loss = {loss.item():.6f}, te_loss = {test_losses[-1]:.6f}"
                    )
            if self.global_step_cl > 0 and self.global_step_cl % 100 == 0:
                with torch.no_grad():
                    self.saver(self.net, self.optimiser, self.global_step_cl)
                break
        return forecast

    def eval_test(self, test_losses, x_test, y_test):
        self.net.eval()
        _, forecast = self.net(
            torch.tensor(np.array(x_test), dtype=torch.float, device=self.device)
        )
        test_losses.append(
            F.mse_loss(
                forecast,
                torch.tensor(np.array(y_test), dtype=torch.float, device=self.device),
            ).item()
        )
        p = forecast.detach().cpu().numpy()
        if self.plot:
            subplots = [221, 222, 223, 224]
            plt.figure(1)
            for plot_id, i in enumerate(
                np.random.choice(range(len(x_test)), size=4, replace=False)
            ):
                ff, xx, yy = p[i], x_test[i], y_test[i]
                plt.subplot(subplots[plot_id])
                plt.grid()
                self.plot_scatter(range(0, self.backcast_length), xx, color="b")
                self.plot_scatter(
                    range(
                        self.backcast_length,
                        self.backcast_length + self.forecast_length,
                    ),
                    yy,
                    color="g",
                )
                self.plot_scatter(
                    range(
                        self.backcast_length,
                        self.backcast_length + self.forecast_length,
                    ),
                    ff,
                    color="r",
                )
            plt.show()

    def fit(
        self,
        train_set,
        test_set,
        batch_size,
        epochs=20,
        optimizer=None,
        loss_fn=F.mse_loss,
        plot=True,
        verbose=True,
    ):
        self.plot = plot
        self.verbose = verbose
        self.epoch = epochs

        if optimizer == None:
            self.optimiser = optim.Adam(self.net.parameters())
        else:
            self.optimiser = optimizer

        x_train = [train_set[i]["x"] for i in range(len(train_set))]
        y_train = [train_set[i]["y"] for i in range(len(train_set))]
        x_test = [test_set[i]["x"] for i in range(len(test_set))]
        y_test = [test_set[i]["y"] for i in range(len(test_set))]

        train_data = self.data_generator(x_train, y_train, batch_size)
        test_losses = []
        for i in range(self.epoch):
            self.eval_test(test_losses, x_test, y_test)
            self.train_100_grad_steps(train_data, loss_fn, test_losses)
        if self.check_save:
            pass
        else:
            if os.path.exists(self.CHECKPOINT_NAME):
                os.remove(self.CHECKPOINT_NAME)

    def predict(self, predict_data):
        if predict_data.shape[0] != self.backcast_length:
            raise Exception(
                "Numpy array for prediction input should be of backcast_length: {} x 1 shape".format(
                    self.backcast_length
                )
            )
        else:
            predict_data = np.reshape(predict_data, (self.backcast_length, 1))
            _, forecasted_values = self.net(
                torch.tensor(predict_data.T, dtype=torch.float)
            )
            forecasted_values = forecasted_values.detach().cpu().numpy()

        return forecasted_values.T.flatten()

    def net_parameters(self):
        return self.net.parameters()
