import numpy as np
import torch
from matplotlib import pyplot as plt

# Adapted from: https://towardsdatascience.com/building-rnn-lstm-and-gru-for-time-series-using-pytorch-a46e5b094e7b


class Optimization:
    def __init__(self, model, loss_fn, optimizer, device):
        self.model = model
        self.loss_fn = loss_fn
        self.optimizer = optimizer
        self.device = device
        self.train_losses = []
        self.test_losses = []

    def train_step(self, x, y):
        # Sets model to train mode
        self.model.train()

        # Makes predictions
        yhat = self.model(x)

        # Computes loss
        loss = self.loss_fn(y, yhat)

        # Computes gradients
        loss.backward()

        # Updates parameters and zeroes gradients
        self.optimizer.step()
        self.optimizer.zero_grad()

        # Returns the loss
        return loss.item()

    def evaluate(self, test_set, batch_size=1, input_dim=1):
        with torch.no_grad():
            predictions = []
            targets = []
            for x_test, y_test in test_set:
                x_test = x_test.view([batch_size, -1, input_dim]).to(self.device)
                y_test = y_test.to(self.device)
                self.model.eval()
                yhat = self.model(x_test)
                predictions.append(yhat.to(self.device).detach().cpu().numpy())
                targets.append(y_test.to(self.device).detach().cpu().numpy())

        return predictions, targets

    def train(self, train_set, test_set, batch_size=16, epochs=10, input_dim=1):
        for epoch in range(1, epochs + 1):
            batch_losses = []
            for x_batch, y_batch in train_set:
                x_batch = x_batch.view([batch_size, -1, input_dim]).to(self.device)
                y_batch = y_batch.to(self.device)
                loss = self.train_step(x_batch, y_batch)
                batch_losses.append(loss)
            training_loss = np.mean(batch_losses)
            self.train_losses.append(training_loss)

            with torch.no_grad():
                batch_val_losses = []
                for x_val, y_val in test_set:
                    x_val = x_val.view([batch_size, -1, input_dim]).to(self.device)
                    y_val = y_val.to(self.device)
                    self.model.eval()
                    yhat = self.model(x_val)
                    val_loss = self.loss_fn(y_val, yhat).item()
                    batch_val_losses.append(val_loss)
                validation_loss = np.mean(batch_val_losses)
                self.test_losses.append(validation_loss)

            print(
                f"[{epoch}/{epochs}] Training loss: {training_loss:.4f}\t Validation loss: {validation_loss:.4f}"
            )

    def plot_losses(self):
        plt.plot(self.train_losses, label="Training loss")
        plt.plot(self.test_losses, label="Validation loss")
        plt.legend()
        plt.title("Losses")
        plt.show()
        plt.close()
