import random
from enum import Enum

import torch
from torch import optim, nn
from torch.utils.data import DataLoader

from experiments.e4.utils.batch import batch_iterator, create_prediction_batch
from experiments.e4.rnn.Optimization import Optimization
from experiments.e4.rnn.lstm import LSTM
from experiments.e4.rnn.rnn import RNN
from experiments.e4.utils.visualisation import plot_prediction
from src.dataset.SyntheticDataset import SyntheticDataset, SeriesType
from src.utils.seed import temp_seed

# https://towardsdatascience.com/building-rnn-lstm-and-gru-for-time-series-using-pytorch-a46e5b094e7b


class RecurrentType(str, Enum):
    RNN = "rnn"
    LSTM = "lstm"


def train(
    recurrent_type,
    batch_size,
    input_dim,
    output_dim,
    train_set,
    test_set,
    epochs,
    device,
    dropout,
):
    hidden_dim = 64
    layer_dim = 2

    model_params = {
        "input_dim": input_dim,
        "output_dim": output_dim,
        "hidden_dim": hidden_dim,
        "layer_dim": layer_dim,
        "dropout": dropout,
    }

    if recurrent_type == RecurrentType.RNN:
        model = RNN(**model_params).to(device=device)
    elif recurrent_type == RecurrentType.LSTM:
        model = LSTM(**model_params).to(device=device)
    else:
        raise ValueError("Unknown recurrent type")

    loss_fn = nn.SmoothL1Loss(reduction="sum")
    optimizer = optim.Adam(model.parameters(), lr=1e-3, weight_decay=1e-6)

    train_batches = list(
        batch_iterator(train_set, batch_size=batch_size, device=device)
    )
    test_batches = list(batch_iterator(test_set, batch_size=batch_size, device=device))

    opt = Optimization(model=model, loss_fn=loss_fn, optimizer=optimizer, device=device)
    opt.train(
        train_batches,
        test_batches,
        batch_size=batch_size,
        epochs=epochs,
        input_dim=input_dim,
    )
    return opt


def make_prediction(opt, source, target, input_dim, device):
    example_batch = create_prediction_batch(source, target, device)
    predictions, values = opt.evaluate(example_batch, batch_size=1, input_dim=input_dim)
    return predictions


def main():
    input_dim = 200
    output_dim = 50
    batch_size = 16
    dropout = 0.1
    epochs = 10

    device = "cuda" if torch.cuda.is_available() else "cpu"

    dataset = SyntheticDataset(
        SeriesType.COMPOSED_WITH_NOISE, x_length=input_dim, y_length=output_dim
    )
    with temp_seed(1):
        train_set, test_set = torch.utils.data.random_split(dataset, [4000, 750])

    # Shuffle datasets!
    with temp_seed(1):
        train_set = [train_set[i] for i in range(len(train_set))]
        random.shuffle(train_set)
        test_set = [test_set[i] for i in range(len(test_set))]
        random.shuffle(test_set)

    # TODO: Choose model
    model = train(
        RecurrentType.RNN,
        batch_size,
        input_dim,
        output_dim,
        train_set,
        test_set,
        epochs,
        device,
        dropout,
    )

    example = dataset[0]
    source = example["x"]
    target = example["y"]
    predictions = make_prediction(model, source, target, input_dim, device)
    plot_prediction(example["x"], example["y"], predictions[0][0])


if __name__ == "__main__":
    main()
