import math

import torch
import torch.nn as nn
import torch.nn.functional as functional

# Adaptive Attention Span
# Adapted from: https://github.com/facebookresearch/adaptive-span/blob/main/adaptive_span.py
# Paper: Adaptive Attention Span in Transformers


class AdaptiveMask(nn.Module):
    """Soft masking function for adaptive size.
    It masks out the last K values of an input. The masking value
    goes from 1 to 0 gradually, so K can be learned with
    back-propagation.

    Args:
        max_size: maximum size (i.e. input dimension)
        ramp_size: size of the ramp going from 0 to 1
        init_val: initial size proportion not to be masked out
        shape: learn multiple sizes independent of each other
    """

    def __init__(self, max_size, ramp_size, init_val=0.0, shape=(1,)):
        nn.Module.__init__(self)
        self._max_size = max_size
        self._ramp_size = ramp_size
        self.current_val = nn.Parameter(torch.zeros(*shape) + init_val)
        mask_template = torch.linspace(1 - max_size, 0, steps=max_size)
        self.register_buffer("mask_template", mask_template)

    def forward(self, x):
        mask = self.mask_template + self.current_val * self._max_size
        mask = mask / self._ramp_size + 1
        mask = mask.clamp(0, 1)
        if x.size(-1) < self._max_size:
            # The input could have been trimmed beforehand to save computation
            mask = mask[:, :, -x.size(-1) :]
        x = x * mask
        return x

    def get_current_max_size(self, include_ramp=True):
        current_size = math.ceil(self.current_val.max().item() * self._max_size)
        if include_ramp:
            current_size += self._ramp_size
        current_size = max(0, min(self._max_size, current_size))
        return current_size

    def get_current_avg_size(self, include_ramp=True):
        current_size = math.ceil(self.current_val.mean().item() * self._max_size)
        if include_ramp:
            current_size += self._ramp_size
        current_size = max(0, min(self._max_size, current_size))
        return current_size

    def clamp_param(self):
        """this need to be called after each update"""
        self.current_val.data.clamp_(0, 1)


class AdaptiveSpanAttention(nn.Module):
    def __init__(
        self,
        embedding_dim,
        head_count,
        input_dim,
        span_length_init=0.0,
        span_loss_coefficient=1.0,
        span_ramp=10,
    ):
        super(AdaptiveSpanAttention, self).__init__()
        self.attention = AdaptiveSpanAttention1(
            embedding_dim,
            head_count,
            input_dim,
            span_length_init,
            span_loss_coefficient,
            span_ramp,
        )

    def forward(self, x, mask=None):
        return self.attention(x, x, x, mask)


class AdaptiveSpanAttention1(nn.Module):
    """Adaptive attention span for Transformer
    This module learns an attention span length from data for each self-attention head

    Args:
        input_dim: maximum attention span i.e. input dimension
        span_loss_coefficient: loss coefficient for the span length
        span_ramp: length of the masking ramp (soft masking)
        span_length_init: initial size ratio
    """

    def __init__(
        self,
        embedding_dim,
        head_count,
        input_dim,
        span_length_init=0.0,
        span_loss_coefficient=1.0,
        span_ramp=10,
    ):
        super(AdaptiveSpanAttention1, self).__init__()
        assert (
            embedding_dim % head_count == 0
        ), "'embedding_dim' needs to be divisible by 'head_count'"
        self.d_k = embedding_dim // head_count
        self.head_count = head_count

        self.queries_linear = nn.Linear(embedding_dim, embedding_dim)
        self.keys_linear = nn.Linear(embedding_dim, embedding_dim)
        self.values_linear = nn.Linear(embedding_dim, embedding_dim)
        self.output_linear = nn.Linear(embedding_dim, embedding_dim)

        self.maximum_span = input_dim
        self.loss_coefficient = span_loss_coefficient
        self.adaptive_mask = AdaptiveMask(
            max_size=input_dim,
            ramp_size=span_ramp,
            init_val=span_length_init,
            shape=(head_count, 1, 1),
        )

    def forward(self, queries, keys, values, mask):
        # Linearly project input to obtain keys, values and queries
        batch_size = queries.size(0)
        queries = (
            self.queries_linear(queries)
            .view(batch_size, -1, self.head_count, self.d_k)
            .transpose(1, 2)
        )
        keys = (
            self.keys_linear(keys)
            .view(batch_size, -1, self.head_count, self.d_k)
            .transpose(1, 2)
        )
        values = (
            self.values_linear(values)
            .view(batch_size, -1, self.head_count, self.d_k)
            .transpose(1, 2)
        )

        queries_x_keys = torch.matmul(queries, keys.transpose(-2, -1))
        scores = queries_x_keys / math.sqrt(self.d_k)

        if mask is not None:
            mask = mask.unsqueeze(1)
            scores = scores.masked_fill(mask == 0, -1e9)

        attention = functional.softmax(scores, dim=-1)

        # Adaptive masking takes place here
        attention = self.mask_attention(attention)

        attention_x_values = torch.matmul(attention, values)

        # Concat the score matrices together (batch_size, -1, head_count * d_k)
        output = (
            attention_x_values.transpose(1, 2)
            .contiguous()
            .view(batch_size, -1, self.head_count * self.d_k)
        )
        return self.output_linear(output)

    def mask_attention(self, attention, normalize=True):
        """Mask attention with the right span"""
        # Batch and head dimensions are merged together, so separate them first
        batch_size = attention.size(0)
        head_count = attention.size(1)
        sequence_length = attention.size(2)

        attention = self.adaptive_mask(attention)
        if normalize:
            attention = attention / (
                attention.sum(-1, keepdim=True) + 1e-8
            )  # normalize so sum is 1

        attention = attention.view(batch_size, head_count, sequence_length, -1)
        return attention

    def get_trim_len(self):
        """How much of memory can be trimmed to reduce computation"""
        max_span = self.maximum_span
        trim_len = min(
            max_span - 1, max_span - self.adaptive_mask.get_current_max_size()
        )
        # Too fine granularity might be bad for the memory management
        trim_len = math.floor(trim_len / 64) * 64
        return trim_len

    def trim_memory(self, query, key, value, key_pe):
        """Trim out unnecessary memory beforehand to reduce computation"""
        trim_len = self.get_trim_len()
        cache_size = key.size(1) - query.size(1)
        trim_len_cache = trim_len - (self.maximum_span - cache_size)
        if trim_len_cache > 0:
            key = key[:, trim_len_cache:, :]
            value = value[:, trim_len_cache:, :]
        elif trim_len_cache < 0:
            # Cache is too short! this happens when validation resumes after a lot of updates
            key = functional.pad(key, [0, 0, -trim_len_cache, 0])
            value = functional.pad(value, [0, 0, -trim_len_cache, 0])
        if trim_len > 0:
            if key_pe is not None:
                key_pe = key_pe[:, :, trim_len:]
        return key, value, key_pe

    def get_loss(self):
        """A loss term for regularizing the span length"""
        return (
            self.loss_coefficient
            * self.maximum_span
            * self.adaptive_mask.current_val.mean()
        )

    def get_current_max_span(self):
        return self.adaptive_mask.get_current_max_size()

    def get_current_avg_span(self):
        return self.adaptive_mask.get_current_avg_size()

    def clamp_param(self):
        self.adaptive_mask.clamp_param()
