from typing import Optional

import torch
from torch import nn

# Attention free attention based on AFT Local
# Adapted from: https://nn.labml.ai/transformers/aft/index.html
# Paper: An Attention Free Transformer


class AttentionFreeLocalAttention(nn.Module):
    def __init__(
        self,
        embedding_dim: int,
        input_dim: int,
        window_size=None,  # 32
        window_size_divider: int = 4,
        max_sequence_length: int = 1024,
    ):
        super(AttentionFreeLocalAttention, self).__init__()
        self.attention = AttentionFreeLocalAttention1(
            embedding_dim,
            input_dim,
            window_size,
            window_size_divider,
            max_sequence_length,
        )

    def forward(self, x, mask=None):
        return self.attention(x, x, x, mask)


class AttentionFreeLocalAttention1(nn.Module):
    """Implement the "local" attention proposed An Attention Free Transformer

    Args:
        window_size: defines the size of local attention window
        max_sequence_length: defines the maximum acceptable sequence
                             length in order to allocate the learnable
                             parameters
    """

    def __init__(
        self,
        embedding_dim: int,
        input_dim: int,
        window_size=None,  # 32
        window_size_divider: int = 4,
        max_sequence_length: int = 1024,
    ):
        super(AttentionFreeLocalAttention1, self).__init__()
        self.embedding_dim = embedding_dim
        self.max_sequence_length = max_sequence_length

        if window_size is None:
            self.local_window_size = input_dim // window_size_divider
        else:
            self.local_window_size = window_size

        self.queries = nn.Linear(embedding_dim, embedding_dim)
        self.keys = nn.Linear(embedding_dim, embedding_dim)
        self.values = nn.Linear(embedding_dim, embedding_dim)
        self.output = nn.Linear(embedding_dim, embedding_dim)

        self.pos_bias = nn.Parameter(
            torch.zeros(max_sequence_length, max_sequence_length), requires_grad=True
        )
        self.local_mask = nn.Parameter(
            self.create_local_mask(max_sequence_length, self.local_window_size),
            requires_grad=False,
        )
        self.sigmoid = nn.Sigmoid()

    @staticmethod
    def create_local_mask(sequence_length, local_window_size):
        local_mask = torch.ones(sequence_length, sequence_length)
        local_mask = torch.tril(local_mask, local_window_size - 1)
        local_mask = torch.triu(local_mask, -(local_window_size - 1))
        return local_mask

    def forward(
        self,
        queries: torch.Tensor,
        keys: torch.Tensor,
        values: torch.Tensor,
        mask: Optional[torch.Tensor] = None,
    ):
        queries = self.queries(queries)
        keys = self.keys(keys)
        values = self.values(values)

        sequence_length = queries.size(1)
        positional_bias = (
            self.pos_bias[:sequence_length, :sequence_length]
            * self.local_mask[:sequence_length, :sequence_length]
        )
        positional_bias = positional_bias.unsqueeze(-1)

        # NOTE: This one uses the mask of the first examples in the batch, should be fine though
        if mask is not None:
            mask_swapped = mask.swapaxes(0, -1).swapaxes(0, 1)
            single_mask = mask_swapped[:, :, :1]
            positional_bias.masked_fill_(single_mask == 0, -1e9)

        max_key = keys.max(dim=1, keepdims=True)[0]
        max_pos_bias = positional_bias.max(dim=1, keepdims=True)[0]
        exp_key = torch.exp(keys - max_key)
        exp_pos_bias = torch.exp(positional_bias - max_pos_bias)

        exp_key_x_values = exp_key * values
        num = torch.einsum("ijb,bjd->bid", exp_pos_bias, exp_key_x_values)
        den = torch.einsum("ijb,bjd->bid", exp_pos_bias, exp_key)
        y = self.sigmoid(queries) * num / den
        return self.output(y)


# Test local mask
if __name__ == "__main__":
    sequence_length = 10
    local_window_size = 4
    print(
        AttentionFreeLocalAttention.create_local_mask(
            sequence_length, local_window_size
        )
    )
