import torch
from torch import nn
from torch.nn import LayerNorm

# Convolution block used in the Conformer Transformer
# Adapted from: https://github.com/pytorch/fairseq/blob/main/fairseq/modules/conformer_layer.py
# Paper: Conformer: Local Features Coupling Global Representations for Visual Recognition


class ConformerConvolutionAttention(nn.Module):
    def __init__(
        self,
        embedding_dim: int,
        channels: int = 32,
        depthwise_kernel_size: int = 31,
        dropout: float = 0.0,
        bias: bool = False,
    ):
        super(ConformerConvolutionAttention, self).__init__()
        self.attention = ConformerConvolutionAttention1(
            embedding_dim, channels, depthwise_kernel_size, dropout, bias
        )

    def forward(self, x, mask=None):
        return self.attention(x, x, x, mask)


class ConformerConvolutionAttention1(torch.nn.Module):
    """Convolution based attention from Conformer Transformer
    Args:
        channels: number of channels in depthwise conv layers
        depthwise_kernel_size: depthwise conv layer kernel size
        dropout: dropout probability at the module output
        bias: if bias should be added to conv layers
    """

    def __init__(
        self,
        embedding_dim: int,
        channels: int = 32,
        depthwise_kernel_size: int = 31,
        dropout: float = 0.0,
        bias: bool = False,
    ):
        super(ConformerConvolutionAttention1, self).__init__()
        assert (
            depthwise_kernel_size - 1
        ) % 2 == 0, "kernel_size should be a odd number for 'SAME' padding"

        self.layer_norm = LayerNorm(embedding_dim)
        self.pointwise_conv1 = torch.nn.Conv1d(
            embedding_dim,
            2 * channels,
            kernel_size=1,
            stride=1,
            padding=0,
            bias=bias,
        )
        self.glu = nn.GLU(dim=1)
        self.depthwise_conv = nn.Conv1d(
            channels,
            channels,
            depthwise_kernel_size,
            stride=1,
            padding=(depthwise_kernel_size - 1) // 2,
            groups=channels,
            bias=bias,
        )
        self.batch_norm = nn.BatchNorm1d(channels)
        self.activation = nn.SiLU()
        self.pointwise_conv2 = nn.Conv1d(
            channels,
            embedding_dim,
            kernel_size=1,
            stride=1,
            padding=0,
            bias=bias,
        )
        self.dropout = nn.Dropout(dropout)

    # NOTE: Masking is not supported
    def forward(self, x, _key, _value, _mask):
        x = self.layer_norm(x)
        # Exchange the temporal dimension and the feature dimension
        x = x.transpose(1, 2)

        # GLU mechanism
        x = self.pointwise_conv1(x)  # (batch, 2*channel, dim)
        x = self.glu(x)  # (batch, channel, dim)

        # 1D Depthwise Conv
        x = self.depthwise_conv(x)
        x = self.batch_norm(x)
        x = self.activation(x)

        x = self.pointwise_conv2(x)
        x = self.dropout(x)
        return x.transpose(1, 2)
