import torch
from torch import nn
from torch.nn import functional as functional

# Factorized Random Synthesizer
# Adapted from: https://github.com/10-zin/Synthesizer/blob/master/synth/synthesizer/modules.py
# Paper: Synthesizer: Rethinking Self-Attention for Transformer Models


class FactorizedRandomAttentionSynthesizer(nn.Module):
    def __init__(
        self,
        embedding_dim,
        head_count,
        dropout=0.0,
        batch_size=16,
        max_sequence_length=1024,
        factor=1,
    ):
        super(FactorizedRandomAttentionSynthesizer, self).__init__()
        self.attention = FactorizedRandomAttentionSynthesizer1(
            embedding_dim, head_count, dropout, batch_size, max_sequence_length, factor
        )

    def forward(self, x, mask=None):
        return self.attention(x, x, x, mask)


class FactorizedRandomAttentionSynthesizer1(nn.Module):
    """Factorized random attention
    Args:
        factor: factorization factor
        max_sequence_length: defines the maximum acceptable sequence length
        dropout: dropout probability at the module output
    """

    def __init__(
        self,
        embedding_dim,
        head_count,
        dropout=0.0,
        batch_size=16,
        max_sequence_length=1024,
        factor=1,
    ):
        super(FactorizedRandomAttentionSynthesizer1, self).__init__()
        self.embedding_dim = embedding_dim
        self.head_count = head_count
        self.max_sequence_length = max_sequence_length
        self.factor = factor

        self.random_attn_1 = torch.randn(
            batch_size, max_sequence_length, factor, requires_grad=True
        )
        self.random_attn_2 = torch.randn(
            batch_size, factor, max_sequence_length, requires_grad=True
        )
        self.dropout = nn.Dropout(dropout)

    def forward(self, q, k, v, mask=None):
        batch_size = q.size(0)
        sequence_length = q.size(1)
        random_attention = torch.matmul(
            self.random_attn_1.to(device=q.device),
            self.random_attn_2.to(device=q.device),
        )[:batch_size, :sequence_length, :sequence_length]

        if mask is not None:
            random_attention = random_attention.masked_fill(mask == 0, -1e9)

        random_attention = self.dropout(functional.softmax(random_attention, dim=-1))
        output = torch.matmul(random_attention, v)

        return output
