import torch
from torch import nn

# Fast weight attention (ironically, it's very slow to calculate)
# Adapted from: https://nn.labml.ai/transformers/fast_weights/index.html
# Paper: Linear Transformers Are Secretly Fast Weight Memory Systems


class DPFP(nn.Module):
    """Deterministic Parameter Free Project"""

    def __init__(self, nu: int = 1, eps: float = 1e-6):
        """
        * `nu` is a hyperparameter
        * `eps` is the small value used to make sure there is no division-by-zero when normalizing
        """
        super().__init__()
        self.nu = nu
        self.relu = nn.ReLU()
        self.eps = eps

    def forward(self, k: torch.Tensor):
        # Expected shape: (seq_len, batch_size, heads, d_k)
        k = self.dpfp(k)
        return k / (torch.sum(k, dim=-1, keepdim=True) + self.eps)

    def dpfp(self, k: torch.Tensor):
        x = self.relu(torch.cat([k, -k], dim=-1))
        x_rolled = [x.roll(shifts=i, dims=-1) for i in range(1, self.nu + 1)]
        x_rolled = torch.cat(x_rolled, dim=-1)
        x_repeat = torch.cat([x] * self.nu, dim=-1)
        return x_repeat * x_rolled


class FastWeightsAttention(nn.Module):
    def __init__(self, embedding_dim: int, head_count: int = 1):
        super(FastWeightsAttention, self).__init__()
        self.attention = FastWeightsAttention1(embedding_dim, head_count)

    def forward(self, x, mask=None):
        return self.attention(x, x, x, mask)


class FastWeightsAttention1(nn.Module):
    """Fast Weights Attention"""

    def __init__(self, embedding_dim: int, head_count: int = 1):
        super(FastWeightsAttention1, self).__init__()
        self.embedding_dim = embedding_dim
        self.head_count = head_count

        assert (
            embedding_dim % head_count == 0
        ), "'embedding_dim' needs to be divisible by 'head_count'"
        self.d_k = embedding_dim // head_count

        self.queries_linear = nn.Linear(embedding_dim, embedding_dim)
        self.keys_linear = nn.Linear(embedding_dim, embedding_dim)
        self.values_linear = nn.Linear(embedding_dim, embedding_dim)
        self.output_linear = nn.Linear(embedding_dim, embedding_dim)

        # Interpolation weight function
        # Shape after the linear: (head_shape, self.heads, self.d_k)
        self.interpolation_weight = nn.Sequential(
            nn.Linear(embedding_dim, head_count, bias=False), nn.Sigmoid()
        )

        self.phi = DPFP()

    # Note: Masking is not supported!
    def forward(self, queries, keys, values, _mask=None):
        batch_size = queries.size(0)
        sequence_length = queries.size(1)
        raw_queries = queries

        # Linearly project input to obtain keys, values and queries
        # Query, key and value result dim: (batch_size, heads, seq_len, d_k)
        queries = (
            self.queries_linear(queries)
            .view(batch_size, -1, self.head_count, self.d_k)
            .transpose(1, 2)
        )
        keys = (
            self.keys_linear(keys)
            .view(batch_size, -1, self.head_count, self.d_k)
            .transpose(1, 2)
        )
        values = (
            self.values_linear(values)
            .view(batch_size, -1, self.head_count, self.d_k)
            .transpose(1, 2)
        )

        queries = self.phi(queries)
        keys = self.phi(keys)

        betas = self.interpolation_weight(raw_queries)
        weights = keys.new_zeros(
            (keys.size(0), keys.size(1), values.size(3), keys.size(3))
        )

        # Iterate through steps
        outputs = []
        for i in range(sequence_length):
            key = keys[:, :, i, :]
            value = values[:, :, i, :]
            query = queries[:, :, i, :]
            beta = betas[:, i, :].unsqueeze(1)

            value_existing = torch.einsum("bhvk,bhk->bhv", weights, key)
            beta_x_value = beta * (value - value_existing)
            weights = weights + torch.einsum("bhv,bhk->bhvk", beta_x_value, key)
            y = torch.einsum("bhvk,bhk->bhv", weights, query)

            # Merge multiple heads and append to `outputs`
            reshaped = y.reshape(y.size(0), -1).unsqueeze(1)
            outputs.append(reshaped)

        x = torch.stack(outputs, dim=1).squeeze(2)
        return self.output_linear(x)
