import torch
from torch import nn

# Linear attention
# Adapted from: https://github.com/idiap/fast-transformers/blob/master/fast_transformers/attention/linear_attention.py
# Paper: Transformers are RNNs: Fast Autoregressive Transformers with Linear Attention
from src.components.activation.Activation import Activation


class LinearAttentionTransformersRNNs(nn.Module):
    def __init__(
        self,
        embedding_dim,
        head_count,
        activation: str = Activation.ELU.value,
        eps=1e-6,
    ):
        super(LinearAttentionTransformersRNNs, self).__init__()
        self.attention = LinearAttentionTransformersRNNs1(
            embedding_dim, head_count, activation, eps
        )

    def forward(self, x, mask=None):
        return self.attention(x, x, x, mask)


class LinearAttentionTransformersRNNs1(nn.Module):
    """Implement unmasked attention using dot product of feature maps in
    O(N D^2) complexity.
    Given the queries, keys and values as Q, K, V instead of computing
        V' = softmax(Q.mm(K.t()), dim=-1).mm(V),
    we make use of a feature map function Φ(.) and perform the following
    computation
        V' = normalize(Φ(Q).mm(Φ(K).t())).mm(V).
    The above can be computed in O(N D^2) complexity where D is the
    dimensionality of Q, K and V and N is the sequence length. Depending on the
    feature map, however, the complexity of the attention might be limited.
    Arguments
    ---------
        activation: activation function used to preprocess k,q,v matrices
        eps: float, a small number to ensure the numerical stability of the
             denominator (default: 1e-6)
    """

    def __init__(
        self,
        embedding_dim,
        head_count,
        activation: str = Activation.ELU.value,
        eps=1e-6,
    ):
        super(LinearAttentionTransformersRNNs1, self).__init__()
        self.embedding_size = embedding_dim
        self.head_count = head_count
        self.activation = Activation.from_str(activation).module()
        self.eps = eps

        # Dimension of queries, keys and values
        # We assume d_v always equals d_k
        assert (
            embedding_dim % head_count == 0
        ), "'embedding_dim' needs to be divisible by 'head_count'"
        self.d_k = embedding_dim // head_count

        self.queries = nn.Linear(embedding_dim, embedding_dim)
        self.keys = nn.Linear(embedding_dim, embedding_dim)
        self.values = nn.Linear(embedding_dim, embedding_dim)
        self.output = nn.Linear(embedding_dim, embedding_dim)

    # NOTE: Masking is not supported
    def forward(self, queries, keys, values, attn_mask):
        batch_size = queries.size(0)
        queries = (
            self.queries(queries)
            .view(batch_size, -1, self.head_count, self.d_k)
            .transpose(1, 2)
        )
        keys = (
            self.keys(keys)
            .view(batch_size, -1, self.head_count, self.d_k)
            .transpose(1, 2)
        )
        values = (
            self.values(values)
            .view(batch_size, -1, self.head_count, self.d_k)
            .transpose(1, 2)
        )

        # Apply the feature map to the queries and keys
        Q = self.activation(queries) + 1
        K = self.activation(keys) + 1

        # Compute the KV matrix, namely the dot product of keys and values so
        # that we never explicitly compute the attention matrix and thus
        # decrease the complexity
        KV = torch.einsum("nshd,nshm->nhmd", K, values)

        # Compute the normalizer
        Z = 1 / (torch.einsum("nlhd,nhd->nlh", Q, K.sum(dim=1)) + self.eps)

        # Finally compute and return the new values
        V = torch.einsum("nlhd,nhmd,nlh->nlhm", Q, KV, Z)

        output = (
            V.contiguous()
            .transpose(1, 2)
            .contiguous()
            .view(batch_size, -1, self.head_count * self.d_k)
        )
        return self.output(output)
