import math

import torch
from torch import nn
from torch.nn import functional as functional

from src.components.positional_encoding.ape.SinusoidalAbsolutePositionalEncoding import (
    SinusoidalAbsolutePositionalEncoding,
)

# Attention with TUPE positional encoding
# Adapted from: https://github.com/jaketae/tupe
# Paper: Rethinking Positional Encoding in Language Pre-training


class TupeMultiHeadAttention(nn.Module):
    def __init__(
        self,
        embedding_dim: int,
        head_count: int,
        dropout: float = 0.1,
        relative_bias: bool = True,
        bidirectional_bias: bool = True,
        bucket_count: int = 32,
        max_distance: int = 128,
        max_length: int = 1024,
        positional_embedding=None,
    ):
        super(TupeMultiHeadAttention, self).__init__()
        self.attention = TupeMultiHeadAttention1(
            embedding_dim,
            head_count,
            dropout,
            relative_bias,
            bidirectional_bias,
            bucket_count,
            max_distance,
            max_length,
            positional_embedding,
        )

    def forward(self, x, mask=None):
        return self.attention(x, x, x, mask)


class TupeMultiHeadAttention1(nn.Module):
    """Tupe multi head attention
    Args:
        bucket_count: number of relative positions buckets
        max_distance: max distance between two elements to be considered
        dropout: dropout probability at the module output
    """

    def __init__(
        self,
        embedding_dim: int,
        head_count: int,
        dropout: float = 0.1,
        relative_bias: bool = True,
        bidirectional_bias: bool = True,
        bucket_count: int = 32,
        max_distance: int = 128,
        max_length: int = 1024,
        positional_embedding=None,
    ) -> None:
        super().__init__()
        self.max_len = max_length
        self.num_heads = head_count
        self.num_buckets = bucket_count
        self.max_distance = max_distance
        self.bidirectional = bidirectional_bias
        self.scale = math.sqrt(2 * head_count)

        self.pos_embed = (
            SinusoidalAbsolutePositionalEncoding(embedding_dim, dropout)
            if positional_embedding is None
            else positional_embedding
        )
        self.dropout = nn.Dropout(dropout)

        # kqv in one pass
        self.pos_kq = nn.Linear(embedding_dim, 2 * embedding_dim, bias=False)
        self.tok_kqv = nn.Linear(embedding_dim, 3 * embedding_dim, bias=False)

        self.relative_bias = relative_bias
        if relative_bias:
            self.bias = nn.Embedding(max_length * 2, head_count)

    def forward(
        self,
        x: torch.tensor,
        _key: torch.tensor,
        _value: torch.tensor,
        _mask: torch.tensor,
    ) -> torch.tensor:
        batch_size, seq_len, embedding_dim = x.shape

        pos_embed = self.pos_embed(torch.zeros(batch_size, seq_len, 1, device=x.device))
        # pos_embed.shape == (batch_size, seq_len, d_model)

        pos_key, pos_query = self.pos_kq(pos_embed).chunk(2, dim=-1)
        pos_key = pos_key.view(batch_size, seq_len, self.num_heads, -1).permute(
            0, 2, 3, 1
        )
        # pos_key.shape == (batch_size, num_heads, d_head, seq_len)
        pos_query = pos_query.view(batch_size, seq_len, self.num_heads, -1).transpose(
            1, 2
        )
        # pos_query.shape == (batch_size, num_heads, seq_len, d_head)
        pos_attn = torch.matmul(pos_query, pos_key)
        # pos_attn.shape == (batch_size, num_heads, seq_len, seq_len)

        tok_key, tok_query, tok_value = self.tok_kqv(x).chunk(3, dim=-1)
        tok_key = tok_key.view(batch_size, seq_len, self.num_heads, -1).permute(
            0, 2, 3, 1
        )
        # tok_key.shape == (batch_size, num_heads, d_head, seq_len)
        tok_query = tok_query.view(batch_size, seq_len, self.num_heads, -1).transpose(
            1, 2
        )
        tok_value = tok_value.view(batch_size, seq_len, self.num_heads, -1).transpose(
            1, 2
        )
        # tok_qv.shape == (batch_size, num_heads, seq_len, d_head)
        tok_attn = torch.matmul(tok_query, tok_key)
        # tok_attn.shape == (batch_size, num_heads, seq_len, seq_len)

        attn = (tok_attn + pos_attn) / self.scale
        if self.relative_bias:
            relative_positions = get_relative_positions(
                seq_len, self.bidirectional, self.num_buckets, self.max_distance
            ).to(attn.device)
            # relative_positions.shape == (seq_len, seq_len)
            bias = self.bias(relative_positions + self.max_len)
            # bias.shape == (seq_len, seq_len, num_heads)
            bias = bias.permute(2, 0, 1).unsqueeze(0)
            # bias.shape == (1, num_heads, seq_len, seq_len)
            attn = attn + bias

        attn = functional.softmax(attn, dim=-1)
        # attn.shape == (batch_size, num_heads, seq_len, seq_len)
        out = torch.matmul(attn, tok_value)
        # out.shape == (batch_size, num_heads, seq_len, d_head)
        out = out.transpose(1, 2).reshape(batch_size, seq_len, -1)
        # out.shape == (batch_size, seq_len, d_model)
        out = self.dropout(out)
        return out


def _get_relative_position_bucket(
    relative_position, bidirectional, num_buckets, max_distance
):
    """
    from https://github.com/huggingface/transformers/blob/master/src/transformers/models/t5/modeling_t5.py
    """
    relative_buckets = 0
    if bidirectional:
        num_buckets //= 2
        relative_buckets += (relative_position > 0).to(torch.long) * num_buckets
        relative_position = torch.abs(relative_position)
    else:
        relative_position = -torch.min(
            relative_position, torch.zeros_like(relative_position)
        )
    # now relative_position is in the range [0, inf)

    # half of the buckets are for exact increments in positions
    max_exact = num_buckets // 2
    is_small = relative_position < max_exact

    # The other half of the buckets are for logarithmically bigger bins in positions up to max_distance
    relative_postion_if_large = max_exact + (
        torch.log(relative_position.float() / max_exact)
        / math.log(max_distance / max_exact)
        * (num_buckets - max_exact)
    ).to(torch.long)
    relative_postion_if_large = torch.min(
        relative_postion_if_large,
        torch.full_like(relative_postion_if_large, num_buckets - 1),
    )

    relative_buckets += torch.where(
        is_small, relative_position, relative_postion_if_large
    )
    return relative_buckets


def get_relative_positions(
    seq_len, bidirectional=True, num_buckets=32, max_distance=128
):
    x = torch.arange(seq_len)[None, :]
    y = torch.arange(seq_len)[:, None]
    relative_positions = _get_relative_position_bucket(
        x - y, bidirectional, num_buckets, max_distance
    )
    return relative_positions
