from collections import namedtuple
import torch
from torch import nn

from src.components.attention.utils.EfficientAttentionMixin import (
    EfficientAttentionMixin,
)

LocalAttentionOutput = namedtuple(
    "LocalSelfAttentionOutput", ["hidden_states", "attention_probs"]
)

# Local attention
# Adapted from: https://github.com/huggingface/transformers/blob/main/src/transformers/models/reformer/modeling_reformer.py
# Paper: Reformer: The Efficient Transformer


class LocalAttentionReformer(nn.Module):
    def __init__(
        self,
        embedding_dim,
        head_count,
        dropout=0.0,
        chunk_length=50,
        max_position_embeddings=1024,
        is_decoder=True,
    ):
        super(LocalAttentionReformer, self).__init__()
        self.attention = LocalAttentionReformer1(
            embedding_dim,
            head_count,
            dropout,
            chunk_length,
            max_position_embeddings,
            is_decoder,
        )

    def forward(self, x, mask=None):
        return self.attention(x, x, x, mask)


class LocalAttentionReformer1(nn.Module, EfficientAttentionMixin):
    """Local attention Reformer
    Args:
        chunk_length: size of chunks of elements that can attend to each other
        dropout: dropout probability at the module output
    """

    def __init__(
        self,
        embedding_dim,
        head_count,
        dropout=0.0,
        chunk_length=50,
        max_position_embeddings=1024,
        is_decoder=True,
    ):
        super(LocalAttentionReformer1, self).__init__()
        self.chunk_length = chunk_length
        self.max_position_embeddings = max_position_embeddings
        self.is_decoder = is_decoder

        self.num_chunks_before = 1
        self.num_chunks_after = 0
        self.hash_seed = 1
        self.num_buckets = None

        self.dropout = dropout

        self.num_attention_heads = head_count
        self.attention_head_size = embedding_dim // head_count  # d_k
        self.all_head_size = self.num_attention_heads * self.attention_head_size
        self.hidden_size = embedding_dim

        # projection matrices
        self.query = nn.Linear(self.hidden_size, self.all_head_size, bias=False)
        self.key = nn.Linear(self.hidden_size, self.all_head_size, bias=False)
        self.value = nn.Linear(self.hidden_size, self.all_head_size, bias=False)

        # save mask value here
        self.register_buffer("mask_value_float16", torch.tensor(-1e4))
        self.register_buffer("mask_value_float32", torch.tensor(-1e9))

    def forward(
        self,
        hidden_states,
        key,
        value,
        attention_mask=None,
        head_mask=None,
        past_buckets_states=None,
        use_cache=False,
        output_attentions=False,
        **kwargs,
    ):
        attention_mask = attention_mask.squeeze(-1)
        sequence_length = hidden_states.shape[1]
        batch_size = hidden_states.shape[0]

        # check if cache shall be used and that hidden states are already cached
        if use_cache and past_buckets_states[1] is not None:
            assert (
                past_buckets_states[0] is None
            ), "LocalSelfAttention should not make use of `buckets`. There seems to be an error when caching hidden_states_and_buckets."
            key_value_hidden_states = self._retrieve_relevant_hidden_states(
                past_buckets_states[1], self.chunk_length, self.num_chunks_before
            )
            key_value_hidden_states = torch.cat(
                [key_value_hidden_states, hidden_states], dim=1
            )

            # only query vector for last token
            query_vectors = self.query(hidden_states)
            # compute key and value for relevant chunk
            key_vectors = self.key(key_value_hidden_states)
            value_vectors = self.value(key_value_hidden_states)

            # free memory
            del key_value_hidden_states
        else:
            # project hidden_states to query, key and value
            query_vectors = self.query(hidden_states)
            key_vectors = self.key(hidden_states)
            value_vectors = self.value(hidden_states)

        # split last dim into `config.num_attention_heads` and `config.attention_head_size`
        query_vectors = self._split_hidden_size_dim(
            query_vectors, self.num_attention_heads, self.attention_head_size
        )
        key_vectors = self._split_hidden_size_dim(
            key_vectors, self.num_attention_heads, self.attention_head_size
        )
        value_vectors = self._split_hidden_size_dim(
            value_vectors, self.num_attention_heads, self.attention_head_size
        )

        assert (
            query_vectors.shape[-1] == self.attention_head_size
        ), f"last dim of query_key_vectors is {query_vectors.shape[-1]} but should be {self.attention_head_size}."
        assert (
            key_vectors.shape[-1] == self.attention_head_size
        ), f"last dim of query_key_vectors is {key_vectors.shape[-1]} but should be {self.attention_head_size}."
        assert (
            value_vectors.shape[-1] == self.attention_head_size
        ), f"last dim of query_key_vectors is {value_vectors.shape[-1]} but should be {self.attention_head_size}."

        if self.chunk_length is None:
            assert (
                self.num_chunks_before == 0 and self.num_chunks_after == 0
            ), "If `config.chunk_length` is `None`, make sure `config.num_chunks_after` and `config.num_chunks_before` are set to 0."

        # normalize key vectors
        key_vectors = key_vectors / torch.sqrt(
            torch.tensor(
                self.attention_head_size,
                device=key_vectors.device,
                dtype=key_vectors.dtype,
            )
        )

        # get sequence length indices
        indices = torch.arange(sequence_length, device=query_vectors.device).repeat(
            batch_size, self.num_attention_heads, 1
        )

        # if one should do normal n^2 self-attention
        do_standard_self_attention = sequence_length <= self.chunk_length

        # if input should be chunked
        if not do_standard_self_attention:
            # chunk vectors
            # B x Num_Attn_Head x Seq_Len // chunk_len x chunk_len  x  attn_head_size
            query_vectors = self._split_seq_length_dim_to(
                query_vectors,
                -1,
                self.chunk_length,
                self.num_attention_heads,
                self.attention_head_size,
            )
            key_vectors = self._split_seq_length_dim_to(
                key_vectors,
                -1,
                self.chunk_length,
                self.num_attention_heads,
                self.attention_head_size,
            )
            value_vectors = self._split_seq_length_dim_to(
                value_vectors,
                -1,
                self.chunk_length,
                self.num_attention_heads,
                self.attention_head_size,
            )

            # chunk indices
            query_indices = self._split_seq_length_dim_to(
                indices, -1, self.chunk_length, self.num_attention_heads
            )
            key_indices = self._split_seq_length_dim_to(
                indices, -1, self.chunk_length, self.num_attention_heads
            )

            # append chunks before and after
            key_vectors = self._look_adjacent(
                key_vectors, self.num_chunks_before, self.num_chunks_after
            )
            value_vectors = self._look_adjacent(
                value_vectors, self.num_chunks_before, self.num_chunks_after
            )
            key_indices = self._look_adjacent(
                key_indices, self.num_chunks_before, self.num_chunks_after
            )
        else:
            query_indices = key_indices = indices

        # query-key matmul: QK^T
        query_key_dots = torch.matmul(query_vectors, key_vectors.transpose(-1, -2))

        # free memory
        del query_vectors, key_vectors

        mask = self._compute_attn_mask(
            query_indices,
            key_indices,
            attention_mask,
            query_key_dots.shape,
            do_standard_self_attention,
        )

        if mask is not None:
            # get mask tensor depending on half precision or not
            if query_key_dots.dtype == torch.float16:
                mask_value = self.mask_value_float16.half()
            else:
                mask_value = self.mask_value_float32

            mask = mask.to(dtype=torch.bool)
            query_key_dots = torch.where(mask, query_key_dots, mask_value)

        # free memory
        del mask

        # softmax
        logits = torch.logsumexp(query_key_dots, dim=-1, keepdim=True)
        attention_probs = torch.exp(query_key_dots - logits)

        # free memory
        del logits

        # dropout
        attention_probs = nn.functional.dropout(
            attention_probs, p=self.dropout, training=self.training
        )

        # Mask heads if we want to
        if head_mask is not None:
            attention_probs = attention_probs * head_mask

        # attend values
        out_vectors = torch.matmul(attention_probs, value_vectors)

        # free memory
        del value_vectors

        # merge chunk length
        if not do_standard_self_attention:
            out_vectors = out_vectors.flatten(start_dim=2, end_dim=3)

        assert out_vectors.shape == (
            batch_size,
            self.num_attention_heads,
            sequence_length,
            self.attention_head_size,
        )

        out_vectors = self._merge_hidden_size_dims(
            out_vectors, self.num_attention_heads, self.attention_head_size
        )

        if output_attentions is False:
            attention_probs = ()

        # NOTE: This is needed if I want to output attention probabilities or logits
        # return LocalAttentionOutput(hidden_states=out_vectors, attention_probs=attention_probs)
        return out_vectors

    def _compute_attn_mask(
        self,
        query_indices,
        key_indices,
        attention_mask,
        query_key_dots_shape,
        do_standard_self_attention,
    ):

        # chunk attention mask and look before and after
        if attention_mask is not None:
            attention_mask = attention_mask.to(torch.uint8)[:, None, :]

            if not do_standard_self_attention:
                attention_mask = self._split_seq_length_dim_to(
                    attention_mask, -1, self.chunk_length, 1
                )
                attention_mask = self._look_adjacent(
                    attention_mask, self.num_chunks_before, self.num_chunks_after
                )
            # create attn_mask
            attention_mask = attention_mask.unsqueeze(-2).expand(query_key_dots_shape)

        # Causal mask
        if self.is_decoder is True:
            causal_mask = torch.ge(
                query_indices.unsqueeze(-1), key_indices.unsqueeze(-2)
            ).to(query_indices.device)

            # add attention mask if not None
            if attention_mask is not None:
                attention_mask = causal_mask * attention_mask
            else:
                attention_mask = causal_mask

        return attention_mask

    @staticmethod
    def _retrieve_relevant_hidden_states(
        previous_hidden_states, chunk_length, num_chunks_before
    ):
        start_position = (
            (previous_hidden_states.shape[1] // chunk_length) - num_chunks_before
        ) * chunk_length
        return previous_hidden_states[:, start_position:]
