import unittest

import torch
from torch import nn


class AddCoordinates(object):
    r"""Coordinate Adder Module as defined in 'An Intriguing Failing of
    Convolutional Neural Networks and the CoordConv Solution'
    (https://arxiv.org/pdf/1807.03247.pdf).
    This module concatenates coordinate information (`x`, `y`, and `r`) with
    given input tensor.
    `x` and `y` coordinates are scaled to `[-1, 1]` range where origin is the
    center. `r` is the Euclidean distance from the center and is scaled to
    `[0, 1]`.
    Args:
        with_r (bool, optional): If `True`, adds radius (`r`) coordinate
            information to input image. Default: `False`
    Shape:
        - Input: `(N, C_{in}, H_{in}, W_{in})`
        - Output: `(N, (C_{in} + 2) or (C_{in} + 3), H_{in}, W_{in})`
    """

    def __init__(self, with_r=False):
        self.with_r = with_r

    def __call__(self, image):
        batch_size, _, image_height, image_width = image.size()

        y_coords = (
            2.0
            * torch.arange(image_height).unsqueeze(1).expand(image_height, image_width)
            / (image_height - 1.0)
            - 1.0
        )
        x_coords = (
            2.0
            * torch.arange(image_width).unsqueeze(0).expand(image_height, image_width)
            / (image_width - 1.0)
            - 1.0
        )

        coords = torch.stack((y_coords, x_coords), dim=0)

        if self.with_r:
            rs = ((y_coords**2) + (x_coords**2)) ** 0.5
            rs = rs / torch.max(rs)
            rs = torch.unsqueeze(rs, dim=0)
            coords = torch.cat((coords, rs), dim=0)

        coords = torch.unsqueeze(coords, dim=0).repeat(batch_size, 1, 1, 1)
        image = torch.cat((coords.to(image.device), image), dim=1)
        return image


class CoordConv(nn.Module):
    r"""2D Convolution Module Using Extra Coordinate Information as defined
    in 'An Intriguing Failing of Convolutional Neural Networks and the
    CoordConv Solution' (https://arxiv.org/pdf/1807.03247.pdf).
    Args:
        Same as `torch.nn.Conv2d` with two additional arguments
        with_r (bool, optional): If `True`, adds radius (`r`) coordinate
            information to input image. Default: `False`
    Shape:
        - Input: `(N, C_{in}, H_{in}, W_{in})`
        - Output: `(N, C_{out}, H_{out}, W_{out})`
    """

    def __init__(self, kernel_size: int = 3, dilation: int = 1, bias=True):
        super(CoordConv, self).__init__()

        # Adjust for even kernel sizes
        if (kernel_size % 2) == 0:
            kernel_size = kernel_size + 1

        self.conv_layer = nn.Conv2d(
            3,
            3,
            kernel_size,
            stride=1,
            padding=(kernel_size // 2) * dilation,
            dilation=dilation,
            groups=1,
            bias=bias,
        )
        self.coord_adder = AddCoordinates()

    def forward(self, x):
        x = x.unsqueeze(1)
        out1 = self.coord_adder(x)
        out2 = self.conv_layer(out1)
        # Sum the generated coordinates with the original input values
        out = torch.sum(out2, 1)
        return out


class TestCoordConv(unittest.TestCase):
    def test(self):
        conv_coord = CoordConv(3, 1)
        input = torch.randn(20, 100, 16)
        output = conv_coord(input)
        self.assertEqual(list(output.shape), [20, 100, 16])
