import torch
from torch import nn
from operator import mul
from functools import reduce

# Axial Positional Encoding
# Adapted from: https://github.com/lucidrains/axial-positional-embedding
# Paper: Reformer: The Efficient Transformer


# Axial shape (32, 32) results to max sequence length of 1024
class AxialPositionalEncoding(nn.Module):
    def __init__(self, embedding_dim, axial_shape=(32, 32), axial_embedding_dims=None):
        super().__init__()
        self.embedding_dim = embedding_dim
        self.shape = axial_shape
        self.max_sequence_length = reduce(mul, axial_shape, 1)

        self.summed = axial_embedding_dims is None
        axial_embedding_dims = (
            ((embedding_dim,) * len(axial_shape))
            if self.summed
            else axial_embedding_dims
        )

        assert len(self.shape) == len(
            axial_embedding_dims
        ), "number of axial dimensions must equal the number of dimensions in the shape"
        assert (
            self.summed
            or not self.summed
            and sum(axial_embedding_dims) == embedding_dim
        ), f"axial dimensions must sum up to the target dimension {embedding_dim}"

        self.weights = ParameterList(self, "weights", len(axial_shape))

        for ind, (shape, axial_dim) in enumerate(zip(self.shape, axial_embedding_dims)):
            ax_shape = [1] * len(self.shape)
            ax_shape[ind] = shape
            ax_shape = (1, *ax_shape, axial_dim)
            ax_emb = nn.Parameter(torch.zeros(ax_shape).normal_(0, 1))
            self.weights.append(ax_emb)

    def forward(self, x):
        batch_size, sequence_length, _ = x.shape
        assert (
            sequence_length <= self.max_sequence_length
        ), f"Sequence length ({sequence_length}) must be less than the maximum sequence length allowed ({self.max_sequence_length})"
        embeddings = []

        for axial_embedding in self.weights.to_list():
            axial_dim = axial_embedding.shape[-1]
            expand_shape = (batch_size, *self.shape, axial_dim)
            emb = axial_embedding.expand(expand_shape).reshape(
                batch_size, self.max_sequence_length, axial_dim
            )
            embeddings.append(emb)

        positional_embedding = (
            sum(embeddings) if self.summed else torch.cat(embeddings, dim=-1)
        )
        return x + positional_embedding[:, :sequence_length].to(x)


class ParameterList(object):
    def __init__(self, kls, prefix, length):
        self.ind = 0
        self.kls = kls
        self.prefix = prefix
        self.length = length

    def _keyname(self, prefix, ind):
        return f"{prefix}_{ind}"

    def append(self, x):
        setattr(self.kls, self._keyname(self.prefix, self.ind), x)
        self.ind += 1

    def to_list(self):
        return [
            getattr(self.kls, self._keyname(self.prefix, i)) for i in range(self.length)
        ]
