import math
import unittest

import numpy as np
import torch
import torch.nn as nn
from matplotlib import pyplot as plt
from torch.autograd import Variable


class SinusoidalAbsolutePositionalEncoding(nn.Module):
    def __init__(self, embedding_dim, dropout, max_sequence_length=1024):
        super(SinusoidalAbsolutePositionalEncoding, self).__init__()
        self.dropout = nn.Dropout(dropout)

        encoding = torch.zeros(max_sequence_length, embedding_dim)
        position = torch.arange(0, max_sequence_length).unsqueeze(1)
        division_term = torch.exp(
            torch.arange(0, embedding_dim, 2)
            * -(math.log(max_sequence_length * 2) / embedding_dim)
        )

        encoding[:, 0::2] = torch.sin(position * division_term)
        encoding[:, 1::2] = torch.cos(position * division_term)
        encoding = encoding.unsqueeze(0)

        self.register_buffer("encoding", encoding)

    def forward(self, x):
        x = x + Variable(self.encoding[:, : x.size(1)], requires_grad=False)
        return self.dropout(x)


class TestTimeSeriesEmbedding(unittest.TestCase):
    def test(self):
        data = np.array([[1], [2]], dtype=np.dtype("i4"))
        x = torch.tensor(data)

        embedding = SinusoidalAbsolutePositionalEncoding(4, 0)
        y = embedding.forward(x)

        print(y)


if __name__ == "__main__":
    batch_size = 1
    sequence_length = 100
    embedding_dim = 20

    positional_encoding = SinusoidalAbsolutePositionalEncoding(embedding_dim, 0)
    x = Variable(torch.zeros(batch_size, sequence_length, embedding_dim))
    y = positional_encoding.forward(x)

    plt.figure(figsize=(15, 5))
    plt.plot(np.arange(100), y[0, :, 4:8].data.numpy())
    plt.legend(["Dim %d" % p for p in [4, 5, 6, 7]])
    plt.show()
