import torch
from torch import nn

# Relative positional encoding used by DeBERTa
# Adapted from: https://github.com/huggingface/transformers/blob/57e6464ac9a31156f1c93e59107323e6ec01309e/src/transformers/models/deberta/modeling_deberta.py
# Paper: DeBERTa: Decoding-enhanced BERT with Disentangled Attention


class RelativePositionalEncodingDeBERTa(nn.Module):
    def __init__(self):
        super().__init__()

    def forward(self, x):
        # (batch_size, head_count, sequence_length, sequence_length)
        batch_size = x.size(0)
        head_count = x.size(1)
        sequence_length = x.size(2)
        positional_encoding = (
            self.build_relative_position(sequence_length, sequence_length, x.device)
            / sequence_length
        )
        positional_encoding_expanded = positional_encoding.expand(
            batch_size, head_count, sequence_length, sequence_length
        )
        x_with_encoding = x + positional_encoding_expanded
        return x_with_encoding

    def build_relative_position(self, query_size, key_size, device):
        """
        Build relative position according to the query and key
        We assume the absolute position of query \\(P_q\\) is range from (0, query_size) and the absolute position of key
        \\(P_k\\) is range from (0, key_size), The relative positions from query to key is \\(R_{q \\rightarrow k} = P_q -
        P_k\\)
        Args:
            query_size (int): the length of query
            key_size (int): the length of key
        Return:
            `torch.LongTensor`: A tensor with shape [1, query_size, key_size]
        """

        q_ids = torch.arange(query_size, dtype=torch.long, device=device)
        k_ids = torch.arange(key_size, dtype=torch.long, device=device)
        rel_pos_ids = q_ids[:, None] - k_ids.view(1, -1).repeat(query_size, 1)
        rel_pos_ids = rel_pos_ids[:query_size, :]
        rel_pos_ids = rel_pos_ids.unsqueeze(0)
        return rel_pos_ids
