import math

import torch
from torch import nn

# TUPE positional encoding
# Adapted from: https://github.com/jaketae/tupe
# Paper: Rethinking Positional Encoding in Language Pre-training


class RelativePositionalEncodingTupe(nn.Module):
    def __init__(
        self,
        head_count: int,
        bucket_count: int = 32,
        max_distance: int = 128,
        max_length: int = 1024,
    ):
        super().__init__()
        self.bucket_count = bucket_count
        self.max_distance = max_distance
        self.max_length = max_length
        self.head_count = head_count

        self.bias = nn.Embedding(max_length * 2, head_count)

    def forward(self, x):
        # (batch_size, head_count, sequence_length, sequence_length)
        sequence_length = x.size(2)
        relative_positions = get_relative_positions(
            sequence_length, True, self.bucket_count, self.max_distance
        ).to(x.device)
        # relative_positions.shape == (seq_len, seq_len)
        bias = self.bias(relative_positions + self.max_length)
        # bias.shape == (seq_len, seq_len, num_heads)
        bias = bias.permute(2, 0, 1).unsqueeze(0)
        # bias.shape == (1, num_heads, seq_len, seq_len)
        x = x + bias

        return x


def _get_relative_position_bucket(
    relative_position, bidirectional, num_buckets, max_distance
):
    """
    from https://github.com/huggingface/transformers/blob/master/src/transformers/models/t5/modeling_t5.py
    """
    relative_buckets = 0
    if bidirectional:
        num_buckets //= 2
        relative_buckets += (relative_position > 0).to(torch.long) * num_buckets
        relative_position = torch.abs(relative_position)
    else:
        relative_position = -torch.min(
            relative_position, torch.zeros_like(relative_position)
        )
    # now relative_position is in the range [0, inf)

    # half of the buckets are for exact increments in positions
    max_exact = num_buckets // 2
    is_small = relative_position < max_exact

    # The other half of the buckets are for logarithmically bigger bins in positions up to max_distance
    relative_postion_if_large = max_exact + (
        torch.log(relative_position.float() / max_exact)
        / math.log(max_distance / max_exact)
        * (num_buckets - max_exact)
    ).to(torch.long)
    relative_postion_if_large = torch.min(
        relative_postion_if_large,
        torch.full_like(relative_postion_if_large, num_buckets - 1),
    )

    relative_buckets += torch.where(
        is_small, relative_position, relative_postion_if_large
    )
    return relative_buckets


def get_relative_positions(
    seq_len, bidirectional=True, num_buckets=32, max_distance=128
):
    x = torch.arange(seq_len)[None, :]
    y = torch.arange(seq_len)[:, None]
    relative_positions = _get_relative_position_bucket(
        x - y, bidirectional, num_buckets, max_distance
    )
    return relative_positions
