import copy
from enum import Enum

import numpy as np
from matplotlib import pyplot as plt
from torch.utils.data import Dataset

from src.dataset.normalization.Normalizer import Normalizer
from src.dataset.utils.signal import sine_wave, square_wave, triangle_wave


class SignalType(Enum):
    SINE = ("sine",)
    SQUARE = ("square",)
    TRIANGLE = ("triangle",)


class SignalDataset(Dataset):
    def __init__(
        self,
        signal_type,
        length=1000,
        sampling_rate=20,
        frequency=0.05,
        amplitude=1,
        x_length=200,
        y_length=40,
        normalizer=Normalizer.ZERO_ONE,
        normalization_factor=1.0,
    ):
        self.signal_type = signal_type
        self.length = length
        self.sampling_rate = sampling_rate
        self.frequency = frequency
        self.amplitude = amplitude

        self.x_length = x_length
        self.y_length = y_length

        if self.signal_type == SignalType.SINE:
            self.series = sine_wave(
                length, sampling_rate, self.frequency, self.amplitude
            )[-1]
        elif self.signal_type == SignalType.SQUARE:
            self.series = square_wave(
                length, sampling_rate, self.frequency, self.amplitude
            )[-1]
        elif self.signal_type == SignalType.TRIANGLE:
            self.series = triangle_wave(
                length, sampling_rate, self.frequency, self.amplitude
            )[-1]
        else:
            raise NotImplementedError()

        # Normalize values
        if normalizer != Normalizer.NONE:
            norm = normalizer.module(self.series)
            self.series = norm.normalize(self.series) * normalization_factor

    def raw_series(self):
        return copy.deepcopy(self.series)

    def __len__(self):
        return (self.length * self.sampling_rate) - (self.x_length + self.y_length)

    def __getitem__(self, idx):
        x = self.series[idx : idx + self.x_length]
        y = self.series[idx + self.x_length : idx + self.x_length + self.y_length]
        return {"x": x, "y": y}

    def show(self, idx=None, max_values=200):
        if idx is not None:
            item = self[idx]
            x1 = np.arange(0, self.x_length)
            x2 = np.arange(self.x_length, self.x_length + self.y_length)
            plt.plot(x1, item["x"])
            plt.plot(x2, item["y"])
            plt.show()
        else:
            if max_values:
                series = self.series[:max_values]
            else:
                series = self.series
            x = np.arange(0, len(series))
            plt.plot(x, series)
            plt.show()


if __name__ == "__main__":
    dataset = SignalDataset(SignalType.SINE)
    print(f"Dataset length: {len(dataset)}")
    dataset.show(max_values=2000)
