import copy

from src.evolutionary.genotype.Genotype import Genotype
from src.evolutionary.genotype.GenotypeCell import GenotypeCell, GenotypeModuleBranch
from src.evolutionary.genotype.modules.ModuleRegistry import ModuleRegistry


def bootstrap_post_norm_transformer_encoder(
    genotype: Genotype, module_registry: ModuleRegistry
) -> Genotype:
    attention = module_registry.find_module(
        "src.components.attention.ScaledDotProductMultiHeadAttention"
    )
    if attention is None:
        raise ValueError("Could not find attention module during initialization")

    feed_forward = module_registry.find_module("src.components.transformer.FeedForward")
    if feed_forward is None:
        raise ValueError("Could not find feed forward module during initialization")

    layer_norm = module_registry.find_module(
        "src.components.normalization.LayerNormalization"
    )
    if layer_norm is None:
        raise ValueError("Could not find layer norm module during initialization")

    for idx in list(set(genotype.layer_indices())):
        cell1 = GenotypeCell()
        cell1.add_module(copy.deepcopy(attention), GenotypeModuleBranch.RIGHT)
        cell2 = GenotypeCell()
        cell2.add_module(copy.deepcopy(layer_norm), GenotypeModuleBranch.LEFT)
        cell2.add_module(copy.deepcopy(layer_norm), GenotypeModuleBranch.RIGHT)
        cell2.add_module(copy.deepcopy(feed_forward), GenotypeModuleBranch.RIGHT)
        cell3 = GenotypeCell()
        cell3.add_module(copy.deepcopy(layer_norm), GenotypeModuleBranch.LEFT)
        cell3.add_module(copy.deepcopy(layer_norm), GenotypeModuleBranch.RIGHT)

        layer = genotype.get_layer(idx)
        layer.add_cell(0, 0, cell1)
        layer.add_cell(1, 1, cell2)
        layer.add_cell(2, 2, cell3)
    return genotype


def bootstrap_pre_norm_transformer_encoder(
    genotype: Genotype, module_registry: ModuleRegistry
) -> Genotype:
    attention = module_registry.find_module(
        "src.components.attention.ScaledDotProductMultiHeadAttention"
    )
    if attention is None:
        raise ValueError("Could not find attention module during initialization")

    feed_forward = module_registry.find_module("src.components.transformer.FeedForward")
    if feed_forward is None:
        raise ValueError("Could not find feed forward module during initialization")

    layer_norm = module_registry.find_module(
        "src.components.normalization.LayerNormalization"
    )
    if layer_norm is None:
        raise ValueError("Could not find layer norm module during initialization")

    genotype.add_postprocess_module(copy.deepcopy(layer_norm))

    for idx in list(set(genotype.layer_indices())):
        cell1 = GenotypeCell()
        cell1.add_module(copy.deepcopy(layer_norm), GenotypeModuleBranch.RIGHT)
        cell1.add_module(copy.deepcopy(attention), GenotypeModuleBranch.RIGHT)
        cell2 = GenotypeCell()
        cell2.add_module(copy.deepcopy(layer_norm), GenotypeModuleBranch.RIGHT)
        cell2.add_module(copy.deepcopy(feed_forward), GenotypeModuleBranch.RIGHT)

        layer = genotype.get_layer(idx)
        layer.add_cell(0, 0, cell1)
        layer.add_cell(1, 1, cell2)
    return genotype


def bootstrap_evolved_transformer_cell(module_registry: ModuleRegistry) -> GenotypeCell:
    attention = module_registry.find_module(
        "src.components.attention.ScaledDotProductMultiHeadAttention"
    )
    if attention is None:
        raise ValueError("Could not find attention module during initialization")

    feed_forward = module_registry.find_module("src.components.transformer.FeedForward")
    if feed_forward is None:
        raise ValueError("Could not find feed forward module during initialization")

    layer_norm = module_registry.find_module(
        "src.components.normalization.LayerNormalization"
    )
    if layer_norm is None:
        raise ValueError("Could not find layer norm module during initialization")

    activation = module_registry.find_module("src.components.activation.ReLU")
    if layer_norm is None:
        raise ValueError("Could not find ReLU module during initialization")

    cell = GenotypeCell()
    cell.add_module(copy.deepcopy(layer_norm), GenotypeModuleBranch.RIGHT)
    cell.add_module(copy.deepcopy(attention), GenotypeModuleBranch.RIGHT)
    cell.add_module(copy.deepcopy(activation), GenotypeModuleBranch.RIGHT)
    cell.add_module(copy.deepcopy(layer_norm), GenotypeModuleBranch.LEFT)
    cell.add_module(copy.deepcopy(feed_forward), GenotypeModuleBranch.LEFT)
    cell.add_module(copy.deepcopy(activation), GenotypeModuleBranch.LEFT)
    return cell


# Paper: Efficient Evolutionary Search of Attention Convolutional Networks via Sampled Training and Node Inheritance
def bootstrap_one_module_per_branch_cell(
    module_registry: ModuleRegistry,
) -> GenotypeCell:
    module1 = module_registry.get_random_module()
    module2 = module_registry.get_random_module()

    cell = GenotypeCell()
    cell.add_module(copy.deepcopy(module1), GenotypeModuleBranch.LEFT)
    cell.add_module(copy.deepcopy(module2), GenotypeModuleBranch.RIGHT)
    return cell
