import unittest
from itertools import combinations
from typing import List

from src.evolutionary.genotype.modules.registries.vanilla_modules import (
    modules as vanilla_modules,
)
from src.evolutionary.core.population import (
    initialize_empty_population,
    randomize_population,
)
from src.evolutionary.genotype.Genotype import Genotype
from src.evolutionary.genotype.GenotypeCell import GenotypeCell, GenotypeModuleBranch
from src.evolutionary.genotype.GenotypeLayer import GenotypeLayer
from src.evolutionary.genotype.GenotypeModule import GenotypeModule
from src.evolutionary.genotype.parameters.IParameter import IParameter
from src.utils.helpers import mean


def argmax(iterable):
    return max(enumerate(iterable), key=lambda x: x[1])[0]


def argmin(iterable):
    return min(enumerate(iterable), key=lambda x: x[1])[0]


def parameter_difference_factor(
    parameter1: IParameter, parameter2: IParameter, eps=1e-5
) -> float:
    allowed_values1 = parameter1.get_allowed_values()
    allowed_values2 = parameter1.get_allowed_values()

    if parameter1.get_type() != parameter2.get_type():
        return 1.0

    # This guard is here because of float imprecision issues in RangeParameter
    if parameter1.get_type() == "RangeParameter":
        # If allowed values have difference greater than eps, return unequal
        allowed_values_differences = [
            abs(v1 - v2) for v1, v2 in zip(allowed_values1, allowed_values2)
        ]
        difference_sum = float(sum(allowed_values_differences))
        if difference_sum > eps:
            return 1.0

        # Find the allowed value closest to the current value
        param1_value = parameter1.get_value()
        param2_value = parameter2.get_value()

        differences_to_current1 = [abs(param1_value - v) for v in allowed_values1]
        parameter1_index = argmin(differences_to_current1)

        differences_to_current2 = [abs(param2_value - v) for v in allowed_values1]
        parameter2_index = argmin(differences_to_current2)
    else:
        if allowed_values1 != allowed_values2:
            return 1.0

        param1_value = parameter1.get_value()
        param2_value = parameter2.get_value()

        parameter1_index = allowed_values1.index(param1_value)
        parameter2_index = allowed_values2.index(param2_value)

    factor = abs(parameter1_index - parameter2_index) / (len(allowed_values1) - 1 + eps)
    return factor


def module_difference_factor(module1: GenotypeModule, module2: GenotypeModule) -> float:
    if module1.get_category() != module2.get_category():
        return 1.0

    # If the module is from the same category but not then same one, return 0.75
    if module1.get_module_path() != module2.get_module_path():
        return 0.75

    parameters1 = module1.get_parameters()
    parameters2 = module2.get_parameters()

    if len(parameters1) == 0 or len(parameters2) == 0:
        return 0.0

    parameters_diff_factors = []
    for idx in range(len(parameters1)):
        p1 = parameters1[idx]
        p2 = parameters2[idx]
        diff_factor = parameter_difference_factor(p1, p2)
        parameters_diff_factors.append(diff_factor)

    mean_parameters_diff_factor = mean(parameters_diff_factors) * 0.75
    return mean_parameters_diff_factor


def cell_difference_factor(cell1: GenotypeCell, cell2: GenotypeCell, eps=1e-5) -> float:
    # Left branch
    left_module_count1 = cell1.module_count(GenotypeModuleBranch.LEFT)
    left_module_count2 = cell2.module_count(GenotypeModuleBranch.LEFT)
    min_length_left = min(left_module_count1, left_module_count2)

    factors_left = []
    for idx in range(min_length_left):
        m1 = cell1.get_module(GenotypeModuleBranch.LEFT, idx)
        m2 = cell2.get_module(GenotypeModuleBranch.LEFT, idx)
        diff_factor = module_difference_factor(m1, m2)
        factors_left.append(diff_factor)
    mean_factor_left = mean(factors_left) if len(factors_left) > 0 else 0
    length_difference_factor = abs(left_module_count1 - left_module_count2) / (
        max(left_module_count1, left_module_count2) + eps
    )
    mean_factor_left = mean([mean_factor_left, length_difference_factor])

    # Right branch
    right_module_count1 = cell1.module_count(GenotypeModuleBranch.RIGHT)
    right_module_count2 = cell2.module_count(GenotypeModuleBranch.RIGHT)
    min_length_right = min(right_module_count1, right_module_count2)

    factors_right = []
    for idx in range(min_length_right):
        m1 = cell1.get_module(GenotypeModuleBranch.RIGHT, idx)
        m2 = cell2.get_module(GenotypeModuleBranch.RIGHT, idx)
        diff_factor = module_difference_factor(m1, m2)
        factors_right.append(diff_factor)
    mean_factor_right = mean(factors_right) if len(factors_right) > 0 else 0
    length_difference_factor = abs(right_module_count1 - right_module_count2) / (
        max(right_module_count1, right_module_count2) + eps
    )
    mean_factor_right = mean([mean_factor_right, length_difference_factor])

    if len(factors_left) == 0 and len(factors_right) == 0:
        return 0.0

    total_difference_factor = mean([mean_factor_left, mean_factor_right])
    return total_difference_factor


def layer_difference_factor(
    layer1: GenotypeLayer, layer2: GenotypeLayer, eps=1e-5
) -> float:
    layer1_cells = layer1.cells()
    layer2_cells = layer2.cells()
    min_length = min(len(layer1_cells), len(layer2_cells))

    factors = []
    for idx in range(min_length):
        c1 = layer1_cells[idx]
        c2 = layer2_cells[idx]
        left_index_factor = abs(c1[0] - c2[0]) / (
            max(len(layer1_cells), len(layer2_cells)) + eps
        )
        right_index_factor = abs(c1[1] - c2[1]) / (
            max(len(layer1_cells), len(layer2_cells)) + eps
        )
        index_diff_factor = mean([left_index_factor, right_index_factor])
        cell_diff_factor = cell_difference_factor(c1[2], c2[2])
        factors.append(mean([index_diff_factor, cell_diff_factor]))

    if len(factors) == 0:
        return 0.0

    total_difference_factor = mean(factors)
    return total_difference_factor


def genotype_difference_factor(genotype1: Genotype, genotype2: Genotype) -> float:
    # Genotype level params
    params1 = genotype1.parameters()
    params2 = genotype2.parameters()

    param_factors = []
    for idx in range(len(params1)):
        param1 = params1[idx]
        param2 = params2[idx]
        diff_factor = parameter_difference_factor(param1, param2)
        param_factors.append(diff_factor)

    # Layers
    if genotype1.has_explicit_layer_count() or genotype2.has_explicit_layer_count():
        min_length = 1
    else:
        layer_count1 = genotype1.layer_count()
        layer_count2 = genotype2.layer_count()
        min_length = min(layer_count1, layer_count2)

    if min_length == 0:
        return 0.0

    layer_factors = []
    for idx in range(min_length):
        l1 = genotype1.get_layer(idx)
        l2 = genotype2.get_layer(idx)
        diff_factor = layer_difference_factor(l1, l2)
        layer_factors.append(diff_factor)

    total_difference_factor = mean(layer_factors + param_factors)
    return total_difference_factor


def population_difference_factor(genotypes: List[Genotype]):
    pairs = list(combinations(genotypes, 2))
    factors = [genotype_difference_factor(g1, g2) for g1, g2 in pairs]
    factor_mean = mean(factors)
    return factor_mean


class TestGenotypeDifferenceFactor(unittest.TestCase):
    def test_empty_population(self):
        genotype1 = Genotype()
        genotype2 = Genotype()
        diff_factor = genotype_difference_factor(genotype1, genotype2)
        self.assertEqual(diff_factor, 0.0)

    def test_no_difference(self):
        genotype = Genotype()
        genotypes = initialize_empty_population(genotype, 2)
        diff_factor = genotype_difference_factor(genotypes[0], genotypes[1])
        self.assertEqual(diff_factor, 0.0)

    def test_some_difference(self):
        genotype = Genotype()
        genotypes = initialize_empty_population(genotype, 2)
        genotype1 = genotypes[0]
        genotype2 = genotypes[1]
        randomize_population([genotype2], vanilla_modules, False, 0, 10, 20, 50)
        diff_factor1 = genotype_difference_factor(genotype1, genotype2)
        self.assertEqual(diff_factor1 > 0.0, True)
