import copy
import time
from typing import List

from src.evolutionary.core.conversion import genotype_to_phenotype
from src.evolutionary.genotype.Genotype import Genotype
from src.evolutionary.phenotype.Phenotype import Phenotype
from src.training.NoamOptimizer import NoamOptimizer
from src.training.TransformerLossCompute import TransformerLossCompute
from src.transformers.encoder_only.train_encoder_only import batch_iterator, run_epoch
from src.utils.helpers import is_nan
from src.utils.logger import logger


class EvaluationSetup:
    batch_size: int
    input_dim: int
    output_dim: int
    max_epochs: int
    train_set: list
    test_set: list
    train_loss: any
    test_loss: any
    optimizer: any


log = logger()


def evaluate_phenotype(phenotype: Phenotype, config: EvaluationSetup, device) -> float:
    model = phenotype.to(device)

    # NOTE: Very important!!! Add model parameters to the optimizer
    config.optimizer.add_param_group({"params": [p for p in model.parameters()]})

    scheduled_optimizer = NoamOptimizer(
        config.optimizer,
        phenotype.embedding_dim(),
        1000,
    )
    scheduled_optimizer.reset()  # Reset the warmup steps

    # Train
    start = time.time()
    model.train()
    train_loss = run_epoch(
        batch_iterator(config.train_set, config.batch_size, config.input_dim, device),
        model,
        TransformerLossCompute(
            model.get_regression(), config.train_loss, scheduled_optimizer
        ),
        0,
        log,
    )
    log.info(f">> Train acc loss: {train_loss:.8f}")

    model.eval()
    eval_loss = run_epoch(
        batch_iterator(config.test_set, config.batch_size, config.input_dim, device),
        model,
        TransformerLossCompute(model.get_regression(), config.test_loss, None),
        0,
        log,
    )
    log.info(f">> Eval acc loss: {eval_loss:.8f}")
    log.info(f">>> Epoch took: {(time.time() - start):.2f}s")

    # We want the smallest loss i.e. multiply by -1 to get fitness
    loss = eval_loss.detach().cpu().numpy()

    # If the resulting loss is NaN, set it to inf to signify bad model
    if is_nan(loss):
        loss = float("inf")

    # Convert loss to fitness
    fitness = float(loss) * (-1)

    # NOTE: Very important!!! Resets optimizer and removes the current model parameters from it!
    config.optimizer.param_groups.clear()  # optim.param_groups = []
    config.optimizer.state.clear()  # optim.state = defaultdict(dict)
    config.optimizer.zero_grad()

    return fitness


def evaluate_phenotype_with_hurdles(
    phenotype: Phenotype, hurdles: list, config: EvaluationSetup, device
) -> float:
    hurdles = copy.deepcopy(hurdles)
    hurdles.append(float("inf"))  # To make sure that we will stop at some point

    idx = 0
    hurdle = hurdles[idx]

    log.info(f"$$$ Epoch: {idx + 1} $$$")
    fitness = evaluate_phenotype(phenotype, config, device)
    if config.max_epochs == 1:
        return fitness

    while fitness > hurdle:
        idx += 1
        log.info(f"$$$ Epoch: {idx + 1} $$$")

        fitness = evaluate_phenotype(phenotype, config, device)
        if idx == len(hurdles) - 1 or idx + 1 == config.max_epochs:
            break

        hurdle = hurdles[idx]

    return fitness


def evaluate_population(genotypes: List[Genotype], config: EvaluationSetup, device):
    for genotype in genotypes:
        if genotype.is_dirty():
            phenotype = genotype_to_phenotype(
                genotype, config.input_dim, config.output_dim, device
            )
            fitness = evaluate_phenotype(phenotype, config, device)
            genotype.set_fitness(fitness)
