import copy
import random
from enum import Enum
from typing import List

import numpy as np

from src.evolutionary.core.conversion import genotype_to_phenotype
from src.evolutionary.core.evaluation import (
    EvaluationSetup,
    evaluate_phenotype_with_hurdles,
)
from src.evolutionary.genotype.Genotype import Genotype


class SelectionMode(Enum):
    RANDOM = "random"
    BEST_FITNESS = "best_fitness"
    WORST_FITNESS = "worst_fitness"
    OLDEST = "oldest"

    @staticmethod
    def from_str(string: str):
        return SelectionMode(string)


def tournament_selection_with_hurdles(
    population: List[Genotype],
    tournament_size: int,
    selection_mode: SelectionMode,
    hurdles: list,
    config: EvaluationSetup,
    device,
) -> Genotype:
    tournament_genotypes: List[(int, Genotype)] = random.sample(
        list(enumerate(population)), tournament_size
    )

    # For the oldest genotype selection we don't need to evaluate fitness
    if selection_mode == SelectionMode.OLDEST:
        tournament_genotypes_timestamps = [
            g[1].timestamp() for g in tournament_genotypes
        ]
        individual_index = np.argmin(tournament_genotypes_timestamps)
    elif selection_mode == SelectionMode.RANDOM:
        individual_index, _ = random.choice(tournament_genotypes)
    else:
        fitnesses = []
        for genotype in tournament_genotypes:
            if genotype[1].is_dirty():
                phenotype = genotype_to_phenotype(
                    genotype[1], config.input_dim, config.output_dim, device
                )
                fitness = evaluate_phenotype_with_hurdles(
                    phenotype, hurdles, config, device
                )
                genotype[1].set_fitness(fitness)  # Cache the fitness
            else:
                fitness = genotype[1].fitness()
            fitnesses.append(fitness)

        if selection_mode == SelectionMode.BEST_FITNESS:
            individual_index = np.argmax(fitnesses)
        elif selection_mode == SelectionMode.WORST_FITNESS:
            individual_index = np.argmin(fitnesses)
        else:
            raise ValueError("Unknown genotype selection mode")

    selected_genotype = copy.deepcopy(
        tournament_genotypes[individual_index]
    )  # Important

    return selected_genotype
