import io
import random

import networkx as nx
import numpy as np
from PIL import Image
from matplotlib import pyplot as plt

from src.evolutionary.genotype.GenotypeCell import GenotypeModuleBranch
from src.evolutionary.genotype.Genotype import Genotype
from src.evolutionary.genotype.GenotypeLayer import GenotypeLayer


def visualize_genotype(genotype: Genotype):
    for idx, lr in enumerate(genotype.layers()):
        visualize_layer(lr, title=f"Layer {idx + 1}")


def visualize_layer(
    layer: GenotypeLayer, node_size=1200, dpi=160, title=None, show=True
):
    edges = []
    colors = []
    labels = {}
    positions = {}

    cells = layer.cells(without_first=True)

    current_y = 1
    positions["0"] = (1, current_y)
    labels["0"] = "Input"

    for idx, cell in enumerate(cells):
        idx += 1  # The first cell was skipped
        left_modules = cell[-1].modules(GenotypeModuleBranch.LEFT)
        right_modules = cell[-1].modules(GenotypeModuleBranch.RIGHT)

        cell_height = (
            len(left_modules)
            if len(left_modules) > len(right_modules)
            else len(right_modules)
        )

        previous_id = f"{cell[0]}"
        for idm, module in enumerate(left_modules):
            new_id = f"{idx}l{idm}"
            edges.append((previous_id, new_id))
            positions[new_id] = (
                0,
                current_y + 1 + (idm / len(left_modules) * cell_height),
            )
            labels[new_id] = module.get_module_label()
            previous_id = new_id
        edges.append((previous_id, f"{idx}"))

        previous_id = f"{cell[1]}"
        for idm, module in enumerate(right_modules):
            new_id = f"{idx}r{idm}"
            edges.append((previous_id, new_id))
            positions[new_id] = (
                2,
                current_y + 1 + (idm / len(right_modules) * cell_height),
            )
            labels[new_id] = module.get_module_label()
            previous_id = new_id
        edges.append((previous_id, f"{idx}"))

        current_y += cell_height + 1
        positions[str(idx)] = (1, current_y)
        labels[str(idx)] = "+"

    current_y += 1
    latest_id = int(edges[-1][-1])
    new_id = latest_id + 1
    edges.append((str(latest_id), str(new_id)))
    positions[str(new_id)] = (1, current_y)
    labels[str(new_id)] = "Output"

    graph = nx.DiGraph()
    graph.add_edges_from(edges)

    for node_key in graph.nodes.keys():
        colors.append(random_color(int(node_key[0])))

    plt.figure("Genotype", figsize=(5, current_y + 1), dpi=dpi)
    plt.xlim(-1, 3)
    plt.ylim(0, current_y + 1)
    plt.autoscale(False)

    if title is not None:
        plt.title(title, fontsize=16)

    nx.draw(
        graph,
        node_size=node_size,
        edgecolors="black",
        node_shape="s",
        node_color=colors,
        labels=labels,
        pos=positions,
        bbox=dict(
            facecolor=(0.9, 0.9, 0.9), edgecolor="black", boxstyle="round,pad=0.4"
        ),
    )

    buffer = io.BytesIO()
    plt.savefig(buffer, format="png")
    buffer.seek(0)
    image = Image.open(buffer)
    exported_image = np.array(image)[:, :, :3]

    if show:
        plt.show()

    return exported_image


def random_color(seed):
    rng = random.Random(seed)
    r = rng.uniform(0.6, 1.0)
    g = rng.uniform(0.6, 1.0)
    b = rng.uniform(0.6, 1.0)
    rgb = (r, g, b)
    return rgb
