from typing import List, Optional, Tuple
import jsonpickle

from src.evolutionary.genotype.GenotypeLayer import GenotypeLayer
from src.evolutionary.genotype.GenotypeModule import GenotypeModule
from src.evolutionary.genotype.parameters.IParameter import IParameter
from src.evolutionary.genotype.parameters.RangeParameter import RangeParameter
from src.evolutionary.genotype.parameters.SetParameter import SetParameter


class GenotypeConfig:
    initial_embedding_dim: int = 8
    allowed_embedding_dim: List[int] = [8, 16, 32, 64, 128]
    initial_patch_size: int = 1
    allowed_patch_size: List[int] = [1, 5, 10, 25, 50]
    initial_dropout: float = 0.1
    allowed_range_dropout: Tuple[float, float] = (0.0, 0.5)
    value_count_dropout: int = 11
    initial_use_memory: bool = False
    allowed_use_memory_states: List[bool] = [True, False]
    initial_use_full_regression: bool = True
    allowed_use_full_regression_states: List[bool] = [True, False]
    explicit_layer_count: bool = True
    initial_layer_count: int = 1
    allowed_range_layer_count: Tuple[int, int] = (1, 16)
    value_count_layer_count: int = 16


class Genotype:
    def __init__(self, config: GenotypeConfig = GenotypeConfig()):
        self._preprocess: List[GenotypeModule] = []
        self._postprocess: List[GenotypeModule] = []
        self._layers: List[GenotypeLayer] = []
        self._fitness: Optional[int] = None
        self._timestamp: int = 0

        # Genotype-level parameters
        self._parameters = []
        self._parameters.append(
            SetParameter(config.initial_embedding_dim, config.allowed_embedding_dim)
        )
        self._parameters.append(
            SetParameter(config.initial_patch_size, config.allowed_patch_size)
        )
        self._parameters.append(
            RangeParameter(
                config.initial_dropout,
                config.allowed_range_dropout,
                config.value_count_dropout,
            )
        )
        self._parameters.append(
            SetParameter(config.initial_use_memory, config.allowed_use_memory_states)
        )
        self._parameters.append(
            SetParameter(
                config.initial_use_full_regression,
                config.allowed_use_full_regression_states,
            )
        )

        # Optional layer count
        self._explicit_layer_count = config.explicit_layer_count
        if self._explicit_layer_count:
            self._parameters.append(
                RangeParameter(
                    config.initial_layer_count,
                    config.allowed_range_layer_count,
                    config.value_count_layer_count,
                )
            )
            self._layers.append(GenotypeLayer())

    # Preprocess modules
    def add_preprocess_module(self, module: GenotypeModule, index=None):
        idx = len(self._preprocess) if index is None else index
        self._preprocess.insert(idx, module)

    def delete_preprocess_module(self, index):
        if index in self.preprocess_module_indices():
            return self._preprocess.pop(index)
        else:
            raise IndexError("Module index is out of bounds")

    def preprocess_modules(self) -> List[GenotypeModule]:
        return self._preprocess

    def preprocess_module_count(self) -> int:
        return len(self._preprocess)

    def preprocess_module_indices(self) -> List[int]:
        return list(range(0, len(self._preprocess)))

    # Postprocess modules
    def add_postprocess_module(self, module: GenotypeModule, index=None):
        idx = len(self._postprocess) if index is None else index
        self._postprocess.insert(idx, module)

    def delete_postprocess_module(self, index):
        if index in self.postprocess_module_indices():
            return self._postprocess.pop(index)
        else:
            raise IndexError("Module index is out of bounds")

    def postprocess_modules(self) -> List[GenotypeModule]:
        return self._postprocess

    def postprocess_module_count(self) -> int:
        return len(self._postprocess)

    def postprocess_module_indices(self) -> List[int]:
        return list(range(0, len(self._postprocess)))

    # Layers
    def add_layer(self, layer: GenotypeLayer, index=None):
        idx = len(self._layers) if index is None else index
        self._layers.insert(idx, layer)

    def delete_layer(self, index):
        if index in self.layer_indices():
            return self._layers.pop(index)
        else:
            raise IndexError("Layer index is out of bounds")

    def get_layer(self, index) -> GenotypeLayer:
        if index in self.layer_indices():
            return self._layers[index]
        else:
            raise IndexError("Layer index is out of bounds")

    def layers(self) -> List[GenotypeLayer]:
        return self._layers

    def layer_count(self) -> int:
        if self._explicit_layer_count:
            return self._parameters[-1].get_value()
        else:
            return len(self._layers)

    def layer_indices(self) -> List[int]:
        if self._explicit_layer_count:
            indices = [0 for _ in range(self.layer_count())]
        else:
            indices = list(range(0, len(self._layers)))
        return indices

    def has_explicit_layer_count(self) -> bool:
        return self._explicit_layer_count

    def has_unbroken_residual_path(self, ignore_normalization: bool) -> bool:
        for layer in self._layers:
            if not layer.has_unbroken_residual_path(ignore_normalization):
                return False
        return True

    # Timestamp
    def timestamp(self):
        return self._timestamp

    def set_timestamp(self, timestamp):
        self._timestamp = timestamp

    # Fitness
    def fitness(self):
        return self._fitness

    def set_fitness(self, fitness):
        self._fitness = float(fitness)

    def is_dirty(self) -> bool:
        return self._fitness is None

    def mark_dirty(self):
        self._fitness = None

    # Genotype level params
    def embedding_dim(self) -> int:
        return self._parameters[0].get_value()

    def patch_size(self) -> int:
        return self._parameters[1].get_value()

    def dropout(self) -> int:
        return self._parameters[2].get_value()

    def use_memory(self) -> bool:
        return self._parameters[3].get_value()

    def use_full_regression(self) -> bool:
        return self._parameters[4].get_value()

    def parameters(self) -> List[IParameter]:
        return self._parameters

    # Serialization
    def to_json(self) -> str:
        jsonpickle.set_encoder_options("simplejson", sort_keys=True, indent=4)
        return jsonpickle.encode(self)

    @staticmethod
    def from_json(json_string: str):
        genotype = jsonpickle.decode(json_string)
        return genotype

    def save_as_json(self, filename: str):
        with open(filename, "w") as outfile:
            outfile.write(self.to_json())

    @staticmethod
    def load_from_json(filename: str):
        with open(filename, "r") as infile:
            string = infile.read()
            return Genotype.from_json(string)
