import unittest
from enum import Enum
from typing import List

from src.evolutionary.genotype.GenotypeModule import GenotypeModule


class GenotypeModuleBranch(Enum):
    LEFT = 0
    RIGHT = 1


class GenotypeCell:
    def __init__(self):
        self._left_branch: List[GenotypeModule] = []
        self._right_branch: List[GenotypeModule] = []

    def add_module(
        self, module: GenotypeModule, branch: GenotypeModuleBranch, index=None
    ):
        if branch is GenotypeModuleBranch.LEFT:
            idx = len(self._left_branch) if index is None else index
            self._left_branch.insert(idx, module)
        elif branch is GenotypeModuleBranch.RIGHT:
            idx = len(self._right_branch) if index is None else index
            self._right_branch.insert(idx, module)
        else:
            raise ValueError("Unknown branch")

    def add_modules(
        self, modules: List[GenotypeModule], branch: GenotypeModuleBranch, index=None
    ):
        if branch is GenotypeModuleBranch.LEFT:
            for module in modules:
                idx = len(self._left_branch) if index is None else index
                self._left_branch.insert(idx, module)
        elif branch is GenotypeModuleBranch.RIGHT:
            for module in modules:
                idx = len(self._right_branch) if index is None else index
                self._right_branch.insert(idx, module)
        else:
            raise ValueError("Unknown branch")

    def delete_module(self, branch: GenotypeModuleBranch, index) -> GenotypeModule:
        if index in self.module_indices(branch):
            if branch is GenotypeModuleBranch.LEFT:
                return self._left_branch.pop(index)
            elif branch is GenotypeModuleBranch.RIGHT:
                return self._right_branch.pop(index)
            else:
                raise ValueError("Unknown branch")
        else:
            raise IndexError("Module index is out of bounds")

    def delete_all_modules(self, branch: GenotypeModuleBranch) -> List[GenotypeModule]:
        if branch is GenotypeModuleBranch.LEFT:
            modules = self._left_branch
            self._left_branch = []
            return modules
        elif branch is GenotypeModuleBranch.RIGHT:
            modules = self._right_branch
            self._right_branch = []
            return modules
        else:
            raise ValueError("Unknown branch")

    def modules(self, branch: GenotypeModuleBranch) -> List[GenotypeModule]:
        if branch is GenotypeModuleBranch.LEFT:
            return self._left_branch
        elif branch is GenotypeModuleBranch.RIGHT:
            return self._right_branch
        else:
            raise ValueError("Unknown branch")

    def get_module(self, branch: GenotypeModuleBranch, index) -> GenotypeModule:
        if branch is GenotypeModuleBranch.LEFT:
            return self._left_branch[index]
        elif branch is GenotypeModuleBranch.RIGHT:
            return self._right_branch[index]
        else:
            raise ValueError("Unknown branch")

    def get_all_modules(self, branch: GenotypeModuleBranch) -> List[GenotypeModule]:
        if branch is GenotypeModuleBranch.LEFT:
            return self._left_branch
        elif branch is GenotypeModuleBranch.RIGHT:
            return self._right_branch
        else:
            raise ValueError("Unknown branch")

    def module_count(self, branch: GenotypeModuleBranch) -> int:
        if branch == GenotypeModuleBranch.LEFT:
            return len(self._left_branch)
        elif branch is GenotypeModuleBranch.RIGHT:
            return len(self._right_branch)
        else:
            raise ValueError("Unknown branch")

    def module_indices(self, branch: GenotypeModuleBranch) -> List[int]:
        if branch == GenotypeModuleBranch.LEFT:
            return list(range(0, len(self._left_branch)))
        elif branch is GenotypeModuleBranch.RIGHT:
            return list(range(0, len(self._right_branch)))
        else:
            raise ValueError("Unknown branch")

    def is_branch_empty(
        self, branch: GenotypeModuleBranch, ignore_normalization: bool
    ) -> bool:
        if branch == GenotypeModuleBranch.LEFT:
            if ignore_normalization:
                left_module_count = len(
                    [m for m in self._left_branch if not m.is_normalization()]
                )
            else:
                left_module_count = len(self._left_branch)
            return left_module_count == 0
        elif branch is GenotypeModuleBranch.RIGHT:
            if ignore_normalization:
                right_module_count = len(
                    [m for m in self._right_branch if not m.is_normalization()]
                )
            else:
                right_module_count = len(self._right_branch)
            return right_module_count == 0
        else:
            raise ValueError("Unknown branch")

    def is_left_branch_empty(self, ignore_normalization: bool) -> bool:
        return self.is_branch_empty(GenotypeModuleBranch.LEFT, ignore_normalization)

    def is_right_branch_empty(self, ignore_normalization: bool) -> bool:
        return self.is_branch_empty(GenotypeModuleBranch.RIGHT, ignore_normalization)


class TestGenotypeCell(unittest.TestCase):
    def test(self):
        cell = GenotypeCell()
        indices_left = cell.module_indices(GenotypeModuleBranch.LEFT)
        self.assertEqual(indices_left, [])
        indices_right = cell.module_indices(GenotypeModuleBranch.RIGHT)
        self.assertEqual(indices_right, [])

        dummy_module = object()
        cell.add_module(dummy_module, GenotypeModuleBranch.LEFT)
        dummy_module1 = object()
        cell.add_module(dummy_module1, GenotypeModuleBranch.LEFT)
        dummy_module2 = object()
        cell.add_module(dummy_module2, GenotypeModuleBranch.LEFT)
        indices_left = cell.module_indices(GenotypeModuleBranch.LEFT)
        self.assertEqual(indices_left, [0, 1, 2])
        indices_right = cell.module_indices(GenotypeModuleBranch.RIGHT)
        self.assertEqual(indices_right, [])

        dummy_module = object()
        cell.add_module(dummy_module, GenotypeModuleBranch.RIGHT)
        dummy_module1 = object()
        cell.add_module(dummy_module1, GenotypeModuleBranch.RIGHT)
        indices_left = cell.module_indices(GenotypeModuleBranch.LEFT)
        self.assertEqual(indices_left, [0, 1, 2])
        indices_right = cell.module_indices(GenotypeModuleBranch.RIGHT)
        self.assertEqual(indices_right, [0, 1])

        cell.delete_module(GenotypeModuleBranch.LEFT, 0)
        cell.delete_module(GenotypeModuleBranch.RIGHT, 1)

        indices_left = cell.module_indices(GenotypeModuleBranch.LEFT)
        self.assertEqual(indices_left, [0, 1])
        indices_right = cell.module_indices(GenotypeModuleBranch.RIGHT)
        self.assertEqual(indices_right, [0])
