import unittest
from typing import List

from src.evolutionary.genotype.GenotypeCell import GenotypeCell


class GenotypeLayer:
    def __init__(self, cells=None):
        if cells is None:
            cells = []
        self._cells: List[list] = cells
        self._cells.insert(0, [-1, -1, None])

    def add_cell(
        self, left_hidden_state_index, right_hidden_state_index, cell, index=None
    ):
        valid_indices = self.hidden_state_indices()
        if (
            left_hidden_state_index in valid_indices
            and right_hidden_state_index in valid_indices
        ):
            idx = len(self._cells) if index is None else index + 1
            self._cells.insert(
                idx, [left_hidden_state_index, right_hidden_state_index, cell]
            )
        else:
            raise IndexError("Hidden state index is out of bounds")

    def delete_cell(self, index) -> GenotypeCell:
        if index in self.cell_indices():
            removed_cell = self._cells.pop(index)
            for cell in self._cells:
                if cell[0] >= index:
                    cell[0] -= 1
                if cell[1] >= index:
                    cell[1] -= 1
            return removed_cell[-1]
        else:
            raise IndexError("Cell index is out of bounds")

    def get_cell(self, index) -> GenotypeCell:
        if index in self.cell_indices():
            return self._cells[index][-1]
        else:
            raise IndexError("Layer index is out of bounds")

    def get_cell_hidden_state_indices(self, cell_index) -> list:
        if cell_index in self.cell_indices():
            return self._cells[cell_index][:2]
        else:
            raise IndexError("Layer index is out of bounds")

    def set_cell_hidden_state_indices(self, cell_index, left_index, right_index):
        if cell_index in self.cell_indices():
            self._cells[cell_index][0] = left_index
            self._cells[cell_index][1] = right_index
        else:
            raise IndexError("Layer index is out of bounds")

    def cells(self, without_first=True) -> List[list]:
        if without_first:
            return self._cells[1:]
        else:
            return self._cells

    def cell_count(self) -> int:
        return len(self._cells) - 1

    def cell_indices(self) -> List[int]:
        return list(range(1, len(self._cells)))

    def hidden_state_indices(self) -> List[int]:
        return list(range(0, len(self._cells)))

    # Test of unbroken residual path
    def _unbroken_path_recursive(self, cell_index: int, ignore_normalization) -> bool:
        input_id1 = self._cells[cell_index][0]
        input_id2 = self._cells[cell_index][1]
        current_cell = self._cells[cell_index][2]

        if input_id1 == -1:
            return False
        elif input_id1 == 0 and current_cell.is_left_branch_empty(ignore_normalization):
            return True
        else:
            input1_has_path = self._unbroken_path_recursive(
                input_id1, ignore_normalization
            )

        if input_id2 == -1:
            return False
        elif input_id2 == 0 and current_cell.is_right_branch_empty(
            ignore_normalization
        ):
            return True
        else:
            input2_has_path = self._unbroken_path_recursive(
                input_id2, ignore_normalization
            )

        if input1_has_path or input2_has_path:
            return True
        else:
            return False

    def has_unbroken_residual_path(self, ignore_normalization: bool) -> bool:
        # All cells that are not referenced (including the last cell) are added to the output
        used_cell_indices = set()
        for cell in self._cells:
            used_cell_indices.add(cell[0])
            used_cell_indices.add(cell[1])
        all_cell_indices = set(list(range(0, len(self._cells))))
        unused_cell_indices = list(all_cell_indices - used_cell_indices)

        for idx in unused_cell_indices:
            if self._unbroken_path_recursive(idx, ignore_normalization):
                return True

        return False


class TestGenotypeLayer(unittest.TestCase):
    def test(self):
        layer = GenotypeLayer()
        indices = layer.cell_indices()
        self.assertEqual(indices, [])
        indices = layer.hidden_state_indices()
        self.assertEqual(indices, [0])

        dummy_cell = object()
        layer.add_cell(0, 0, dummy_cell)
        indices = layer.cell_indices()
        self.assertEqual(indices, [1])
        indices = layer.hidden_state_indices()
        self.assertEqual(indices, [0, 1])

        dummy_cell2 = object()
        layer.add_cell(0, 1, dummy_cell2)
        dummy_cell3 = object()
        layer.add_cell(1, 2, dummy_cell3)
        dummy_cell4 = object()
        layer.add_cell(3, 2, dummy_cell4)
        dummy_cell5 = object()
        layer.add_cell(3, 4, dummy_cell5)
        indices = layer.cell_indices()
        self.assertEqual(indices, [1, 2, 3, 4, 5])
        indices = layer.hidden_state_indices()
        self.assertEqual(indices, [0, 1, 2, 3, 4, 5])

        layer.delete_cell(3)
        indices = layer.cell_indices()
        self.assertEqual(indices, [1, 2, 3, 4])
        indices = layer.hidden_state_indices()
        self.assertEqual(indices, [0, 1, 2, 3, 4])
