import copy
import json
import random
from typing import List, Optional

from src.evolutionary.genotype.GenotypeModule import GenotypeModule
from src.evolutionary.genotype.modules.ModuleCategory import ModuleCategory


class ModuleRegistry:
    def __init__(self, modules: List[GenotypeModule] = None):
        if modules is None:
            modules = []
        self._modules = modules

    def add_module(self, module):
        self._modules.append(module)

    def find_module(self, path: str, absolute: bool = True) -> Optional[GenotypeModule]:
        found_module = None
        for module in self._modules:
            if module.get_module_path(absolute) == path:
                found_module = copy.deepcopy(module)
                break
        return found_module

    def pop_module(self, path: str, absolute: bool = True) -> Optional[GenotypeModule]:
        popped_module = None
        for idx, module in enumerate(self._modules):
            if module.get_module_path(absolute) == path:
                popped_module = self._modules.pop(idx)
                break
        return popped_module

    def get_random_module(self) -> GenotypeModule:
        random_module = random.choice(self._modules)
        return copy.deepcopy(random_module)

    def modules(self) -> List[GenotypeModule]:
        return self._modules

    def modules_for_categories(
        self, category_filter: List[ModuleCategory]
    ) -> List[GenotypeModule]:
        categories = [c.value for c in category_filter]
        modules = [
            copy.deepcopy(m) for m in self._modules if m.get_category() in categories
        ]
        return modules

    def module_count(self) -> int:
        return len(self._modules)

    def to_dict(self) -> dict:
        return {"modules": [m.to_dict() for m in self._modules]}

    @staticmethod
    def from_dict(data: dict):
        modules_list = data["modules"]
        modules = [GenotypeModule.from_dict(m_dict) for m_dict in modules_list]
        return ModuleRegistry(modules)

    @staticmethod
    def from_list(data: list):
        modules = {"modules": data}
        return ModuleRegistry.from_dict(modules)

    def save_as_json(self, filename: str):
        with open(filename, "w") as outfile:
            json_str = json.dumps(self.to_dict())
            outfile.write(json_str)

    @staticmethod
    def load_from_json(filename: str):
        with open(filename, "r") as infile:
            json_string = infile.read()
            dict_data = json.loads(json_string)
            return ModuleRegistry.from_dict(dict_data)
