import unittest
import torch

from src.evolutionary.core.conversion import create_phenotype_module
from src.evolutionary.genotype.GenotypeModule import GenotypeModule
from src.evolutionary.genotype.modules.ModuleCategory import ModuleCategory
from src.utils.helpers import save_json


def save_as_json(path):
    dicts = [m.to_dict() for m in modules]
    save_json(path, dicts)


modules = [
    GenotypeModule(
        "src.components.activation.ELU",
        {},
        ModuleCategory.ACTIVATION.value,
    ),
    GenotypeModule(
        "src.components.activation.GeLU",
        {},
        ModuleCategory.ACTIVATION.value,
    ),
    GenotypeModule(
        "src.components.activation.LeakyReLU",
        {},
        ModuleCategory.ACTIVATION.value,
    ),
    GenotypeModule(
        "src.components.activation.ReLU",
        {},
        ModuleCategory.ACTIVATION.value,
    ),
    GenotypeModule(
        "src.components.activation.SiLU",
        {},
        ModuleCategory.ACTIVATION.value,
    ),
    GenotypeModule(
        "src.components.activation.SquaredReLU",
        {},
        ModuleCategory.ACTIVATION.value,
    ),
]


class TestActivationModules(unittest.TestCase):
    def test_instantiation_of_all_modules(self):
        batch_size = 16
        sequence_length = 200
        embedding_dim = 16
        dropout = 0.1

        shared_params = {
            "embedding_dim": embedding_dim,
            "input_dim": sequence_length,
            "dropout": dropout,
        }
        input_tensor = torch.rand(batch_size, sequence_length, embedding_dim)

        phenotype_modules = []
        output_tensors = []
        for module in modules:
            phenotype_module = create_phenotype_module(
                module, shared_params=shared_params
            )
            phenotype_modules.append(phenotype_module)
            output_tensor = phenotype_module(input_tensor)
            output_tensors.append(output_tensor)

        self.assertEqual(len(phenotype_modules), len(modules))
        for t in output_tensors:
            self.assertEqual(t.shape, (batch_size, sequence_length, embedding_dim))

    def test_all_parameters(self):
        batch_size = 16
        dropout = 0.1

        # The shortest possible sequence length is 8 i.e. patch_size = 50 / input_dim = 400
        # Available patch sizes: [1, 5, 10, 25, 50]
        sequence_lengths = [8, 20, 40, 200]
        embedding_dims = [16, 32, 64]

        for ed in embedding_dims:
            for sl in sequence_lengths:
                for module in modules:
                    input_tensor = torch.rand(batch_size, sl, ed)
                    shared_params = {
                        "embedding_dim": ed,
                        "input_dim": sl,
                        "dropout": dropout,
                    }

                    for key, param in module.parameters.items():
                        allowed_values = param.get_allowed_values()
                        for value in allowed_values:
                            override_params = {key: value}
                            try:
                                phenotype_module = create_phenotype_module(
                                    module,
                                    shared_params=shared_params,
                                    override_params=override_params,
                                )
                                output_tensor = phenotype_module(input_tensor)
                            except Exception as e:
                                print(
                                    f"embedding_dim: {ed}, input_dim: {sl}, module: {module.module_path}"
                                )
                                print(f"param: {key}, value: {value}")
                                raise e

                            self.assertEqual(output_tensor.shape, (batch_size, sl, ed))
