import unittest
import torch

from src.components.activation.Activation import Activation
from src.components.attention.preprocess.QKVPreprocess import QKVPreprocess
from src.components.attention.preprocess.ScoresPreprocess import ScoresPreprocess
from src.evolutionary.core.conversion import create_phenotype_module
from src.evolutionary.genotype.GenotypeModule import GenotypeModule
from src.evolutionary.genotype.modules.ModuleCategory import ModuleCategory
from src.evolutionary.genotype.parameters.ConstantParameter import ConstantParameter
from src.evolutionary.genotype.parameters.RangeParameter import RangeParameter
from src.evolutionary.genotype.parameters.SetParameter import SetParameter
from src.utils.helpers import save_json


def save_as_json(path):
    dicts = [m.to_dict() for m in modules]
    save_json(path, dicts)


# NOTE: Big patch sizes can cause issues as they produce very short sequences!!!
modules = [
    GenotypeModule(
        "src.components.attention.AdaptiveSpanAttention",
        {
            "head_count": SetParameter(4, [1, 2, 4, 8]),
            "span_length_init": RangeParameter(0.1, (0.0, 1.0), 11),
            "span_ramp": SetParameter(10, [10, 20, 40, 80]),
        },
        ModuleCategory.ATTENTION.value,
    ),
    GenotypeModule(
        "src.components.attention.AttentionFreeFullAttention",
        {
            "head_count": SetParameter(4, [1, 2, 4, 8]),
        },
        ModuleCategory.ATTENTION.value,
    ),
    GenotypeModule(
        "src.components.attention.AttentionFreeLocalAttention",
        {
            "window_size_divider": SetParameter(4, [2, 4, 8]),
        },
        ModuleCategory.ATTENTION.value,
    ),
    GenotypeModule(
        "src.components.attention.AttentionFreeSimpleAttention",
        {
            "head_count": SetParameter(4, [1, 2, 4, 8]),
        },
        ModuleCategory.ATTENTION.value,
    ),
    # NOTE: This one is very sensitive to short sequences (i.e. big patch sizes)!!! Length below 8 can crash it
    GenotypeModule(
        "src.components.attention.BigBirdBlockSparseAttention",
        {
            "head_count": SetParameter(4, [1, 2, 4, 8]),
            "random_block_count": ConstantParameter(1),
            "block_size": ConstantParameter(1),
        },
        ModuleCategory.ATTENTION.value,
    ),
    GenotypeModule(
        "src.components.attention.ConformerConvolutionAttention",
        {
            "dropout": RangeParameter(0.1, (0.0, 0.3), 7),
        },
        ModuleCategory.ATTENTION.value,
    ),
    GenotypeModule(
        "src.components.attention.FactorizedDenseAttentionSynthesizer",
        {
            "head_count": SetParameter(4, [1, 2, 4, 8]),
            "dropout": RangeParameter(0.1, (0.0, 0.3), 7),
        },
        ModuleCategory.ATTENTION.value,
    ),
    GenotypeModule(
        "src.components.attention.FactorizedRandomAttentionSynthesizer",
        {
            "head_count": SetParameter(4, [1, 2, 4, 8]),
            "dropout": RangeParameter(0.1, (0.0, 0.3), 7),
        },
        ModuleCategory.ATTENTION.value,
    ),
    # NOTE: Supports only 1 attention head
    GenotypeModule(
        "src.components.attention.FastWeightsAttention",
        {},
        ModuleCategory.ATTENTION.value,
    ),
    # NOTE: This one has to have at least two heads!
    GenotypeModule(
        "src.components.attention.FavorAttentionPerformer",
        {
            "head_count": SetParameter(4, [2, 4, 8]),
            "dropout": RangeParameter(0.1, (0.0, 0.3), 7),
        },
        ModuleCategory.ATTENTION.value,
    ),
    GenotypeModule(
        "src.components.attention.FNetFourierTransformAttention",
        {
            "head_count": SetParameter(4, [1, 2, 4, 8]),
        },
        ModuleCategory.ATTENTION.value,
    ),
    GenotypeModule(
        "src.components.attention.KmeansRoutingAttention",
        {
            "head_count": SetParameter(4, [1, 2, 4, 8]),
            "cluster_count": SetParameter(10, [5, 10, 20]),
            "window_size_divider": SetParameter(4, [2, 4, 8]),
            "dropout": RangeParameter(0.1, (0.0, 0.3), 7),
        },
        ModuleCategory.ATTENTION.value,
    ),
    GenotypeModule(
        "src.components.attention.LambdaLayerAttention",
        {
            "head_count": SetParameter(4, [1, 2, 4, 8]),
            "dropout": RangeParameter(0.1, (0.0, 0.3), 7),
        },
        ModuleCategory.ATTENTION.value,
    ),
    GenotypeModule(
        "src.components.attention.LinearAttentionTransformersRNNs",
        {
            "head_count": SetParameter(4, [1, 2, 4, 8]),
            "activation": SetParameter(
                Activation.ELU.value, Activation.module_string_list()
            ),
        },
        ModuleCategory.ATTENTION.value,
    ),
    GenotypeModule(
        "src.components.attention.LinformerAttention",
        {
            "head_count": SetParameter(2, [1, 2, 4, 8]),
        },
        ModuleCategory.ATTENTION.value,
    ),
    GenotypeModule(
        "src.components.attention.LocalAttention",
        {
            "head_count": SetParameter(4, [1, 2, 4, 8]),
            "window_size_divider": SetParameter(4, [2, 4, 8]),
            "dropout": RangeParameter(0.1, (0.0, 0.3), 7),
        },
        ModuleCategory.ATTENTION.value,
    ),
    GenotypeModule(
        "src.components.attention.LocalitySensitiveHashingAttentionReformer",
        {
            "head_count": SetParameter(4, [1, 2, 4, 8]),
            "chunk_length_divider": SetParameter(2, [2, 4, 8]),
            "dropout": RangeParameter(0.1, (0.0, 0.3), 7),
        },
        ModuleCategory.ATTENTION.value,
    ),
    GenotypeModule(
        "src.components.attention.LongformerAttention",
        {
            "head_count": SetParameter(4, [1, 2, 4, 8]),
            "window_size_divider": ConstantParameter(2),
            "dropout": RangeParameter(0.1, (0.0, 0.3), 7),
        },
        ModuleCategory.ATTENTION.value,
    ),
    # Also needs to have sequence of length 8 at least
    GenotypeModule(
        "src.components.attention.LongShortAttention",
        {
            "head_count": SetParameter(4, [1, 2, 4, 8]),
            "window_size_divider": SetParameter(2, [2, 4, 8]),
            "dropout": RangeParameter(0.1, (0.0, 0.3), 7),
        },
        ModuleCategory.ATTENTION.value,
    ),
    GenotypeModule(
        "src.components.attention.RectifiedLinearAttention",
        {
            "head_count": SetParameter(4, [1, 2, 4, 8]),
        },
        ModuleCategory.ATTENTION.value,
    ),
    GenotypeModule(
        "src.components.attention.RelativeMultiHeadAttentionXL",
        {
            "head_count": SetParameter(4, [1, 2, 4, 8]),
        },
        ModuleCategory.ATTENTION.value,
    ),
    GenotypeModule(
        "src.components.attention.ScaledDotProductMultiHeadAttention",
        {
            "head_count": SetParameter(4, [1, 2, 4, 8]),
            "q_preprocess": SetParameter(
                QKVPreprocess.NoPreprocess.value, QKVPreprocess.module_string_list()
            ),
            "k_preprocess": SetParameter(
                QKVPreprocess.NoPreprocess.value, QKVPreprocess.module_string_list()
            ),
            "v_preprocess": SetParameter(
                QKVPreprocess.NoPreprocess.value, QKVPreprocess.module_string_list()
            ),
            "scores_preprocess": SetParameter(
                ScoresPreprocess.NoPreprocess.value,
                ScoresPreprocess.module_string_list(),
            ),
        },
        ModuleCategory.ATTENTION.value,
    ),
    GenotypeModule(
        "src.components.attention.TupeMultiHeadAttention",
        {
            "head_count": SetParameter(4, [1, 2, 4, 8]),
            "bucket_count": SetParameter(16, [4, 8, 16, 32]),
            "dropout": RangeParameter(0.1, (0.0, 0.3), 7),
        },
        ModuleCategory.ATTENTION.value,
    ),
]


class TestAttentionModules(unittest.TestCase):
    def test_instantiation_of_all_modules(self):
        batch_size = 16
        sequence_length = 200
        embedding_dim = 16
        dropout = 0.1

        shared_params = {
            "embedding_dim": embedding_dim,
            "input_dim": sequence_length,
            "dropout": dropout,
        }
        input_tensor = torch.rand(batch_size, sequence_length, embedding_dim)

        phenotype_modules = []
        output_tensors = []
        for module in modules:
            phenotype_module = create_phenotype_module(
                module, shared_params=shared_params
            )
            phenotype_modules.append(phenotype_module)
            output_tensor = phenotype_module(input_tensor)
            output_tensors.append(output_tensor)

        self.assertEqual(len(phenotype_modules), len(modules))
        for t in output_tensors:
            self.assertEqual(t.shape, (batch_size, sequence_length, embedding_dim))

    def test_all_parameters(self):
        batch_size = 16
        dropout = 0.1

        # The shortest possible sequence length is 8 i.e. patch_size = 50 / input_dim = 400
        # Available patch sizes: [1, 5, 10, 25, 50]
        sequence_lengths = [8, 20, 40, 200]
        embedding_dims = [16, 32, 64]

        for ed in embedding_dims:
            for sl in sequence_lengths:
                for module in modules:
                    input_tensor = torch.rand(batch_size, sl, ed)
                    shared_params = {
                        "embedding_dim": ed,
                        "input_dim": sl,
                        "dropout": dropout,
                    }

                    for key, param in module.parameters.items():
                        allowed_values = param.get_allowed_values()
                        for value in allowed_values:
                            override_params = {key: value}
                            try:
                                phenotype_module = create_phenotype_module(
                                    module,
                                    shared_params=shared_params,
                                    override_params=override_params,
                                )
                                output_tensor = phenotype_module(input_tensor)
                            except Exception as e:
                                print(
                                    f"embedding_dim: {ed}, input_dim: {sl}, module: {module.module_path}"
                                )
                                print(f"param: {key}, value: {value}")
                                raise e

                            self.assertEqual(output_tensor.shape, (batch_size, sl, ed))
