import unittest
import torch

from src.evolutionary.core.conversion import create_phenotype_module
from src.evolutionary.genotype.GenotypeModule import GenotypeModule
from src.evolutionary.genotype.modules.ModuleCategory import ModuleCategory
from src.evolutionary.genotype.parameters.SetParameter import SetParameter
from src.utils.helpers import save_json


def save_as_json(path):
    dicts = [m.to_dict() for m in modules]
    save_json(path, dicts)


modules = [
    GenotypeModule(
        "src.components.convolution.Conv1d",
        {
            "kernel_size": SetParameter(3, [3, 6, 12]),
            "dilation": SetParameter(1, [1, 2, 3]),
        },
        ModuleCategory.CONVOLUTION.value,
    ),
    GenotypeModule(
        "src.components.convolution.Conv2d",
        {
            "kernel_width": SetParameter(3, [3, 6, 12]),
            "kernel_height": SetParameter(3, [3, 6, 12]),
            "dilation_width": SetParameter(1, [1, 2, 3]),
            "dilation_height": SetParameter(1, [1, 2, 3]),
        },
        ModuleCategory.CONVOLUTION.value,
    ),
    GenotypeModule(
        "src.components.convolution.CoordConv",
        {
            "kernel_size": SetParameter(3, [1, 3, 6, 12]),
            "dilation": SetParameter(1, [1, 2, 3]),
        },
        ModuleCategory.CONVOLUTION.value,
    ),
    GenotypeModule(
        "src.components.convolution.DepthWiseSeparableConv",
        {
            "kernel_size": SetParameter(3, [3, 6, 12]),
            "dilation": SetParameter(1, [1, 2, 3]),
        },
        ModuleCategory.CONVOLUTION.value,
    ),
    GenotypeModule(
        "src.components.convolution.InvertedConvBottleneck",
        {
            "forward_expansion": SetParameter(1, [1, 2, 4, 8]),
            "kernel_size": SetParameter(3, [3, 6, 12]),
            "dilation": SetParameter(1, [1, 2, 3]),
        },
        ModuleCategory.CONVOLUTION.value,
    ),
    GenotypeModule(
        "src.components.convolution.MaxPool1d",
        {
            "kernel_size": SetParameter(3, [3, 6, 12, 24]),
        },
        ModuleCategory.CONVOLUTION.value,
    ),
    GenotypeModule(
        "src.components.convolution.MaxPool2d",
        {
            "kernel_width": SetParameter(3, [3, 6, 12, 24]),
            "kernel_height": SetParameter(3, [3, 6, 12, 24]),
        },
        ModuleCategory.CONVOLUTION.value,
    ),
]


class TestConvolutionModules(unittest.TestCase):
    def test_instantiation_of_all_modules(self):
        batch_size = 16
        sequence_length = 200
        embedding_dim = 16
        dropout = 0.1

        shared_params = {
            "embedding_dim": embedding_dim,
            "input_dim": sequence_length,
            "dropout": dropout,
        }
        input_tensor = torch.rand(batch_size, sequence_length, embedding_dim)

        phenotype_modules = []
        output_tensors = []
        for module in modules:
            phenotype_module = create_phenotype_module(
                module, shared_params=shared_params
            )
            phenotype_modules.append(phenotype_module)
            output_tensor = phenotype_module(input_tensor)
            output_tensors.append(output_tensor)

        self.assertEqual(len(phenotype_modules), len(modules))
        for t in output_tensors:
            self.assertEqual(t.shape, (batch_size, sequence_length, embedding_dim))

    def test_all_parameters(self):
        batch_size = 16
        sequence_length = 200
        embedding_dim = 16
        shared_params = {"embedding_dim": embedding_dim, "input_dim": sequence_length}

        kernel_widths = [3, 6, 12, 24]
        kernel_heights = [3, 6, 12, 24]
        dilation_widths = [1, 2, 3]
        dilation_heights = [1, 2, 3]

        input_tensor = torch.rand(batch_size, sequence_length, embedding_dim)
        for module in modules:
            for kw in kernel_widths:
                for kh in kernel_heights:
                    for dw in dilation_widths:
                        for dh in dilation_heights:
                            override_params = {
                                "kernel_size": kw,
                                "kernel_width": kw,
                                "kernel_height": kh,
                                "dilation": dw,
                                "dilation_width": dw,
                                "dilation_height": dh,
                            }
                            phenotype_module = create_phenotype_module(
                                module,
                                shared_params=shared_params,
                                override_params=override_params,
                            )
                            output = phenotype_module(input_tensor)
                            self.assertEqual(
                                output.shape,
                                (batch_size, sequence_length, embedding_dim),
                            )
