import unittest
from typing import Tuple, List, TypeVar
import numpy as np

from src.evolutionary.genotype.parameters.IParameter import IParameter

T = TypeVar("T", int, float)


class RangeParameter(IParameter):
    def __init__(
        self, initial_value: T, allowed_range: Tuple[T, T], value_count: int = 100
    ):
        self.value: T = initial_value
        self.allowed_range = list(allowed_range)

        if isinstance(initial_value, int):
            self.allowed_values = list(
                np.arange(allowed_range[0], allowed_range[1] + 1)
            )
            self.allowed_values = [int(v) for v in self.allowed_values]
        elif isinstance(initial_value, float):
            self.allowed_values = list(
                np.round(
                    np.linspace(allowed_range[0], allowed_range[1], value_count), 3
                )
            )
            self.allowed_values = [float(v) for v in self.allowed_values]
        self.value_count = len(self.allowed_values)
        self.type = "range"

    def get_value(self):
        return self.value

    def set_value(self, value: T):
        if value not in self.allowed_values:
            raise ValueError("Value not allowed")
        self.value = value

    def get_allowed_values(self) -> List[T]:
        return self.allowed_values

    # Serialize/deserialize
    def to_dict(self) -> dict:
        data = {
            "value": self.value,
            "allowed_range": self.allowed_range,
            "value_count": self.value_count,
            "type": self.type,
        }
        return data

    @staticmethod
    def from_dict(data: dict):
        value = data["value"]
        allowed_range = data["allowed_range"]
        value_count = data["value_count"]
        return RangeParameter(value, allowed_range, value_count)


class TestRangeParameter(unittest.TestCase):
    def test1(self):
        parameter = RangeParameter(0, (0, 5))
        self.assertEqual(parameter.get_allowed_values(), [0, 1, 2, 3, 4, 5])
        self.assertEqual(parameter.get_value(), 0)
        parameter.set_value(2)
        self.assertEqual(parameter.get_value(), 2)

    def test2(self):
        parameter = RangeParameter(0.0, (0.0, 10.0), 6)
        self.assertEqual(
            parameter.get_allowed_values(), [0.0, 2.0, 4.0, 6.0, 8.0, 10.0]
        )
        self.assertEqual(parameter.get_value(), 0)
        parameter.set_value(2.0)
        self.assertEqual(parameter.get_value(), 2.0)

    def test_serialize_deserialize(self):
        parameter1 = RangeParameter(0.0, (0.0, 10.0), 6)
        dict_data = parameter1.to_dict()
        parameter2 = RangeParameter.from_dict(dict_data)

        self.assertEqual(parameter1.get_value(), parameter2.get_value())
        self.assertEqual(
            parameter1.get_allowed_values(), parameter2.get_allowed_values()
        )
