import json
from typing import NamedTuple

from src.utils.helpers import namedtuple_fields_with_types


class DatasetConfig(NamedTuple):
    input_dim: int = 400  # This has to be at least 400 because of Big Bird attention
    output_dim: int = 50
    batch_size: int = 16  # Better to keep it at 16 (best compromise)
    train_test_ratio: float = 0.8
    dataset_type: str = "libra"
    synthetic_length: int = 5000
    libra_categories: list[str] = ["economics", "finance", "human", "nature"]
    libra_dataset_filter: list[list] = [
        [71, 86, 69, 75, 70, 83, 80, 74, 81, 77],
        [63, 96, 49, 39, 71, 18, 99, 66, 82, 17],
        [86, 56, 46, 21, 67, 5, 2, 92, 68, 97],
        [0, 91, 39, 70, 23, 42, 73, 1, 53, 24],
    ]
    libra_max_dataset_values_crop_limit: int = 3000
    libra_min_training_examples_per_dataset = 100

    @staticmethod
    def fields() -> dict:
        return namedtuple_fields_with_types(DatasetConfig)

    def to_dict(self) -> dict:
        return dict(self._asdict())

    @staticmethod
    def from_dict(data: dict):
        data_fields = list(data.keys())
        data_fields.sort()
        type_fields = list(DatasetConfig._fields)
        type_fields.sort()

        if tuple(data_fields) != tuple(type_fields):
            raise ValueError("Field mismatch during deserialization")

        return DatasetConfig(**data)

    @staticmethod
    def load_from_json(filename: str):
        with open(filename, "r") as infile:
            string = infile.read()
            data = json.loads(string)
            return DatasetConfig.from_dict(data)


class EvaluationConfig(NamedTuple):
    max_epochs: int = 5
    hurdle_after_iterations: int = 40
    train_loss: str = "smoothl1"
    test_loss: str = "smoothl1"

    @staticmethod
    def fields() -> dict:
        return namedtuple_fields_with_types(EvaluationConfig)

    def to_dict(self) -> dict:
        return dict(self._asdict())

    @staticmethod
    def from_dict(data: dict):
        data_fields = list(data.keys())
        data_fields.sort()
        type_fields = list(EvaluationConfig._fields)
        type_fields.sort()

        if tuple(data_fields) != tuple(type_fields):
            raise ValueError("Field mismatch during deserialization")

        return EvaluationConfig(**data)

    @staticmethod
    def load_from_json(filename: str):
        with open(filename, "r") as infile:
            string = infile.read()
            data = json.loads(string)
            return EvaluationConfig.from_dict(data)


class EvolutionConfig(NamedTuple):
    iterations: int = 200
    population_size: int = 100
    parent_tournament_size: int = 2
    replacement_tournament_size: int = 4
    replacement_candidate_selection: str = "worst_fitness"
    enable_independent_layer_evolution = False

    # Initialization
    population_initialization_method: str = "full_encoder"
    prepend_positional_encoding: bool = True
    keep_positional_encoding_in_registry: bool = True
    transformer_encoder_type: str = "pre_norm"  # "full_encoder" initialization only
    non_encoder_number_of_cells: int = 2  # Non "full_encoder" initialization only

    randomize_initial_population: bool = True
    non_randomized_genotype_count: int = 4
    init_layer_mutation_iterations: int = 5
    init_cell_mutation_iterations: int = 5
    init_module_mutation_iterations: int = 10
    init_parameter_mutation_iterations: int = 20

    # Layers
    # mutations
    global_layer_mutation_probability: float = 0.2
    add_layer_mutation_probability: float = 0.1
    duplicate_layer_mutation_probability: float = 0.1
    swap_two_layers_mutation_probability: float = 0.2
    remove_layer_mutation_probability: float = 0.2
    # crossover
    swap_random_layers_crossover_probability: float = 0.2

    # Cells
    # mutations
    global_cell_mutation_probability: float = 0.2
    add_cell_mutation_probability: float = 0.1
    duplicate_cell_mutation_probability: float = 0.1
    swap_two_cells_mutation_probability: float = 0.2
    change_cell_inputs_mutation_probability: float = 0.2
    remove_cell_mutation_probability: float = 0.2
    # crossover
    swap_random_cells_crossover_probability: float = 0.2
    swap_random_branches_crossover_probability: float = 0.2

    # Modules
    # mutations
    global_module_mutation_probability: float = 0.3
    add_module_mutation_probability: float = 0.2
    duplicate_module_mutation_probability: float = 0.1
    swap_two_modules_mutation_probability: float = 0.2
    swap_module_for_new_mutation_probability: float = 0.2
    respect_module_category_probability: float = 0.5
    remove_module_mutation_probability: float = 0.3

    # Parameters
    # mutations
    global_parameter_mutation_probability: float = 0.3
    change_genotype_parameter_mutation_probability: float = 0.1
    change_parameter_mutation_probability: float = 0.2

    @staticmethod
    def fields() -> dict:
        return namedtuple_fields_with_types(EvolutionConfig)

    def to_dict(self) -> dict:
        return dict(self._asdict())

    @staticmethod
    def from_dict(data: dict):
        data_fields = list(data.keys())
        data_fields.sort()
        type_fields = list(EvolutionConfig._fields)
        type_fields.sort()

        if tuple(data_fields) != tuple(type_fields):
            raise ValueError("Field mismatch during deserialization")

        return EvolutionConfig(**data)

    @staticmethod
    def load_from_json(filename: str):
        with open(filename, "r") as infile:
            string = infile.read()
            data = json.loads(string)
            return EvolutionConfig.from_dict(data)


class GenotypeConfig(NamedTuple):
    embedding_dim: list[int, list[int]] = (16, [16, 32, 64])
    patch_size: list[int, list[int]] = (1, [1, 5, 10, 25, 50])
    layer_count: list[int, list[int]] = (1, (1, 8), 8)
    dropout: list[float, list[float], float] = (0.1, (0.0, 0.3), 7)
    use_memory: list[bool, list[bool]] = (False, [False])
    use_full_regression: list[bool, list[bool]] = (True, [True, False])

    @staticmethod
    def fields() -> dict:
        return namedtuple_fields_with_types(GenotypeConfig)

    def to_dict(self) -> dict:
        return dict(self._asdict())

    @staticmethod
    def from_dict(data: dict):
        data_fields = list(data.keys())
        data_fields.sort()
        type_fields = list(GenotypeConfig._fields)
        type_fields.sort()

        if tuple(data_fields) != tuple(type_fields):
            raise ValueError("Field mismatch during deserialization")

        return GenotypeConfig(**data)

    @staticmethod
    def load_from_json(filename: str):
        with open(filename, "r") as infile:
            string = infile.read()
            data = json.loads(string)
            return GenotypeConfig.from_dict(data)


class GlobalConfig(NamedTuple):
    dataset: dict
    evaluation: dict
    evolution: dict
    genotype: dict
    modules: list

    @staticmethod
    def fields() -> dict:
        return namedtuple_fields_with_types(GlobalConfig)

    def to_dict(self) -> dict:
        data = dict(self._asdict())
        return data

    @staticmethod
    def from_dict(data: dict):
        data_fields = list(data.keys())
        data_fields.sort()
        type_fields = list(GlobalConfig._fields)
        type_fields.sort()

        if tuple(data_fields) != tuple(type_fields):
            raise ValueError("Field mismatch during deserialization")

        return GlobalConfig(**data)

    @staticmethod
    def load_from_json(filename: str):
        with open(filename, "r") as infile:
            string = infile.read()
            data = json.loads(string)
            return GlobalConfig.from_dict(data)

    def save_as_json(self, filename: str):
        with open(filename, "w") as outfile:
            json_str = json.dumps(self.to_dict(), indent=4)
            outfile.write(json_str)
