import copy
import datetime
import json
import math
import random
import time
from typing import List, Tuple

import torch

from src.dataset.DatasetType import DatasetType
from src.dataset.LibraDataset import LibraCategory
from src.evolutionary.core.difference import population_difference_factor
from src.evolutionary.core.evaluation import EvaluationSetup, evaluate_population
from src.evolutionary.core.population import (
    randomize_population,
    TransformerEncoderType,
    PopulationInitializationMethod,
    initialize_population,
)
from src.evolutionary.core.selection import (
    SelectionMode,
    tournament_selection_with_hurdles,
)
from src.evolutionary.core.visualization import visualize_genotype
from src.evolutionary.dataset.bootstrap import (
    boostrap_synthetic_dataset,
    boostrap_libra_dataset,
)
from src.evolutionary.genotype.Genotype import (
    Genotype,
    GenotypeConfig as NativeGenotypeConfig,
)
from src.evolutionary.genotype.GenotypeModule import GenotypeModule
from src.evolutionary.genotype.modules.ModuleRegistry import ModuleRegistry
from src.evolutionary.main.config import (
    DatasetConfig,
    EvaluationConfig,
    EvolutionConfig,
    GenotypeConfig,
    GlobalConfig,
)
from src.evolutionary.operators.mutation import (
    add_new_cell,
    add_new_random_module,
    duplicate_random_cell,
    swap_two_random_modules,
    delete_random_cell,
    remove_random_module,
    duplicate_random_module,
    change_random_cell_inputs,
    change_random_parameter_genotype,
    change_random_parameter,
    swap_two_random_cells,
    swap_module_for_new_random_module,
    add_new_layer,
    duplicate_random_layer,
    swap_two_random_layers,
    delete_random_layer,
)
from src.evolutionary.operators.crossover import (
    swap_random_cells as swap_random_cells_crossover,
    swap_random_branches as swap_random_branches_crossover,
    swap_random_layers as swap_random_layers_crossover,
)
from src.training.loss import Loss
from src.utils.Tensorboard import Tensorboard
from src.utils.helpers import mean
from src.utils.logger import logger

log = logger()


def bootstrap_dataset_from_config(config: DatasetConfig) -> Tuple[list, list]:
    if config.dataset_type == DatasetType.SYNTHETIC:
        train_set, test_set = boostrap_synthetic_dataset(
            config.input_dim,
            config.output_dim,
            config.synthetic_length,
            config.train_test_ratio,
        )
    elif config.dataset_type == DatasetType.LIBRA:
        categories = [LibraCategory.from_str(c) for c in config.libra_categories]
        train_set, test_set = boostrap_libra_dataset(
            config.input_dim,
            config.output_dim,
            categories,
            config.libra_dataset_filter,
            config.libra_min_training_examples_per_dataset,
            config.libra_max_dataset_values_crop_limit,
            config.train_test_ratio,
        )
    else:
        raise ValueError("Unknown dataset type")

    return train_set, test_set


def initialize_and_randomize_population(
    genotype_config: GenotypeConfig,
    evolution_config: EvolutionConfig,
    population_size: int,
    cell_count: int,
    module_registry: ModuleRegistry,
    transformer_encoder_type: TransformerEncoderType,
    prepend_positional_encoding: bool,
    keep_positional_encoding_in_registry: bool,
    population_initialization_method: PopulationInitializationMethod,
) -> List[Genotype]:
    native_genotype_config = NativeGenotypeConfig()
    native_genotype_config.initial_embedding_dim = genotype_config.embedding_dim[0]
    native_genotype_config.allowed_embedding_dim = genotype_config.embedding_dim[1]
    native_genotype_config.initial_patch_size = genotype_config.patch_size[0]
    native_genotype_config.allowed_patch_size = genotype_config.patch_size[1]
    native_genotype_config.initial_dropout = genotype_config.dropout[0]
    native_genotype_config.allowed_range_dropout = genotype_config.dropout[1]
    native_genotype_config.value_count_dropout = genotype_config.dropout[2]
    native_genotype_config.initial_use_memory = genotype_config.use_memory[0]
    native_genotype_config.allowed_use_memory_states = genotype_config.use_memory[1]
    native_genotype_config.initial_use_full_regression = (
        genotype_config.use_full_regression[0]
    )
    native_genotype_config.allowed_use_full_regression_states = (
        genotype_config.use_full_regression[1]
    )

    native_genotype_config.explicit_layer_count = (
        not evolution_config.enable_independent_layer_evolution
    )
    native_genotype_config.initial_layer_count = genotype_config.layer_count[0]
    native_genotype_config.allowed_range_layer_count = genotype_config.layer_count[1]
    native_genotype_config.value_count_layer_count = genotype_config.layer_count[2]

    template_genotype = Genotype(config=native_genotype_config)
    population = initialize_population(
        population_initialization_method,
        template_genotype,
        population_size,
        cell_count,
        module_registry,
        transformer_encoder_type,
        prepend_positional_encoding,
        keep_positional_encoding_in_registry,
    )

    randomize_initial_population = evolution_config.randomize_initial_population
    if randomize_initial_population:
        non_randomized_genotype_count = evolution_config.non_randomized_genotype_count
        randomize_population(
            population[non_randomized_genotype_count:],
            module_registry.modules(),
            evolution_config.enable_independent_layer_evolution,
            evolution_config.init_layer_mutation_iterations,
            evolution_config.init_cell_mutation_iterations,
            evolution_config.init_module_mutation_iterations,
            evolution_config.init_parameter_mutation_iterations,
        )

    return population


def mutate_genotype(
    genotype: Genotype, config: EvolutionConfig, available_modules: List[GenotypeModule]
):
    # NOTE: Always reset the cached fitness after a mutation!!!

    # Layer mutations
    if (
        config.enable_independent_layer_evolution
        and random.random() < config.global_layer_mutation_probability
    ):
        if random.random() < config.add_layer_mutation_probability:
            add_new_layer(genotype)
            genotype.mark_dirty()

        if random.random() < config.duplicate_layer_mutation_probability:
            duplicate_random_layer(genotype)
            genotype.mark_dirty()

        if random.random() < config.swap_two_layers_mutation_probability:
            swap_two_random_layers(genotype)
            genotype.mark_dirty()

        if random.random() < config.remove_layer_mutation_probability:
            delete_random_layer(genotype)
            genotype.mark_dirty()

    # Cell mutations
    if random.random() < config.global_cell_mutation_probability:
        if random.random() < config.add_cell_mutation_probability:
            add_new_cell(genotype)
            genotype.mark_dirty()

        if random.random() < config.duplicate_cell_mutation_probability:
            duplicate_random_cell(genotype)
            genotype.mark_dirty()

        if random.random() < config.swap_two_cells_mutation_probability:
            swap_two_random_cells(genotype)
            genotype.mark_dirty()

        if random.random() < config.change_cell_inputs_mutation_probability:
            change_random_cell_inputs(genotype)
            genotype.mark_dirty()

        if random.random() < config.remove_cell_mutation_probability:
            delete_random_cell(genotype)
            genotype.mark_dirty()

    # Module mutations
    if random.random() < config.global_module_mutation_probability:
        if random.random() < config.add_module_mutation_probability:
            add_new_random_module(genotype, available_modules)
            genotype.mark_dirty()

        if random.random() < config.duplicate_module_mutation_probability:
            duplicate_random_module(genotype)
            genotype.mark_dirty()

        if random.random() < config.swap_two_modules_mutation_probability:
            swap_two_random_modules(genotype)
            genotype.mark_dirty()

        if random.random() < config.swap_module_for_new_mutation_probability:
            respect_module_category = (
                random.random() < config.respect_module_category_probability
            )
            swap_module_for_new_random_module(
                genotype, available_modules, respect_module_category
            )
            genotype.mark_dirty()

        if random.random() < config.remove_module_mutation_probability:
            remove_random_module(genotype)
            genotype.mark_dirty()

    # Parameter mutations
    if random.random() < config.global_parameter_mutation_probability:
        if random.random() < config.change_genotype_parameter_mutation_probability:
            change_random_parameter_genotype(genotype)
            genotype.mark_dirty()

        if random.random() < config.change_parameter_mutation_probability:
            change_random_parameter(genotype)
            genotype.mark_dirty()


def crossover_genotypes(
    genotype1: Genotype, genotype2: Genotype, config: EvolutionConfig
):
    if (
        config.enable_independent_layer_evolution
        and random.random() < config.swap_random_layers_crossover_probability
    ):
        swap_random_layers_crossover(genotype1, genotype2)
        genotype1.mark_dirty()
        genotype2.mark_dirty()

    if random.random() < config.swap_random_cells_crossover_probability:
        swap_random_cells_crossover(genotype1, genotype2)
        genotype1.mark_dirty()
        genotype2.mark_dirty()

    if random.random() < config.swap_random_branches_crossover_probability:
        swap_random_branches_crossover(genotype1, genotype2)
        genotype1.mark_dirty()
        genotype2.mark_dirty()


def main(config: GlobalConfig):
    # Configs
    dataset_config = DatasetConfig(**config.dataset)
    evaluation_config = EvaluationConfig(**config.evaluation)
    evolution_config = EvolutionConfig(**config.evolution)
    genotype_config = GenotypeConfig(**config.genotype)

    # Dataset
    batch_size = dataset_config.batch_size
    input_dim = dataset_config.input_dim
    output_dim = dataset_config.output_dim
    train_set, test_set = bootstrap_dataset_from_config(dataset_config)
    log.info(f">>> Dataset length: {len(train_set) + len(test_set)}")

    # Evaluation
    max_epochs = evaluation_config.max_epochs
    hurdle_after_iterations = evaluation_config.hurdle_after_iterations
    hurdles = []

    train_loss = Loss.from_str(evaluation_config.train_loss).module(reduction="sum")
    test_loss = Loss.from_str(evaluation_config.train_loss).module(reduction="sum")
    optimizer = torch.optim.AdamW(
        [torch.zeros(1, 1)], lr=0, betas=(0.9, 0.98), eps=1e-9
    )
    evaluation_config = EvaluationSetup()
    evaluation_config.batch_size = batch_size
    evaluation_config.input_dim = input_dim
    evaluation_config.output_dim = output_dim
    evaluation_config.max_epochs = max_epochs
    evaluation_config.train_set = train_set
    evaluation_config.test_set = test_set
    evaluation_config.train_loss = train_loss
    evaluation_config.test_loss = test_loss
    evaluation_config.optimizer = optimizer

    # Evolution
    iterations = evolution_config.iterations
    population_size = evolution_config.population_size
    parent_tournament_size = evolution_config.parent_tournament_size
    replacement_tournament_size = evolution_config.replacement_tournament_size
    replacement_candidate_selection = SelectionMode.from_str(
        evolution_config.replacement_candidate_selection
    )
    module_registry = ModuleRegistry.from_list(config.modules)

    # Initialize population
    init_number_of_cells = evolution_config.non_encoder_number_of_cells
    prepend_positional_encoding = evolution_config.prepend_positional_encoding
    keep_positional_encoding_in_registry = (
        evolution_config.keep_positional_encoding_in_registry
    )
    transformer_encoder_type = TransformerEncoderType.from_str(
        evolution_config.transformer_encoder_type
    )
    population_initialization_method = PopulationInitializationMethod.from_str(
        evolution_config.population_initialization_method
    )
    population = initialize_and_randomize_population(
        genotype_config,
        evolution_config,
        population_size,
        init_number_of_cells,
        module_registry,
        transformer_encoder_type,
        prepend_positional_encoding,
        keep_positional_encoding_in_registry,
        population_initialization_method,
    )

    # Tracking of progress via Tensorboard
    tb_run_timestamp = datetime.datetime.now().strftime("%d-%m-%Y_%H:%M:%S")
    tb_run_folder = f"./.runs/{tb_run_timestamp}/"
    tb = Tensorboard(log_dir=tb_run_folder)

    # Add global config to Tensorboard and also save as json
    config.save_as_json(f"{tb_run_folder}config.json")
    tb.add_json("config", config.to_dict(), 0)

    # Device
    device = "cuda" if torch.cuda.is_available() else "cpu"
    log.info(f">>> Used device: {device}")

    evolution_start = time.time()
    iteration_duration_history = []
    for i in range(iterations):
        iteration_start = time.time()
        iterations_left = iterations - (i + 1)
        mean_iteration_duration = mean(iteration_duration_history)
        evolution_time_left = mean_iteration_duration * iterations_left

        log.info("#######################################################")
        log.info(
            f"###### Iteration: {i + 1}, time left: {evolution_time_left:.2f}s ({(evolution_time_left/3600):.2f}h) ######"
        )
        log.info("#######################################################")

        # Add hurdle
        if i > 0 and i % hurdle_after_iterations == 0:
            log.info(f"%%% Evaluate population to obtain a hurdle %%%")
            evaluate_population(population, evaluation_config, device)
            mean_population_fitness = mean(
                [g.fitness() for g in population if not math.isnan(g.fitness())]
            )
            log.info(f"%%%%%% New hurdle ({mean_population_fitness}) %%%%%%")
            tb.add_scalar("hurdles", mean_population_fitness, i)
            hurdles.append(mean_population_fitness)

        # Selection
        _, parent1 = tournament_selection_with_hurdles(
            population,
            parent_tournament_size,
            SelectionMode.BEST_FITNESS,
            hurdles,
            evaluation_config,
            device,
        )
        _, parent2 = tournament_selection_with_hurdles(
            population,
            parent_tournament_size,
            SelectionMode.BEST_FITNESS,
            hurdles,
            evaluation_config,
            device,
        )

        # Important! Clone parents and set their timestamp
        child1 = copy.deepcopy(parent1)
        child2 = copy.deepcopy(parent2)
        child1.set_timestamp(i)
        child2.set_timestamp(i)

        # Apply crossover
        crossover_genotypes(child1, child2, evolution_config)

        # Apply mutations
        mutate_genotype(child1, evolution_config, module_registry.modules())
        mutate_genotype(child2, evolution_config, module_registry.modules())

        # Survivor selection
        _, survivor1 = tournament_selection_with_hurdles(
            [parent1, child1],
            2,
            SelectionMode.BEST_FITNESS,
            hurdles,
            evaluation_config,
            device,
        )
        _, survivor2 = tournament_selection_with_hurdles(
            [parent2, child2],
            2,
            SelectionMode.BEST_FITNESS,
            hurdles,
            evaluation_config,
            device,
        )

        individual_to_remove_index1, _ = tournament_selection_with_hurdles(
            population,
            replacement_tournament_size,
            replacement_candidate_selection,
            hurdles,
            evaluation_config,
            device,
        )
        population.pop(individual_to_remove_index1)

        individual_to_remove_index2, _ = tournament_selection_with_hurdles(
            population,
            replacement_tournament_size,
            replacement_candidate_selection,
            hurdles,
            evaluation_config,
            device,
        )
        population.pop(individual_to_remove_index2)

        population.insert(0, survivor1)
        population.append(survivor2)

        # Fitness tracking
        tb.add_scalar("survivor1_fitness", survivor1.fitness(), i)
        tb.add_scalar("survivor2_fitness", survivor2.fitness(), i)

        # Diversity tracking
        difference_factor = population_difference_factor(population)
        tb.add_scalar("difference_factor", difference_factor, i)
        log.info(f">>> Difference factor: {difference_factor:.6f}")

        # Save and visualize results
        if i == iterations - 1:
            evaluate_population(population, evaluation_config, device)
            sorted_population = list(
                sorted(population, key=lambda g: g.fitness(), reverse=True)
            )

            log.info(f">>> Best genotype fitness: {sorted_population[0].fitness()}")

            # Save top genotype jsons and visualize them
            best_genotypes = sorted_population[:10]
            for idx, g in enumerate(best_genotypes):
                g.save_as_json(f"{tb_run_folder}genotype_best{idx + 1}.json")
                tb.add_json(f"genotype_best{idx + 1}", json.loads(g.to_json()), i)
                visualize_genotype(g)

        iteration_duration = time.time() - iteration_start
        iteration_duration_history.append(iteration_duration)

    evolution_duration = time.time() - evolution_start
    log.info(f"### Evolution took: {evolution_duration:.2f}s ###")
