import torch
from torch import nn

from src.components.embedding.TimeSeriesPatchEmbedding import TimeSeriesPatchEmbedding
from src.components.transformer.EncoderFirstElementRegression import (
    EncoderFirstElementRegression,
)
from src.components.transformer.EncoderRegression import EncoderRegression
from src.evolutionary.phenotype.PhenotypeCell import PhenotypeCell
from src.evolutionary.phenotype.PhenotypeLayer import PhenotypeLayer


# NOTE: use_memory is VERY memory expensive, 8 GB might not be enough
class Phenotype(nn.Module):
    def __init__(
        self,
        embedding_dim: int,
        input_dim: int,
        output_dim: int,
        patch_size: int,
        use_memory: bool = False,
        use_full_regression: bool = True,
    ):
        super(Phenotype, self).__init__()
        self._embedding_dim = embedding_dim
        self._patch_size = patch_size
        self._use_memory = use_memory
        self._use_full_regression = use_full_regression

        self._embedding = TimeSeriesPatchEmbedding(embedding_dim, patch_size)

        if self._use_full_regression:
            self._regression = EncoderRegression(
                embedding_dim, input_dim, output_dim, patch_size
            )
        else:
            self._regression = EncoderFirstElementRegression(
                embedding_dim, input_dim, output_dim, patch_size
            )

        self._preprocess = nn.ModuleList()
        self._postprocess = nn.ModuleList()
        self._layers = nn.ModuleList()
        self._layer_count = 0

    def init_parameters(self):
        # Initialize parameters with Glorot / fan_avg
        for p in self.parameters():
            if p.dim() > 1:
                nn.init.xavier_uniform_(p)

    def get_regression(self):
        return self._regression

    def add_preprocess_module(self, module):
        self._preprocess.append(module)

    def add_postprocess_module(self, module):
        self._postprocess.append(module)

    def add_layer(self, layer):
        self._layers.append(layer)
        self._layer_count += 1

    def embedding_dim(self):
        return self._embedding_dim

    def patch_size(self):
        return self._patch_size

    # NOTE: Mask is not supported yet (and also not needed)
    def forward(self, x, mask=None):
        sequence_length = x.size(1)

        # Time series embedding + data preprocess
        x = self._embedding(x)
        for pre in self._preprocess:
            x = pre(x)

        # Inter layer memory from Transformer XL
        memory = []
        for i, layer in enumerate(self._layers):
            if self._use_memory:
                memory.append(x.detach())

            if i > 0 and self._use_memory:
                from_memory = memory[
                    -2
                ]  # Always want the second last i.e. memory from previous step
                memory.pop(0)  # Remove old snapshots to save memory
            else:
                from_memory = None

            # Concat input with memory if available
            if from_memory is not None:
                concatenated = torch.cat((from_memory, x), dim=1)
            else:
                concatenated = x

            x = layer(concatenated)

        # Remove the memory part
        if self._use_memory and len(self._layers) > 1:
            split_index = sequence_length // self._patch_size
            x = x[:, -split_index:, :]

        # Postprocess
        for post in self._postprocess:
            x = post(x)

        return x


if __name__ == "__main__":
    embedding_dim = 3

    cell1 = PhenotypeCell()
    layer = PhenotypeLayer()
    layer.add_cell(0, 0, cell1)

    phenotype = Phenotype(embedding_dim, 3, 3, 1)
    phenotype.add_layer(layer)

    t1 = torch.tensor([[1.0], [2.0], [3.0]])
    t2 = phenotype(t1, None)
    print(t2)
