import copy
from enum import Enum
from typing import List

import torch
import torch.nn as nn

# In general, RMSE and MSE with reduction="sum" seems to work the best. MSE seems to produce a bit smoother curve
# LogCosh also works well, also quite smooth curve. XSigmoid had similar performance
# Huber loss is another possibility (combines advantages of both L1Loss and MSELoss). Seems quite stable
# SmoothL1Loss looks good performance-wise (at least for decoder-only)

# torch.nn.L1Loss -> MAE
# torch.nn.MSE -> MSE


class Loss(str, Enum):
    MAE = "mae"
    MSE = "mse"
    RMSE = "rmse"
    SmoothL1 = "smoothl1"
    LogCosh = "logcosh"
    XTanh = "xtanh"
    XSigmoid = "xsigmoid"

    @staticmethod
    def from_str(string: str):
        return Loss(string)

    @staticmethod
    def module_list() -> List[Enum]:
        return [loss for loss in Loss]

    def module(self, reduction="mean"):
        registry = {
            Loss.MAE: torch.nn.L1Loss(reduction=reduction),
            Loss.MSE: torch.nn.MSELoss(reduction=reduction),
            Loss.RMSE: RMSELoss(reduction=reduction),
            Loss.SmoothL1: torch.nn.SmoothL1Loss(reduction=reduction),
            Loss.LogCosh: LogCoshLoss(),
            Loss.XTanh: XTanhLoss(),
            Loss.XSigmoid: XSigmoidLoss(),
        }
        return copy.deepcopy(registry[self])


class RMSELoss(nn.Module):
    def __init__(self, reduction="mean", eps=1e-12):
        super().__init__()
        # 'none' | 'mean' | 'sum'
        self.mse = nn.MSELoss(reduction=reduction)
        self.eps = eps

    def forward(self, output, target):
        loss = self.mse(output, target)
        return torch.sqrt(loss + self.eps)


class LogCoshLoss(nn.Module):
    def __init__(self, eps=1e-12):
        super().__init__()
        self.eps = eps

    def forward(self, output, target):
        diff = output - target
        return torch.mean(torch.log(torch.cosh(diff + self.eps)))


class XTanhLoss(nn.Module):
    def __init__(self):
        super().__init__()

    def forward(self, output, target):
        diff = output - target
        return torch.mean(diff * torch.tanh(diff))


class XSigmoidLoss(nn.Module):
    def __init__(self):
        super().__init__()

    def forward(self, output, target):
        diff = output - target
        return torch.mean(2 * diff / (1 + torch.exp(-diff)) - diff)
