import copy
import torch.nn as nn

from src.components.embedding.TimeSeriesPatchEmbedding import TimeSeriesPatchEmbedding
from src.components.transformer.Regression import Regression
from src.components.attention.ScaledDotProductMultiHeadAttention import (
    ScaledDotProductMultiHeadAttention1,
)
from src.components.transformer.FeedForward import FeedForward
from src.components.positional_encoding.ape.SinusoidalAbsolutePositionalEncoding import (
    SinusoidalAbsolutePositionalEncoding,
)
from src.components.transformer.Decoder import DecoderLayer, Decoder
from src.components.transformer.Encoder import EncoderLayer, Encoder


class TimeSeriesTransformer(nn.Module):
    def __init__(
        self, encoder, decoder, source_embedding, target_embedding, regression
    ):
        super(TimeSeriesTransformer, self).__init__()
        self.encoder = encoder
        self.decoder = decoder
        self.source_embedding = source_embedding
        self.target_embedding = target_embedding
        self.regression = regression

    @staticmethod
    def create(
        encoder_layers=4,
        decoder_layers=4,
        embedding_dim=20,
        forward_expansion=4,
        head_count=4,
        dropout=0.1,
        patch_size=1,
        xavier_param_init=True,
    ):
        attention = ScaledDotProductMultiHeadAttention1(embedding_dim, head_count)
        feed_forward = FeedForward(embedding_dim, forward_expansion, dropout)
        positional_encoding = SinusoidalAbsolutePositionalEncoding(
            embedding_dim, dropout
        )
        embedding = TimeSeriesPatchEmbedding(embedding_dim, patch_size)
        transformer = TimeSeriesTransformer(
            Encoder(
                EncoderLayer(
                    embedding_dim,
                    copy.deepcopy(attention),
                    copy.deepcopy(feed_forward),
                    dropout,
                ),
                encoder_layers,
            ),
            Decoder(
                DecoderLayer(
                    embedding_dim,
                    copy.deepcopy(attention),
                    copy.deepcopy(attention),
                    copy.deepcopy(feed_forward),
                    dropout,
                ),
                decoder_layers,
            ),
            nn.Sequential(
                copy.deepcopy(embedding),
                copy.deepcopy(positional_encoding),
            ),
            nn.Sequential(
                copy.deepcopy(embedding),
                copy.deepcopy(positional_encoding),
            ),
            Regression(embedding_dim, patch_size),
        )

        # Initialize parameters with Glorot / fan_avg
        if xavier_param_init:
            for p in transformer.parameters():
                if p.dim() > 1:
                    nn.init.xavier_uniform_(p)

        return transformer

    def get_regression(self):
        return self.regression

    def embedding_dim(self):
        return self.regression.embedding_dim

    def patch_size(self):
        return self.regression.patch_size

    def forward(self, source, target, source_mask, target_mask):
        return self.decode(
            self.encode(source, source_mask), source_mask, target, target_mask
        )

    def encode(self, source, source_mask):
        return self.encoder(self.source_embedding(source), source_mask)

    def decode(self, memory, source_mask, target, target_mask):
        return self.decoder(
            self.target_embedding(target), memory, source_mask, target_mask
        )
