import time

import numpy as np
import torch
from colorama import Fore, Style
from matplotlib import pyplot as plt
from torch import nn

from src.training.TransformerLossCompute import TransformerLossCompute
from src.training.evaluation import mae, rmse
from src.transformers.encoder_decoder.TimeSeriesTransformer import TimeSeriesTransformer
from src.dataset.SignalDataset import (
    SignalDataset,
    SignalType,
)
from src.training.NoamOptimizer import NoamOptimizer


class Batch:
    def __init__(
        self, source, target, split_size=50, padding_value=0.0, use_source_mask=False
    ):
        source_split, target_split = source.split(
            [source.size(1) - split_size, split_size], dim=1
        )
        self.source = source_split

        if use_source_mask:
            self.source_mask = (source_split != padding_value).unsqueeze(1)
        else:
            self.source_mask = None

        target_padding = torch.full(
            (target.size(0), target.size(1)),
            padding_value,
            device=target.device,
        )
        self.target = torch.cat(
            [
                target_split,
                target_padding,
            ],
            dim=1,
        )
        self.y_target = torch.cat(
            [
                target_split,
                target,
            ],
            dim=1,
        )

        self.target_mask = None
        self.y_token_count = self.y_target.data.sum()


def batch_iterator(dataset, batch_size, split_size, device):
    for i in range(0, len(dataset) - batch_size, batch_size):
        batch = [dataset[j] for j in range(i, i + batch_size)]
        x = np.array([b["x"] for b in batch], dtype=np.dtype("f4"))
        y = np.array([b["y"] for b in batch], dtype=np.dtype("f4"))

        source = torch.tensor(x, requires_grad=True).to(device=device)
        target = torch.tensor(y, requires_grad=True).to(device=device)
        yield Batch(source, target, split_size)


def run_epoch(dataset_iterator, model, loss_compute):
    """Helper for running one epoch"""
    total_tokens = 0
    total_loss = 0
    tokens = 0

    start = time.time()
    for i, batch in enumerate(dataset_iterator):
        output = model.forward(
            batch.source, batch.target, batch.source_mask, batch.target_mask
        )
        loss = loss_compute(output, batch.y_target, batch.y_token_count)
        total_loss += loss
        tokens += batch.y_token_count
        total_tokens += batch.y_token_count

        if i % 50 == 1:
            elapsed = time.time() - start
            print(
                "> #%d, loss: %f, tokens per sec: %f"
                % (i, loss / batch.y_token_count, tokens / elapsed)
            )
            start = time.time()
            tokens = 0

    return total_loss / total_tokens


def make_prediction(
    model, source, predict_tokens, split_size, use_mask=False, padding_value=0.0
):
    source_split, target_split = source.split(
        [source.size(1) - split_size, split_size], dim=1
    )

    if use_mask:
        source_mask = (source_split != padding_value).unsqueeze(1)
    else:
        source_mask = None
    memory = model.encode(source_split, source_mask)

    target_padding = torch.full(
        (target_split.size(0), predict_tokens),
        padding_value,
        device=target_split.device,
    )
    target = torch.cat(
        [
            target_split,
            target_padding,
        ],
        dim=1,
    )

    target_mask = None
    output = model.decode(
        memory,
        source_mask,
        target,
        target_mask,
    )
    values = model.regression(output)
    values = values.view(1, -1)
    prediction = values[:, split_size:]
    return prediction


def show_results(source, target, prediction):
    print("### Results ###")
    print(f"MAE: {mae(prediction, target)}")
    print(f"RMSE: {rmse(prediction, target)}")

    source_length = source.size
    target_length = target.size

    x_source = np.arange(0, source_length)
    x_target = np.arange(source_length, source_length + target_length)
    plt.plot(x_source, source)
    plt.plot(x_target, target)
    plt.plot(x_target, prediction)
    plt.show()


if __name__ == "__main__":
    print(Fore.LIGHTBLUE_EX, end="")
    print("### Train encoder-decoder ###")

    device = "cuda" if torch.cuda.is_available() else "cpu"
    print(f"# CUDA: {torch.cuda.is_available()}")

    # Dataset
    x_length = 200
    y_length = 400
    dataset = SignalDataset(SignalType.SINE, x_length=x_length, y_length=y_length)
    print(f"# Dataset length: {len(dataset)}")
    train_set, test_set = torch.utils.data.random_split(dataset, [17000, 2400])

    # Model
    batch_size = 20
    embedding_dim = 16
    split_size = 50
    patch_size = 1
    criterion = nn.SmoothL1Loss(reduction="sum")
    model = TimeSeriesTransformer.create(
        encoder_layers=2,
        decoder_layers=2,
        embedding_dim=embedding_dim,
        head_count=4,
        dropout=0.01,
        patch_size=patch_size,
    ).to(device)
    optimizer = torch.optim.AdamW(model.parameters(), lr=0, betas=(0.9, 0.98), eps=1e-9)
    scheduled_optimizer = NoamOptimizer(
        optimizer,
        embedding_dim,
        1000,
    )

    # Train
    epochs = 3
    for epoch in range(epochs):
        print(Fore.LIGHTRED_EX, end="")
        print(f"### Epoch: {epoch + 1} ###")

        print(Style.RESET_ALL, end="")
        model.train()
        train_loss = run_epoch(
            batch_iterator(train_set, batch_size, split_size, device),
            model,
            TransformerLossCompute(model.regression, criterion, scheduled_optimizer),
        )
        print(Fore.LIGHTRED_EX, end="")
        print(f"## Train acc loss: {train_loss} ##")

        print(Fore.LIGHTYELLOW_EX, end="")
        model.eval()
        eval_loss = run_epoch(
            batch_iterator(test_set, batch_size, split_size, device),
            model,
            TransformerLossCompute(model.regression, criterion, None),
        )
        print(Fore.LIGHTRED_EX, end="")
        print(f"## Eval acc loss: {eval_loss} ##")
        print(Style.RESET_ALL, end="")

    model.eval()
    test_sample = test_set[0]
    target = test_sample["y"]
    source = test_sample["x"]
    source_tensor = torch.Tensor(test_sample["x"]).unsqueeze(0).to(device=device)
    prediction_tensor = make_prediction(
        model, source_tensor, predict_tokens=y_length, split_size=split_size
    )
    prediction = prediction_tensor.squeeze(0).detach().cpu().numpy()

    show_results(source, target, prediction)
