import time

import numpy as np
import torch
from colorama import Fore, Style
from matplotlib import pyplot as plt
from torch import nn

from src.dataset.SignalDataset import (
    SignalDataset,
    SignalType,
)
from src.training.NoamOptimizer import NoamOptimizer
from src.training.evaluation import rmse, mae
from src.training.TransformerLossCompute import TransformerLossCompute
from src.transformers.encoder_only.TimeSeriesEncoderTransformer import (
    TimeSeriesEncoderTransformer,
)


class Batch:
    # NOTE: The padding value might interfere with dataset values. Choose carefully!
    def __init__(self, source, target, padding_value=0.0):
        self.source = source
        self.source_mask = source != padding_value
        self.y_target = target
        self.y_token_count = (target != padding_value).data.sum()


def batch_iterator(dataset, batch_size, input_dim, device):
    for i in range(0, len(dataset) - batch_size, batch_size):
        batch = [dataset[j] for j in range(i, i + batch_size)]
        x = np.array([[[s] for s in b["x"]] for b in batch], dtype=np.dtype("f4"))
        y = np.array([[[s] for s in b["y"]] for b in batch], dtype=np.dtype("f4"))

        source = torch.tensor(x, requires_grad=True).to(device=device)
        target_y = torch.tensor(y, requires_grad=True).to(device=device)

        sequence_length = source.size(1)
        if sequence_length < input_dim:
            source = nn.functional.pad(
                source, (0, 0, 0, input_dim - sequence_length), "constant", 0
            )

        yield Batch(source, target_y)


def run_epoch(dataset_iterator, model, loss_compute, patch_size: int, logger=None):
    """Helper for running one epoch"""
    total_tokens = 0
    total_loss = 0
    tokens = 0

    start = time.time()
    for i, batch in enumerate(dataset_iterator):
        mask = batch.source_mask if patch_size == 1 else None
        output = model.forward(batch.source, mask)
        loss = loss_compute(output, batch.y_target, batch.y_token_count)
        total_loss += loss
        tokens += batch.y_token_count
        total_tokens += batch.y_token_count

        if i % 50 == 1:
            elapsed = time.time() - start
            loss = loss / batch.y_token_count
            tokens_per_sec = tokens / elapsed
            if logger:
                logger.debug(
                    f"> {i}, loss: {loss:.6f}, tokens per sec: {tokens_per_sec}"
                )
            else:
                print(f"> {i}, loss: {loss:.6f}, tokens per sec: {tokens_per_sec}")
            start = time.time()
            tokens = 0

    return total_loss / total_tokens


def make_prediction(model, source, input_dim, patch_size, padding_value=0.0):
    sequence_length = source.size(1)
    if sequence_length < input_dim:
        source = nn.functional.pad(
            source, (0, input_dim - sequence_length), "constant", 0
        )
    source_mask = (source != padding_value).unsqueeze(-1) if patch_size == 1 else None

    output = model(
        source,
        source_mask,
    )
    values = model.get_regression()(output)
    return values


def show_results(source, target, prediction):
    print("### Results ###")
    print(f"MAE: {mae(prediction, target)}")
    print(f"RMSE: {rmse(prediction, target)}")

    source_length = source.size
    target_length = target.size

    x_source = np.arange(0, source_length)
    x_target = np.arange(source_length, source_length + target_length)
    plt.plot(x_source, source)
    plt.plot(x_target, target)
    plt.plot(x_target, prediction)
    plt.show()


if __name__ == "__main__":
    print(Fore.LIGHTBLUE_EX, end="")
    print("### Train encoder only ###")

    device = "cuda" if torch.cuda.is_available() else "cpu"
    print(f"# CUDA: {torch.cuda.is_available()}")

    # Dataset
    x_length = 200
    y_length = 400
    dataset = SignalDataset(SignalType.SINE, x_length=x_length, y_length=y_length)
    print(f"# Dataset length: {len(dataset)}")
    train_set, test_set = torch.utils.data.random_split(dataset, [17000, 2400])

    # Model
    batch_size = 20
    input_dim = 200
    output_dim = 400
    embedding_dim = 16
    # NOTE: Patch size >1 doesn't support masking
    patch_size = 1
    criterion = nn.SmoothL1Loss(reduction="sum")
    model = TimeSeriesEncoderTransformer.create(
        layers=4,
        embedding_dim=embedding_dim,
        input_dim=input_dim,
        output_dim=output_dim,
        head_count=4,
        dropout=0.01,
        patch_size=patch_size,
    ).to(device)
    optimizer = torch.optim.AdamW(model.parameters(), lr=0, betas=(0.9, 0.98), eps=1e-9)
    scheduled_optimizer = NoamOptimizer(
        optimizer,
        embedding_dim,
        1000,
    )

    # Train
    epochs = 2
    for epoch in range(epochs):
        print(Fore.LIGHTRED_EX, end="")
        print(f"### Epoch: {epoch + 1} ###")

        start = time.time()

        print(Style.RESET_ALL, end="")
        model.train()
        train_loss = run_epoch(
            batch_iterator(train_set, batch_size, input_dim, device),
            model,
            TransformerLossCompute(model.regression, criterion, scheduled_optimizer),
            patch_size,
        )
        print(Fore.LIGHTRED_EX, end="")
        print(f"## Train acc loss: {train_loss} ##")

        print(Fore.LIGHTYELLOW_EX, end="")
        model.eval()
        eval_loss = run_epoch(
            batch_iterator(test_set, batch_size, input_dim, device),
            model,
            TransformerLossCompute(model.regression, criterion, None),
            patch_size,
        )
        print(Fore.LIGHTRED_EX, end="")
        print(f"## Eval acc loss: {eval_loss} ##")
        print(f"## Epoch took: {time.time() - start}s ##")
        print(Style.RESET_ALL, end="")

    model.eval()
    test_sample = test_set[0]
    target = test_sample["y"]
    source = test_sample["x"]
    source_tensor = torch.Tensor(test_sample["x"]).unsqueeze(0).to(device=device)
    prediction_tensor = make_prediction(model, source_tensor, input_dim, patch_size)
    prediction = prediction_tensor.squeeze(0).detach().cpu().numpy()

    show_results(source, target, prediction)
